# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.core import menu, plugin, interfaces, media_uri
from elisa.core import common, utils
from elisa.core import media_uri
from elisa.core import task
from elisa.core.gst import data_access_src
from elisa.core.utils.misc import handle_failure, get_host_address
from elisa.extern import natural_sort
from elisa.core.interfaces import *


from elisa.core.signals.data import new_device, del_device
from elisa.core.signals.player import stopped_signal, playing_signal
from elisa.core.signals.skin import new_menu_item, del_menu_item
from elisa.core.signals.application import application_started

from pgm import render

from twisted.internet import threads
import mutex

BANNED_URI_SCHEMES = []

def uri_cacheable(uri):
    banned = False
    if not isinstance(uri, media_uri.Uri):
        uri = media_uri.Uri(location=uri)
    for banned_scheme in BANNED_URI_SCHEMES:
        if uri.scheme.startswith(banned_scheme):
            banned = True
            break
    return not banned

class TreePlugin(plugin.Plugin, menu.MenuItem):
    " tree navigation Plugin Class "

    data_access = plugin.ExtensionPoint(interfaces.IDataAccess)

    icons = ['remove', 'rate_up', 'rate_down', ]

    item_filter = None
    item_icon_path = None
    item_folder_path = None
    item_action = None

    parent = None
    menu_name = "Default"
    menu_help_string = ""
    
    def __init__(self):
        """
        Constructor
        """
        plugin.Plugin.__init__(self)
        menu.MenuItem.__init__(self)
        
        new_device.connect(self.got_new_device)
        del_device.connect(self.device_gone)

        self._locations = []
        self.set_short_name(self.get_menu_name())
        self.set_help_string(self.menu_help_string)
        
        self.load_icons()

        # FIXME: philn, you must do it better !!!
        self.set_icon_path(self.get_name() + '.png')
        self.set_target_path(self.get_name())

        locations = self.get_config().get('locations',[])

        for location in locations:                
                    
            if location.endswith('*'):
                self.add_sub_locations(location)
                l = location[:-1]
            else:
                self.add_location(location)
                l = location
                
            if l not in self._locations:
                self._locations.append(l)
            
                
    def post_init(self):
        plugin_manager = common.get_plugin_manager()
        loaded = plugin_manager.is_registered(self.__class__)
        if loaded:
            if self.parent:
                parent_plugin = plugin_manager.get_plugin_with_name(self.parent)
                if parent_plugin:
                    if isinstance(parent_plugin, TreePlugin):
                        parent_plugin.add_item(self)

            self.add_wizard_shortcuts()
                 
        plugin.Plugin.post_init(self)

    def add_wizard_shortcuts(self):
        # add content menu
        icon_name = 'add_%s_folder.png' % self.get_name()
        add_folder = menu.MenuItem(short_name=_("Add folder"))
        add_folder.set_target_path('add_folder')
        add_folder.set_icon_path(icon_name)
        add_folder.set_action_callback(self.add_folder_wizard)
        #self.insert_item(0, add_folder)
        
    def get_menu_name(self):
        return self.menu_name

    def got_new_device(self, resource_name, location, plugin_name):
        self.logger.debug("[%s] Got new device at %s" % (self.name,location))
        locations = self.get_config().get('locations',[])
        parent = self.get_parent()
        if not locations:
            self.logger.debug("[%s] Empty menu... this will be fixed, sorry" % self.name)
            return
        
        uri = media_uri.Uri(location)

        icon_path = None
        data_access = self.get_data_plugin(location)
        if data_access and data_access.root_folder_icon:
            icon_path = data_access.root_folder_icon
        elif uri.scheme == 'file':
            icon_path = 'generic-usb.png'

        if isinstance(parent, menu.MenuTree):
            if (not plugin_name) or (plugin_name == self.get_name()):
                self.add_location(location, resource_name, icon_path=icon_path,
                                  emit_signal=True)
            else:
                self.logger.debug("Not adding location in plugin %r" % plugin_name)
        else:
            self.logger.debug("Not a menu tree: %r" % parent)

    def device_gone(self, resource_name, location, plugin_name):
        locations = self.get_config().get('locations',[])
        if locations and not plugin_name:
            self.del_location(location, emit_signal=True)
            
    def load_icons(self):
        for icon_name in self.icons:
            name = "%s_icon" % icon_name
            setattr(self, name, "%s.png" % icon_name)

    def get_data_plugin(self, location):
        # extract the type or uri (file:// .. smb:// ....)
        if isinstance(location, media_uri.Uri):
            uri = location
        else:
            uri = media_uri.Uri(location=location)

        # look for the plugin supporting this type of uri
        data_access_plugin = self.data_access.get("uri:%s" % uri.scheme)

        return data_access_plugin

    def get_locations(self):
        return self._locations

    def add_location(self, location, short_name=None, icon_path=None,
                     emit_signal=False, check_cache=True):
        if location in self:
            self.logger.info("[%s] Location %r already added" % (self.get_name(),
                                                                 location))
            return

        media_manager = common.get_application().get_media_manager()

        available = True
        if check_cache and media_manager.enabled() and uri_cacheable(location):
            source = media_manager.get_source_for_uri(location)
            if source:
                available = source.available
                if emit_signal and not available:
                    media_manager.add_source(source.uri, self)
                    available = True
                # TODO: re-enable this when we have acceptable monitoring support
##                 if available:
##                     media_manager.monitor_source(media_uri.Uri(source.uri))
                media_manager.update_source(source.uri, self, True)
            else:
                data_plugin = self.get_data_plugin(location)
                if data_plugin:
                    if not data_plugin.validate_uri(location):
                        self.logger.info("Invalid URI: %r" % location)
                        return
                
                media_manager.add_source(location, self)
                
        if not available:
            self.logger.debug("Source at %r doesn't seem to be available" % location)
            return

        if short_name:
            loc_name = "%r at" % short_name
        else:
            loc_name = ''
            
        self.logger.info("[%s] Adding location %s %s" % (self.get_name(),
                                                         loc_name,
                                                         repr(location)))
            
        def add_item(new_item, called_from):
            if new_item and new_item not in called_from:
                if icon_path:
                    new_item.set_icon_path(icon_path)
                elif data_plugin.root_folder_icon:
                    new_item.set_icon_path(data_plugin.root_folder_icon)
                called_from.add_item(new_item)
                if location not in self._locations:
                    self._locations.append(location)
                    if emit_signal:
                        index = called_from.get_items_count() - 1
                        new_menu_item.emit(new_item, index)

        data_plugin = self.get_data_plugin(location)
        if data_plugin:
            ok = False
            if not data_plugin.validate_uri(location):
                self.logger.info("Invalid URI: %r" % location)
                return
            elif not data_plugin.is_dir(location):
                handle = data_plugin.open(location)
                if handle:
                    ok = True
                    data_plugin.close(handle)
            else:
                ok = True
            if ok:
                item = data_plugin.create_menu_for_uri(uri=location,
                                                       for_plugin=self,
                                                       short_name=short_name)
                add_item(item, self)
            else:
                self.logger.info("Not found: %s" % location)
        else:
            self.logger.info("No data access plugin found for : %s" % location)

    def add_sub_locations(self, head_uri):
        head_uri = head_uri[:-1]
        
        data_plugin = self.get_data_plugin(head_uri)
        if data_plugin:
            media_manager = common.get_application().get_media_manager()

            available = True
            if media_manager.enabled() and uri_cacheable(head_uri):
                source = media_manager.get_source_for_uri(head_uri)
                if source:
                    available = source.available
                else:
                    media_manager.add_source(head_uri, self)

            if available:
                children = data_plugin.get_direct_children(head_uri)
                locations = children.keys()
                locations.sort()
                for location in locations:
                    typ, media_type, name, index = children[location]
                    self.add_location(location, name, check_cache=False)

    def del_location(self, location, emit_signal=False):
        self.logger.info("[%s] Deleting location %s" % (self.get_name(),
                                                        repr(location)))
        uri = str(media_uri.Uri(location))
        menu_item = self.get_item_with_target(uri)
        if emit_signal:
            if menu_item:
                del_menu_item.emit(menu_item)

            media_manager = common.get_application().get_media_manager()

            if media_manager.enabled() and uri_cacheable(uri):
                source = media_manager.get_source_for_uri(uri)
                if source:
                    media_manager.remove_source(uri, self)

        self.del_item_with_target(uri)

    def handle_failure(self, failure):
        self.logger.info("Error : %s" % failure.getTraceback())
        failure.trap(Exception)

    def item_action_menu(self, menu_item):

        rate_up = menu.MenuItem(short_name=_("Rate+"))
        rate_up.set_target_path("rate+")
        rate_up.set_icon_path(self.rate_up_icon)
        rate_up.set_action_callback(self.rate_up_parent_item)

        rate_down = menu.MenuItem(short_name=_("Rate-"))
        rate_down.set_target_path("rate-")
        rate_down.set_icon_path(self.rate_down_icon)
        rate_down.set_action_callback(self.rate_down_parent_item)

        remove = menu.MenuItem(short_name=_("Remove"))
        remove.set_target_path("remove")
        remove.set_icon_path(self.remove_icon)
        remove.set_action_callback(self.remove_parent_item)

        #menu_item.add_item(rate_up)
        #menu_item.add_item(rate_down)
        menu_item.add_item(remove)
        
    def _set_rating_delta(self, location, delta):
        media_manager = common.get_application().get_media_manager()
        media_node = media_manager.get_media_node_for_uri(location)
        if media_node:
            new_rating = media_node.rating + delta
            if new_rating in range(101):
                media_manager.update_db_row(media_node, rating=new_rating)
                self.logger.info("New rating for %s : %s" % (location,
                                                             new_rating))


    def add_folder_wizard(self, menu_item):
        print 'add folder', menu_item
        
    def rate_up_parent_item(self, menu_item):
        parent = menu_item.get_parent()
        location = parent.get_target_path()
        self._set_rating_delta(location, 5)
        
    def rate_down_parent_item(self, menu_item):
        parent = menu_item.get_parent()
        location = parent.get_target_path()
        self._set_rating_delta(location, -5)

    def remove_parent_item(self, menu_item):
        """ TODO:
        - update widgets?
        """
        def on_ok():
            parent = menu_item.get_parent()
            location = parent.get_target_path()
            data_plugin = self.get_data_plugin(location)
            if data_plugin:
                #data_plugin.remove_location(location)
                print 'remove!'

        def on_cancel():
            pass
                
        appli = common.get_application()
        lines = (_("Are you sure you want to delete this file?"),
                 "", menu_item.get_parent().get_short_name())
        appli.get_skin().open_popup(lines, ok_callback=on_ok,
                                    cancel_callback=on_cancel)

    def get_folder_thumbnail(self, uri, widget_item, size):
        pass

    def get_item_thumbnail(self, uri, widget_item, size):
        pass


class PlayerPlugin(TreePlugin):

    icons = ['pause', 'play', 'fastb', 'fastf', 'seekf', 'seekb',
             'stop', 'volume_up', 'volume_down', 'previous_track',
             'next_track'] + TreePlugin.icons

    def __init__(self):
        TreePlugin.__init__(self)
        stopped_signal.connect(self._got_stop_signal)
        playing_signal.connect(self._got_playing_signal)
        self._mutex = mutex.mutex()

    def unload(self):
        if self.is_locked():
            self.unlock()
        return TreePlugin.unload(self)

    def lock(self):
        r = self._mutex.testandset()
        if r:
            self.logger.debug("Locking player")
        return r

    def is_locked(self):
        return self._mutex.locked
    
    def unlock(self):
        self.logger.debug("Unlocking player")
        self._mutex.unlock()

    def seek_forward(self, step):
        player = self.get_player()
        player.seek_forward(step)

    def seek_backward(self, step):
        player = self.get_player()
        player.seek_backward(step)

    def _got_stop_signal(self, playable):
        uri = playable.get_uri()
        self.stopped(location=uri)

    def _got_playing_signal(self, playable, started):
        uri = playable.get_uri()
        menu_item = self.get_item_with_target(uri)
        if menu_item:
            self.playing(menu_item)
        
    def item_filter(self, uri):
        uri = media_uri.Uri(location=uri)
        data_plugin = self.get_data_plugin(uri)
        is_item = False
        if data_plugin and not data_plugin.is_dir(str(uri)):
            is_item = True
        return is_item

    def folder_action_menu(self, menu_item):
        play_all_image = 'play_all.png'

        play_all = menu.MenuItem(short_name="Play all")
        play_all.set_target_path("play_all")
        play_all.set_icon_path(play_all_image)
        play_all.set_action_callback(self.play_all_parent_item)

        if not menu_item.get_target_path().startswith('upnp://'):
            menu_item.add_item(play_all)

    def item_action_menu(self, menu_item):
        
        play_pause = menu.MenuItem(short_name=_("Play"))
        play_pause.set_target_path("play/pause")
        play_pause.set_icon_path(self.play_icon)
        play_pause.set_action_callback(self.play_parent_item)

        fastf = menu.MenuItem(short_name=_("Speed+"))
        fastf.set_target_path("speed+")
        fastf.set_icon_path(self.fastf_icon)
        fastf.set_action_callback(self.fastf_parent_item)
        fastf.disable()

        fastb = menu.MenuItem(short_name=_("Speed-"))
        fastb.set_target_path("speed-")
        fastb.set_icon_path(self.fastb_icon)
        fastb.set_action_callback(self.fastb_parent_item)
        fastb.disable()

        seekf = menu.MenuItem(short_name=_("Seek+"))
        seekf.set_target_path("seek+")
        seekf.set_icon_path(self.seekf_icon)
        seekf.set_action_callback(self.seekf_parent_item)
        seekf.disable()

        seekb = menu.MenuItem(short_name=_("Seek-"))
        seekb.set_target_path("seek-")
        seekb.set_icon_path(self.seekb_icon)
        seekb.set_action_callback(self.seekb_parent_item)
        seekb.disable()

        next = menu.MenuItem(short_name=_("Next"))
        next.set_target_path("next")
        next.set_icon_path(self.next_track_icon)
        next.set_action_callback(self.next_parent_item)
        next.disable()

        prev = menu.MenuItem(short_name=_("Previous"))
        prev.set_target_path("prev")
        prev.set_icon_path(self.previous_track_icon)
        prev.set_action_callback(self.prev_parent_item)
        prev.disable()


        stop = menu.MenuItem(short_name=_("Stop"))
        stop.set_target_path("stop")
        stop.set_icon_path(self.stop_icon)
        stop.set_action_callback(self.stop_parent_item)
        stop.disable()

        volp = menu.MenuItem(short_name=_("Volume+"))
        volp.set_target_path("vol+")
        volp.set_icon_path(self.volume_up_icon)
        volp.set_action_callback(self.volp_parent_item)
        volp.disable()

        volm = menu.MenuItem(short_name=_("Volume-"))
        volm.set_target_path("vol-")
        volm.set_icon_path(self.volume_down_icon)
        volm.set_action_callback(self.volm_parent_item)
        volm.disable()

        menu_item.add_item(prev)        
        menu_item.add_item(seekb)
        menu_item.add_item(fastb)
        menu_item.add_item(play_pause)
        menu_item.add_item(fastf)
        menu_item.add_item(seekf)
        menu_item.add_item(next)
        menu_item.add_item(stop)
        menu_item.add_item(volm)
        menu_item.add_item(volp)

        TreePlugin.item_action_menu(self, menu_item)
        
        player = self.get_player()
        if player and str(menu_item.get_target_path()) == \
               str(player.get_playing_uri()):
            self.playing(menu_item)

    def play_item(self, menu_item):
        """ By default, update the buttons of the action
        menu. Inheriting player plugins should override this to
        actually play something.
        """
        self.playing(menu_item)

    def play_parent_item(self, menu_item):
        self.play_item(menu_item.get_parent())

        menu_item.set_short_name(_("Pause"), True)
        menu_item.set_icon_path(self.pause_icon, True)
        menu_item.set_action_callback(self.pause_parent_item, True)

    def pause_parent_item(self, menu_item):
        self.pause_item(menu_item.get_parent())
        menu_item.set_short_name(_("Play"), True)
        menu_item.set_icon_path(self.play_icon, True)
        menu_item.set_action_callback(self.play_parent_item, True)

    def fastf_parent_item(self, menu_item):
        self.fastf_item(menu_item.get_parent())

    def fastb_parent_item(self, menu_item):
        self.fastb_item(menu_item.get_parent())

    def seekf_parent_item(self, menu_item):
        self.seekf_item(menu_item.get_parent())

    def seekb_parent_item(self, menu_item):
        self.seekb_item(menu_item.get_parent())

    def next_parent_item(self, menu_item):
        self.next_item(menu_item.get_parent())

    def prev_parent_item(self, menu_item):
        self.prev_item(menu_item.get_parent())

    def stop_parent_item(self, menu_item):
        self.stop_item(menu_item.get_parent())

    def volm_parent_item(self, menu_item):
        self.volm_item(menu_item.get_parent())

    def volp_parent_item(self, menu_item):
        self.volp_item(menu_item.get_parent())

    def get_player(self):
        return common.get_application().get_player()

    def volm_item(self, menu_item):
        player = self.get_player()
        player.set_volume_incr(-5)

    def volp_item(self, menu_item):
        player = self.get_player()
        player.set_volume_incr(5)

    def pause_item(self, menu_item):
        player = self.get_player()
        player.pause()

    def fastf_item(self, menu_item):
        player = self.get_player()
        player.increment_playback_speed(0.1)

    def fastb_item(self, menu_item):
        player = self.get_player()
        player.decrement_playback_speed(0.1)


    def seekf_item(self, menu_item):
        player = self.get_player()
        player.seek_forward()

    def seekb_item(self, menu_item):
        player = self.get_player()
        player.seek_backward()

    def next_item(self, menu_item):
        player = self.get_player()
        player.next(at_user_request=True)
        
    def prev_item(self, menu_item):
        player = self.get_player()
        player.previous()

    def stop_item(self, menu_item):
        player = self.get_player()
        player.stop()
        player.close()
        self.stopped(menu_item)

    def stopped(self, menu_item=None, location=None):
        if menu_item:
            for target in ("seek-", "seek+", "speed-", "speed+", "stop", "vol+",
                           "vol-"):
                item = menu_item.get_item_with_target(target)
                if item:
                    item.disable()

            play_pause = menu_item.get_item_with_target("play/pause")
            if play_pause:
                play_pause.set_icon_path(self.play_icon, True)
                play_pause.set_action_callback(self.play_parent_item, True)
                play_pause.set_short_name(_("Play"), True)
            location = menu_item.get_target_path()
        
        manager = common.get_application().get_task_manager()
        task = manager.get_task_with_target_path(location)
        if task:
            manager.del_task(task)
            
    def playing(self, menu_item):
        for target in ("seek-", "seek+", "speed+", "speed-", "stop", "vol+",
                       "vol-"):
            item = menu_item.get_item_with_target(target)
            if item:
                item.enable()
                
        play_pause = menu_item.get_item_with_target("play/pause")
        play_pause.set_short_name(_("Pause"), True)
        play_pause.set_icon_path(self.pause_icon, True)
        play_pause.set_action_callback(self.pause_parent_item, True)

        common.get_application().get_task_manager().add_task(task.Task(menu_item))
        

    def play_all_parent_item(self, menu_item):
        self.play_all_item(menu_item.get_parent())

    def play_all_item(self, menu_item):
        location_path = menu_item.get_target_path()
        
        player = common.get_application().get_player()
        player.stop()
        player.set_uri(location_path)
        player.next()

    folder_action = play_all_item

    def en_queue_item(self, menu_item):
        pass

class DataAccessPlugin(plugin.Plugin):

    root_folder_icon = ''

    uncacheable_uri_schemes = []
    
    data_access = plugin.ExtensionPoint(interfaces.IDataAccess)

    gst_source = None

    def __init__(self):
        for feature in self.provides:
            parsed = feature.split(':')
            if len(parsed) == 2 and parsed[0] == 'uri':
                media_uri.register_scheme(parsed[1])
        plugin.Plugin.__init__(self)
        self._history = {}

        for scheme in self.uncacheable_uri_schemes:
            if scheme not in BANNED_URI_SCHEMES:
                BANNED_URI_SCHEMES.append(scheme)

    def validate_uri(self, uri):
        return True

    def get_data_plugin(self, location):
        # extract the type or uri (file:// .. smb:// ....)
        if isinstance(location, media_uri.Uri):
            uri = location
        else:
            uri = media_uri.Uri(location=location)

        # look for the plugin supporting this type of uri
        data_access_plugin = self.data_access.get("uri:%s" % uri.scheme)

        return data_access_plugin


    def _add_to_history(self, head_uri, uri, short_name):
        if head_uri not in self._history:
            self._history[head_uri] = []
        self._history[head_uri].append((uri, short_name))
        
    def get_cached_children(self, uri, for_plugin, short_name):
        children = {}
        media_manager = common.get_application().get_media_manager()
        media = media_manager.get_media_for_uri(uri)
        if not media:
            media_manager.add_source(uri, for_plugin, short_name)
        else:
            children = media_manager.get_children(media, uri, for_plugin)
        return children

    def create_child_menuitems_for_uri(self, uri, start_item, nb_items,
                                       for_plugin, short_name, menu_item=None):
        """
        Build a menu representing the data holded at the given location.
        """
        items = []
        current_item_rank = 0
        folder_action_menu = for_plugin.folder_action_menu

        media_manager = common.get_application().get_media_manager()

        if uri_cacheable(uri) and media_manager.enabled() and \
               not str(uri).startswith('file://'):
            children = self.get_cached_children(uri, for_plugin, short_name)
        else:
            children = self.get_direct_children(uri)

        for child_uri, infos in children.iteritems():
            (typ, media_type, filename, index) = infos
            if current_item_rank >= start_item and \
                   (len(items) < nb_items or nb_items < 0):

                new_item = self.create_menu_for_uri(child_uri, for_plugin,
                                                    filename)
                if new_item:
                    items.append(new_item)

            if nb_items > 0 and nb_items <= len(items):
                break

            current_item_rank += 1

        # split folders and normal items
        new_items = []
        new_folders = []
        for item in items:
            folder_icon_path = for_plugin.folder_icon_path
            # FIXME: better way to detect if it is a folder or not!
            if item.get_icon_path() == folder_icon_path:
                new_folders.append(item)
            else:
                new_items.append(item)
                
        key = lambda item: item.get_short_name().lower()
        new_folders = natural_sort.obj_natsorted(new_folders,
                                                 natural_sort.natcasecmp,
                                                 key)
        new_items = natural_sort.obj_natsorted(new_items,
                                               natural_sort.natcasecmp,
                                               key)
        
        # put folders first
        items = new_folders + new_items
            
        # build action menu if folder
        if items and folder_action_menu:
            folder_action_menu(menu_item)

        return items

    def create_menu_for_uri(self, uri, for_plugin, short_name=None):
        item_filter = for_plugin.item_filter
        item_icon_path = for_plugin.item_icon_path
        folder_icon_path = for_plugin.folder_icon_path
        item_thumbnailer = for_plugin.get_item_thumbnail
        folder_thumbnailer = for_plugin.get_folder_thumbnail
        item_action = for_plugin.item_action
        folder_action = for_plugin.folder_action
        item_action_menu = for_plugin.item_action_menu
        
        media_manager = common.get_application().get_media_manager()
        if media_manager.enabled():
            media = media_manager.get_media_for_uri(str(uri))
            # XXX: we should display non available medias using a special icon
            #      instead of displaying nothing
            if media and media.deleted:
                return None
        else:
            media = None

        if not short_name:
            if media:
                short_name = media.short_name
            else:
                short_name = media_uri.Uri(str(uri)).get_label()
        
        is_dir = self.is_dir(uri)
        is_item = item_filter(uri)

        item = menu.MenuItem(short_name=short_name)
        if is_dir or is_item:

            if is_dir:
                icon_path = folder_icon_path
                thumbnailer = folder_thumbnailer
                item.set_action_callback(folder_action)
            else:
                icon_path = item_icon_path
                thumbnailer = item_thumbnailer
                item.set_action_callback(item_action)

            callback = self.create_child_menuitems_for_uri
            item.set_loadmore_callback(callback, start_item=0,
                                       nb_items=-1,
                                       for_plugin=for_plugin,
                                       short_name=short_name)

            uri = utils.misc.fs_encode(str(uri))
                        
            item.set_target_path(uri)
            item.set_icon_path(icon_path)
            item.set_thumbnailer(thumbnailer)

            # build action menu if leaf
            if is_item and item_action_menu:
                item_action_menu(item)
            return item

        return None

    def has_children(self, uri):
        children = self.get_direct_children(uri)
        return len(children) > 0

    def get_parent(self, uri):
        # FIXME: implement this correctly in all data_access plugins
        return None

    def previous_location(self, uri):
        history = self._history.get(uri,[])
        previous_uri = None
        short_name = None
        if history and len(history) > 1:
            previous_uri, short_name = history[-2]
            self._history[uri].remove((previous_uri, short_name))
        return previous_uri, short_name

    def get_file_type(self, uri):
        """
        Return the type of file. If directory return empty string, either
        return the content-type of the uri
        """
        return ''
    
    def close(self, data_access_file):
        """
        Close the DataAccessFile
        """
        handle = data_access_file.get_file_handle()
        handle.close()

    def seek(self, data_access_file, offset, whence=0):
        """
        Close the DataAccessFile
        """
        handle = data_access_file.get_file_handle()
        handle.seek(offset, whence)

    def read(self, data_access_file, size=-1):
        """
        Close the DataAccessFile
        """
        handle = data_access_file.get_file_handle()
        return handle.read(size)

    def get_gst_pipeline(self):
        return None

    def get_gst_source(self):
        # FIXME: avoid recreating the source multiple times
        if not self.gst_source:
            self.gst_source = data_access_src.DataAccessFileSource('src')

        #self.gst_source = data_access_gst.DataAccessFileSource('src')
            
        return self.gst_source

    def monitor_uri(self, uri, callback):
        pass

    def unmonitor_uri(self, uri):
        pass
    
    def uri_is_monitorable(self, uri):
        return False

    def uri_is_monitored(self, uri):
        return False

class DataAccessFile:

    def __init__(self, plugin, file_handle):
        self._dataaccess_plugin = plugin
        self._file_handle = file_handle

    def close(self):
        self._dataaccess_plugin.close(self)

    def get_file_handle(self):
        return self._file_handle

    def read(self, size=-1):
        return self._dataaccess_plugin.read(self, size)

    def seek(self, offset, whence=0):
        return self._dataaccess_plugin.seek(self, offset, whence)

    def get_gst_source(self):
        return self._dataaccess_plugin.get_gst_source()

class QueuePlugin(plugin.Plugin):
    " "

    data_access = plugin.ExtensionPoint(interfaces.IDataAccess)

    def get_data_plugin(self, location):
        # extract the type or uri (file:// .. smb:// ....)
        uri = media_uri.Uri(location=location)

        # look for the plugin supporting this type of uri
        data_access_plugin = self.data_access.get("uri:%s" % uri.scheme)

        return data_access_plugin


class CustomPlugin(plugin.Plugin):
    " custom interface Plugin Class "

class ScreenLessPlugin(plugin.Plugin):
    " Screen-less plugin "

class ServerPlugin(plugin.Plugin):
    """
    Implements a server plugin
    """
    __implements__ = IServerProvider
    __platforms__ = ['Linux', 'Windows']

    def get_server_port(self):
        return 80

    def get_host_address(self):
        return get_host_address()

class UserInputPlugin(plugin.Plugin):

    __implements__ = IPlugin

    # List of type of sources for events
    # this permits to identify the kind of hardware provided by the plugin
    source_types = {'local': 0,
                    'remote_control': 1,
                    'other': 2}

    def __init__(self, source_type):
        plugin.Plugin.__init__(self)
        self._source_type = UserInputPlugin.source_types[source_type]

    def create_pgm_render_event(self, event_type):
        event = render.get_renderer().PgmRenderEvent(event_type)
        event.source = self._source_type
        return event

