
# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.extern.configobj import ConfigObj, Section
from elisa.core.log import Logger, LOG_FILE
from elisa.core import common
import os
import textwrap

CONFIG_DIR = os.path.join(os.path.expanduser('~'), ".elisa")
CONFIG_FILE = os.path.join(os.path.curdir, "elisa.conf")

DEFAULT_CONFIG = """\
[general]
# the Pigment renderer to use. Only 'opengl1' supported for now
renderer = 'opengl1'

# window resolution
# resolution = [800, 600]

# should elisa switch to fullscreen on startup? 0: no, 1: yes
start_fullscreen = 0

# the main menu plugins to use
plugins = ['pictures', 'music', 'movies', 'dvd', 'services']

# the data access plugins, each plugin handles a set of URI schemes
# possible values: 'vfs', 'upnp', 'local_fs', 'meta_fs', 'flickr','ipod','daap_fs'
plugins.data = ['vfs', 'upnp', 'local_fs', 'meta_fs','daap_fs','ipod']

# the plugins to load under the "services" menu
# possible values: 'weather', 'config'
plugins.services = ['weather']

# miscellaneous plugins
# possible values: 'audioscrobbler','lastfm','lirc', 'web_radio', 'hal','http_server','upnp_server'
plugins.misc = ['audioscrobbler','lirc', 'web_radio', 'hal', 'lastfm']

# the player queue to use
# possible values: 'default_queue', 'smart_queue'
active_queue = 'default_queue'

# the skin to load. set to empty string '' to not load any skin
skin = 'default_skin'

# the keyboard shortcut to quit elisa
close_key = 'q'

# Logging levels:
#  INFO
#  DEBUG
#  DEBUG_DETAILLED
#  DEBUG_VERBOSE
log_level = 'INFO'

# enable that to make elisa display its log on the console/terminal
console_log_output = 1

enable_profiling = 0

[media_manager]
# enable or disable the media manager
enable_cache = 0

# the database filename. the file will be stored in the same directory as
# the config file
db_name = 'elisa.db'

# TODO: move this to [general] section
cache_directory = 'cache'

# source locations to not monitor. Typically readonly partitions for instance
unmonitored_locations = []

# the following 4 options are used to schedule database updates of the media
# sources declared by each plugin having a "locations" option. Scheduling
# starts from elisa startup time

# source locations to update every 5 minutes. Use with care :)
fivemin_location_updates = []

# source locations to update every hour
hourly_location_updates = []

# source locations to update daily
daily_location_updates = []

# source locations to update every week 
weekly_location_updates = []

[mail]
# email error reporting. Enable if you want elisa to send errors by
# mail on failure
enable = 0
mail_from = 'elisa@localhost'
mail_to = ['phil@localhost']
mail_subject = 'Elisa error'
mail_server = 'localhost'"""

class Config:
    """ Elisa's configuration system

    Each configuration is stored in a text file. The configuration is
    structured in sections. There's at least one section called
    'general'. Each plugin has its own section (ex: 'plugin.foo').


    """

    def __init__(self, config_file=None):

        if not config_file:
            if not os.path.exists(CONFIG_DIR):
                os.makedirs(CONFIG_DIR)
            elif not os.path.isdir(CONFIG_DIR):
                os.unlink(CONFIG_DIR)
                os.makedirs(CONFIG_DIR)
            config_dir = CONFIG_DIR
            config_file = os.path.join(config_dir, CONFIG_FILE)
        else:
            config_dir = os.path.dirname(config_file)
            if not config_dir:
                config_dir = os.getcwd()
                
        if not os.path.exists(config_file):
            self._create_config(config_file)
            
        self._logger = Logger(os.path.join(config_dir, LOG_FILE))
        self._filename = config_file
        self._config_dir = config_dir
        
        try:
            self._config = ConfigObj(config_file, unrepr=True)
        except Exception,ex:
            errors = [ error.msg for error in ex.errors ]
            errors = ';'.join(errors)
            self._logger.info("Config format error in %s: %s" % (config_file,
                                                                 errors))
            raise

    def get_config_dir(self):
        return self._config_dir

    def get_filename(self):
        return self._filename

    def _create_config(self, config_file):
        f = open(config_file,'w')
        f.write(DEFAULT_CONFIG)
        f.close()
        common.get_application().first_run = True
        
    def get_option(self, key, section='general', default=None):
        """ Fetch the option value stored in the given section, at the
        given key. Return a default value if the key is not found.
        """
        return self.get_section(section).get(key, default)

    def set_option(self, key, value, section='general'):
        """ Store an option value under key id at the given section.
        """
        self._config.setdefault(section,{})[key] = value
        #self._config[section][key] = value

    def del_option(self, key, section_name='general'):
        """ Remove the option identified by key under the specified
        section.
        """
        section = self.get_section(section_name)
        if section and section.has_key(key):
            del section[key]
        self.set_section(section_name, section)

    def write(self):
        """
        save the config in a text file (handled by ConfigObj)

        """
        self._logger.info('saving config...')
        self._config.write()

    def get_section(self, section_name, default=None):
        """
        @return the ConfigObj section identified by section_name
        """
        if default is None:
            default = {}
        return self._config.get(section_name, default)

    def set_section(self, section_name, section={}, doc={}):
        """
        Store section_data in a new section identified by section_name
        in the config
        """
        if not isinstance(section, Section):
            section = Section(self._config,
                              self._config.depth+1,
                              self._config.main,
                              indict=section,
                              name=section_name)
            
            for key in section.keys():
                doc.setdefault(key, '')
            section.comments = dict([(k, ['# %s' % line
                                          for line in textwrap.wrap(v, 77)])
                                     for k,v in doc.iteritems()])
            
        self._config[section_name] = section

    def del_section(self, section_name):
        """ Remove the section identified by section_name
        """
        if self._config.has_key(section_name):
            del self._config[section_name]

    def as_dict(self):
        r = {}
        for name, section in self._config.iteritems():
            r.update(section.dict())
        return r

if __name__ == "__main__":
    try:
        config_file = "test.conf"
        
        data = open("test.conf").read().splitlines();
        _config = ConfigObj(data, unrepr=True)
    except Exception,ex:
        errors = [ error.msg for error in ex.errors ]
        errors = ';'.join(errors)
        print ("Config format error in %s: %s" % (config_file,
                                                             errors))
        raise
