# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.


import logging
import logging.handlers
import sys, os
import copy

logger = None

class CustomFile:

    def __init__(self, f):
        self.f = f
        self.enable()

    def enable(self):
        self.do_write = True

    def disable(self):
        self.do_write = False

    def write(self, data):
        if self.do_write:
            self.f.write(data)

    def __getattr__(self, attr):
        return getattr(self.f, attr)

class debug:
    """
    The debug decorator. It basically logs the name and arguments of
    the function/method it's decorating.

    Use it like:

       @log.debug("some message")
       def foo(self, bar, bad=0):
           # do something ...
           pass

    This would result in a debug message like:

        foo(<some instance>, <some bar>,) bad=0 some message
        
    """

    def __init__(self, message):
        self.message = message
    
    def __call__(self, f):
        def newf(*args, **kwds):
            kw_values = ', '.join([ '%s=%s' % (a,v) for (a,v) in
                                    kwds.iteritems() ])
            Logger().debug("%s%s %s %s" % (f.func_name, args, kw_values,
                                           self.message))
            f_result = f(*args, **kwds)
            # we could even log the result of the function/method
            return f_result
        newf.__doc__ = f.__doc__
        return newf

LOG_FILE='elisa.log'

class _Logger:
    """
    The Elisa logger, which basically log messages to a file named
    core.log. Use it like this:

      from log import Logger

      logger.debug("hey this is a log message")
      logger.info("need some info ?")

    Logging levels:

    - INFO            : 20
    - DEBUG           : 10
    - DEBUG_DETAILLED : 5
    - DEBUG_VERBOSE   : 1

    FIXME : Line insered in double in logfile
    """

    levels = {'DEBUG_DETAILLED': 5,
              'DEBUG_VERBOSE': 1
              }

    stdout = CustomFile(sys.stdout)


    def __init__(self, log_file=LOG_FILE):
        common = '%(asctime)s.%(msecs)-4d %(levelname)-8s '
        default_format = common + '%(message)s'
        datefmt = '%d/%m/%Y %H:%M:%S'

        formatter = logging.Formatter(default_format,
                                      datefmt)
        
        handler = logging.handlers.RotatingFileHandler(log_file,
                                                       maxBytes=10000000,
                                                       backupCount=5)
        handler.setFormatter(formatter)

        for name, lvl in self.levels.iteritems():
            logging.addLevelName(lvl, name)

        self._log = logging.getLogger('Elisa')
        self._log.addHandler(handler)

        self.set_level('DEBUG_DETAILLED')

        logging.basicConfig(stream=self.stdout)

        root = logging.getLogger()
        handler = root.handlers[0]
        handler.setFormatter(formatter)

        self.enable_console_output()

    def enable_console_output(self):
        """ enable log output to stdout """
        self.stdout.enable()
        self.stdout.flush()

    def disable_console_output(self):
        """ disable log output to stdout """
        self.stdout.disable()
        self.stdout.flush()

    def set_level(self, name):
        """ Set the log level. name can be one of:

        - 'INFO'
        - 'DEBUG'
        - 'DEBUG_DETAILLED'
        - 'DEBUG_VERBOSE'

        """
        level = logging.getLevelName(name)
        self._log.setLevel(level)
        self._level_name = name

    def get_level(self):
        """ Fetch current log level string identifier
        """
        return self._level_name

    def info(self, info):
        """ Log some basic info (in >= INFO log level)
        """
        self._log.log(logging.INFO, info)

    def debug(self, msg, obj=None):
        """ Log functional data (in >= DEBUG log level)
        """
        self._log.log(logging.DEBUG, msg)

    def debug_detailled(self, msg, obj=None):
        """ Log verbose functional data (in >= DEBUG_DETAILLED log level)
        """
        self._log.log(self.levels['DEBUG_DETAILLED'], msg)

    def debug_verbose(self, msg, obj=None):
        """ Detailled logging (in >= DEBUG_VERBOSE log level)
        """
        self._log.log(self.levels['DEBUG_VERBOSE'], msg)

def Logger(log_file=LOG_FILE):
    global logger
    if not logger:
        logger = _Logger(log_file)
    return logger

if __name__ == '__main__':
    l = Logger()

    l.info("Hello World")
    l.debug("Argh")
