# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.core import plugin
from elisa.core import interfaces
from elisa.core import log
from elisa.core import common
from elisa.core import media_uri
from elisa.core.signals.data import data_updated
from elisa.core.utils import misc, priority_queue

import sys, os
import mimetypes

from elisa.core.utils import tag_browser
    
from elisa.core.utils.misc import handle_failure
from elisa.core.base_plugins import uri_cacheable

import threading, time
from twisted.internet import threads, task, defer, reactor
from twisted.spread import pb

try:
    from elisa.core.media_db import MediaDB, Row
except ImportError, error:
    media_db_error = error
else:
    media_db_error = None

global DB_CONNECTIONS
DB_CONNECTIONS = {}

"""
TODO:

- add signals notifications for the skin
- backport the 2 methods used by elisa_storage Coherence backend from
  the cache manager
"""

class MediaManager(pb.Referenceable):

    data_access = plugin.ExtensionPoint(interfaces.IDataAccess)

    def __init__(self, config):
        self.logger = log.Logger()
        self._config = config
        self._queue = priority_queue.PriorityQueue()

        self._times = {}

        self._update_intervals = dict(fivemin=60*5,
                                      hour = 60*60, day = 60* 60 * 24,
                                      week = 60 * 60 * 24 * 7)
        self._fivemin_locations = self._get_option('fivemin_location_updates',[])
        self._hourly_locations = self._get_option('hourly_location_updates',[])
        self._daily_locations = self._get_option('daily_location_updates',[])
        self._weekly_locations = self._get_option('weekly_location_updates',[])
        
        self._tag_browser = tag_browser.TagBrowser()
        
        self._db_lock = threading.Lock()
        
        self._main_loop_lock = threading.Lock()
        self._main_loop_running = True

        config_dir = config.get_config_dir()
        self.db_name = os.path.join(config_dir,
                                    self._get_option('db_name',
                                                     default='elisa.db'))

        self.cache_dir = os.path.join(config_dir,
                                      self._get_option('cache_directory',
                                                       default='cache'))
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

        self._enable_cache = self._get_option('enable_cache', default=1)

        locations = self._get_option('unmonitored_locations', default=[])
        self._unmonitored_locations = [str(media_uri.Uri(l)) for l in locations]
        
        if self._enable_cache:
            if media_db_error:
                self.logger.info("Media management is disabled. "
                                 "Please install sqlite3 and PySQLite2")
                self.logger.info("Error was: %s" % media_db_error)

    def enabled(self):
        return self._enable_cache and not media_db_error

    def start(self):
        if self.enabled():            
            main_thread = threading.Thread(target=self._process_queue,
                                           name="media manager main loop")
            main_thread.start()
            self._schedule_updates()
            
    def stop(self):
        if self.enabled():
            self.logger.info("Stopping Media manager")
            self._main_loop_lock.acquire()
            self._main_loop_running = False
            self._main_loop_lock.release()

            # save the config
            self._set_option('fivemin_location_updates', self._fivemin_locations)
            self._set_option('hourly_location_updates', self._hourly_locations)
            self._set_option('daily_location_updates',self._daily_locations)
            self._set_option('weekly_location_updates',self._weekly_locations)
            self._set_option('unmonitored_locations',self._unmonitored_locations)
            
    def _is_running(self):
        self._main_loop_lock.acquire()
        running = self._main_loop_running
        self._main_loop_lock.release()
        return running

    def _get_option(self, name, default=None):
        # Keep backward compatibility. Cache manager options used to be stored
        # in [general] section
        section = self._config.get_section("media_manager")
        if section:
            section = "media_manager"
        else:
            section = "general"
        opt = self._config.get_option(name, default=default, section=section)
        return opt

    def _set_option(self, name, value):
        section = self._config.get_section("media_manager")
        if section:
            section = "media_manager"
        else:
            section = "general"
        if type(value) == unicode:
            value = value.encode('utf-8')
        self._config.set_option(name, value, section=section)
        
    def get_data_plugin(self, uri):
        # extract the type or uri (file:// .. smb:// ....)
        if not isinstance(uri, media_uri.Uri):
            uri = media_uri.Uri(misc.fs_decode(uri))
        data_access_plugin = self.data_access.get("uri:%s" % uri.scheme)
        return data_access_plugin
                    
    def get_db_connection(self):
        conn = None
        global DB_CONNECTIONS
        if self.enabled():
            try:
                self._db_lock.acquire()
                current_thread = threading.currentThread()
                thread_id = id(current_thread)
                conn = DB_CONNECTIONS.get(thread_id)
                if not conn:
                    conn = MediaDB(self.db_name)
                    DB_CONNECTIONS[thread_id] = conn
            finally:
                self._db_lock.release()
        return conn

    def get_tag_browser(self):
        """
        TODO: refactor this one if we decide to allow // media files parsing
        """
        return self._tag_browser

    def _enqueue(self, item, priority):
        self._queue.put((item, priority))

    def _dequeue(self):
        item = None
        if not self._queue.empty():
            item = self._queue.get()
        return item

    def _process_queue(self):
        while 1:
            if not self._is_running():
                break

            item = self._dequeue()
            if item:
                source, for_plugin, uri, short_name,parent = item
                try:
                    self._process_source(source, for_plugin, uri, short_name,
                                         parent)
                except Exception, exc:
                    self.logger.info("Source parsing failed: %s" % exc)
            else:
                # keeps CPU available when queue empty
                time.sleep(0.05)

    def _schedule_updates(self):        
        # schedule some events
        for label, interval in self._update_intervals.iteritems():
            reactor.callLater(interval, self._scheduled_update, label)
                    
    def _scheduled_update(self, interval_label):
        interval = self._update_intervals.get(interval_label)
        reactor.callLater(interval, self._scheduled_update, interval_label)
        interval_locations = {'fivemin': self._fivemin_locations,
                              'hour': self._hourly_locations,
                              'day': self._daily_locations,
                              'week': self._weekly_locations}
        sources = interval_locations.get(interval_label)
        if sources:
            msg = "Launching scheduled update of the %s: %s" % (interval_label,
                                                                sources)
            self.logger.info(msg)
            for source_uri in sources:
                self.update_source(media_uri.Uri(source_uri))

    def _is_scheduled(self, source_uri):
        all_updates = self._fivemin_locations + self._hourly_locations + \
                      self._daily_locations + self._weekly_locations
        all_updates = set([str(media_uri.Uri(location=location))
                           for location in all_updates])
        scheduled = source_uri in all_updates
        return scheduled
    
    #from elisa.core.utils import profiling
    #@profiling.profileit(40)
    def _process_source(self, source, for_plugin=None, uri=None,
                        short_name=None, parent=None):
        # source can be either
        # - a db Row instance
        # - an URI string
        # uri is the current item being processed in the source
        #print 'Processing %r for %s at %r' % (source, for_plugin, uri)
        
        available = True
        conn = self.get_db_connection()
        
        if isinstance(source, basestring):
            # add the source to db
            source_name = media_uri.Uri(source).get_label()
            db_source = self.get_source_for_uri(source)
            if db_source:
                source = db_source
                available = source.available
                if not available:
                    conn.show_source(source)
            else:
                source = conn.add_source(source, source_name)
        elif not isinstance(source, Row):
            source = self.get_source_for_uri(source)
            available = source.available

        if not self.monitored_source(source.uri) or not available:
            self.logger.info("Not monitored: %s or not available: %s" % (source.uri,
                             available))
            return

        if source.uri not in self._times:
            self._times[source.uri] = time.time()
            
        if not uri:
            uri = source.uri
            parent = None
            
        data_access = self.get_data_plugin(uri)
        if not data_access:
            self.logger.info("No DataAccess plugin found for %s" % uri)
            return
        
        next_location, next_short_name = data_access.next_location(uri,
                                                                   files_only=False)
        if data_access.is_dir(uri):
            self._process_media_directory(source, uri, short_name,
                                          parent)
            if next_location:
                self._enqueue((source, for_plugin, next_location,
                               next_short_name,
                               uri), 0)
                processed = True
            else:
                processed = False
        else:
            processed = self._process_media_for_plugin(source, for_plugin, uri,
                                                       short_name, parent)
            if next_location:
                parent = data_access.get_parent(next_location)
                self._enqueue((source, for_plugin, next_location,
                               next_short_name, parent), 0)

        if not next_location:
            self._source_updated(source)

    def _source_updated(self, source):
        conn = self.get_db_connection()

        # displays profiling information
        # FIXME: this does not always reflect reality
        now = time.time()
        delta = now - self._times[source.uri]
        count = conn.get_files_count_for_source(source)
        if count:
            speed = "(%s s/file)"  % (delta / count,)
        else:
            speed = ""
        msg = 'Parse of %s took %s seconds %s' % (source.uri, delta, speed)
        self.logger.info(msg)

        del self._times[source.uri]
        conn.updated_source(source)
        conn.hide_un_updated_medias_for_source(source)

        # schedule source daily cyclic update
        if source.uri not in self._daily_locations:
            self._daily_locations.append(source.uri)
        
    def _process_media_directory(self, source, uri, short_name, parent_uri):
        file_type = 'directory'
        conn = self.get_db_connection()
        media_node_id = 0
        if not parent_uri:
            media_node_id = -1
            source_id = source.id
        else:
            # XXX: hmm this should be avoidable
            if parent_uri.startswith('file:') and not parent_uri.endswith('/'):
                parent_uri = '%s/' % parent_uri
                
            media_parent = self.get_media_for_uri(parent_uri)
            if not media_parent:
                parent_media_uri = media_uri.Uri(parent_uri)
                # XXX
                parent_parent = parent_media_uri.get_parent()
                parent_name = parent_media_uri.get_label()
                media_parent = self.get_media_for_uri(parent_parent)
                if media_parent:
                    args = (parent_uri, parent_name)
                    kw = {'format':'', 'typ': file_type,
                          'source_id': source.id,
                          'media_node_id': media_parent.id}
                    inserted, updated, media_parent = conn.add_media_node(*args,
                                                                          **kw)
                    media_node_id = media_parent.id
                    source_id = source.id
            else:
                media_node_id = media_parent.id
                source_id = source.id
            
        #print 'Processing dir: %s (source=%r, parent=%r)' % (uri,source_id,
        #                                                     media_node_id)

        if media_node_id:
            args = (uri, short_name)
            kw = {'format':'', 'typ': file_type,
                  'source_id': source_id, 'media_node_id': media_node_id}
            inserted, updated, media_node = conn.add_media_node(*args, **kw)
            
    def _process_media_for_plugin(self, source, for_plugin, uri,
                                  short_name, parent_uri):
        data_access = self.get_data_plugin(uri)
        media_type = data_access.get_file_type(uri)
        if not for_plugin or media_type in for_plugin.provides:
            processed = self._process_media(source, uri, short_name, parent_uri)
        else:

            if for_plugin.get_name() == "music":
                cover_file = for_plugin.get_config().get('cover_file')
            else:
                cover_file = None
                
            if media_type == "image" and cover_file and parent_uri:
                parent = self.get_media_for_uri(parent_uri)
                if parent and short_name == cover_file:
                    conn = self.get_db_connection()
                    #media = self.get_media_for_uri(uri)
                    conn.update_metadata(None, parent, cover_uri=uri)
                processed = True
            else:
                processed = True

        return processed
 
    def _process_media(self, source, uri, short_name, parent):
        # FIXME: if media does not exist anymore, quit
        conn = self.get_db_connection()
        data_access = self.get_data_plugin(uri)

        media_type = data_access.get_file_type(uri)
        inserted, updated = False, False
        
        if parent:
            parent = self.get_media_for_uri(parent)
            if parent:
                media_node_id = parent.id
            else:
                media_node_id = -1
        else:
            media_node_id = -1

        if uri.startswith('upnp'):
            uri = "http://%s" % uri[7:]

            
        media_type = data_access.get_file_type(uri)

        file_type = 'file'
        args = (uri, short_name)
        kw = {'format': media_type, 'typ': file_type,
              'source_id': source.id, 'media_node_id': media_node_id}
        metadata = self._parse_media(source, media_type, uri, short_name)
        kw.update({'metadata':metadata})
        inserted, updated, media_node = conn.add_media_node(*args, **kw)
        if not inserted or not updated:
            conn.update(media_node, updated=1)
        artist = metadata.get('artist')
        if artist and parent and parent.artist != artist:
            album = metadata.get('album')
            conn.update_metadata(media_node, parent, artist=artist, album=album)

        return inserted or updated

    def _parse_media(self, source, media_type, uri, short_name):
        # parse tags (audio, picture)
        tags = {}

        if media_type == 'audio':
            tag_browser = self.get_tag_browser()
            uri = misc.fs_encode(uri)
            b = self.get_tag_browser()
            
            #self.logger.info("Parsing tags on %s" % uri)
            try:
                tags = b.get_tags(uri)
            except Exception, ex:
                print 'Tag parsing failed', ex
                tags = {}
            else:
                artist = tags.get('artist')
                album = tags.get('album')
                song = tags.get('title')
                track = tags.get('track-number')

                # TODO: move this to tag_browser.get_tags
                if type(artist) == list:
                    artist = artist[0]
                if type(song) == list:
                    song = song[0]
                if type(album) == list:
                    album = album[0]
                if not song:
                    song = short_name.replace('_',' ')
                    
                tags = {'artist': artist, 'album': album, 'song': song,
                        'track': track}
        return tags

    def _monitor_event_handler(self, event, uri):
        #self.logger.info("Event %s occured on %s" % (event, uri))
        if event == "ADDED" or event == "CHANGED":
            #self._enqueue((uri, None, None, None, None), 0)
            self.add_source(uri)
        elif event == "REMOVED":
            # FIXME: hrmm this is wrong if we're dealing with non-source files
            #self.remove_source(uri)
            pass
   

    ## Public API

    def add_source(self, source_uri, for_plugin=None, priority=0):
        """
        Add a new source in DB. If it's already there, just mark it as
        available again (and start monitoring it).

        :param source_uri: The location of the source
        :type uri: string
        :param for_plugin: The plugin for which to add the source
        :type for_plugin: elisa.core.plugin.Plugin
        :param priority: source update priority. 0 = higher
        :type priority: int
        """
        self.logger.info("Adding source %s" % source_uri)
        self._enqueue((source_uri, for_plugin, None, None, None), priority)

        if not isinstance(source_uri, media_uri.Uri):
            source_uri = media_uri.Uri(misc.fs_decode(source_uri))

        # TODO: re-enable this when we have acceptable monitoring support
##         self.monitor_source(source_uri)

    def remove_source(self, source_uri, for_plugin):
        """
        Mark a source as unavailable in the db and stop monitoring it
        
        :param source_uri: The location of the source
        :type uri: string
        :param for_plugin: The plugin for which to add the source
        :type for_plugin: elisa.core.plugin.Plugin
        """
        source = self.get_source_for_uri(source_uri)
        if not isinstance(source_uri, media_uri.Uri):
            source_uri = media_uri.Uri(misc.fs_decode(source_uri))
            
        if source:
            self.unmonitor_source(source_uri)

            # mark the source as unavailable in the db
            conn = self.get_db_connection()
            conn.hide_source(source)

    def update_source(self, source_uri, for_plugin=None, probe=False):

        if not for_plugin:
            plugin_manager = common.get_plugin_manager()
            for plugin in plugin_manager.get_main_plugins():
                if not hasattr(plugin, 'get_locations'):
                    continue
                source_uri = str(source_uri)
                locations = [str(media_uri.Uri(l)) for l in plugin.get_locations()]
                if source_uri in locations:
                    for_plugin = plugin
                    break
        if for_plugin:
            do_update = True
            if probe:
                do_update = self.monitored_source(source_uri) and \
                            not self._is_scheduled(source_uri)
            if do_update:
                conn = self.get_db_connection()
                source = self.get_source_for_uri(source_uri)
                conn.prepare_source_for_update(source)
                self._enqueue((source_uri, for_plugin, None, None, None), 0)
                
    def monitored_source(self, source_uri):
        """
        Check wether a source located at `source_uri` is monitored or not.

        :param source_uri: The location of the source
        :type uri: string
        :rtype: bool
        """
        monitored = source_uri not in self._unmonitored_locations
        return monitored

    def monitor_source(self, source_uri):            
        """
        Start monitoring the source located at
        `source_uri`. Monitoring is handled by the underlying
        DataAccess plugin.

        :param source_uri: The location of the source
        :type uri: elisa.core.media_uri.Uri
        """            
        data_access_plugin = self.get_data_plugin(source_uri)
        
        # start monitoring this source if monitorable
        if self.monitored_source(source_uri) and data_access_plugin:
            self.logger.info("Starting to monitor %s" % source_uri)
            data_access_plugin.monitor_uri(source_uri,
                                           self._monitor_event_handler)

    def unmonitor_source(self, source_uri):
        """
        Stop monitoring the source located at `source_uri`.
        
        :param source_uri: The location of the source
        :type uri: elisa.core.media_uri.Uri
        """
        data_access_plugin = self.get_data_plugin(source_uri)
        if self.monitored_source(source_uri) and data_access_plugin:
            data_access_plugin.unmonitor_uri(source_uri)
            
    def get_children(self, media, uri, for_plugin):
        """
        Used by TreePlugins to build their menus according to db contents
        """
        available = True
        conn = self.get_db_connection()

        if type(for_plugin) == basestring:
            manager = common.get_plugin_manager()
            for_plugin = manager.get_plugin_with_name(for_plugin)

        media_node = conn.get_media_for_uri(uri)
        if not media_node:
            media_node = media
            available = not media.deleted
            
        item_filter = for_plugin.item_filter
        children = {}
        if available:
            nodes = conn.get_nodes(media_node)
            for node in nodes:
                index = nodes.index(node)
                if node.deleted:
                    continue
                node_uri = misc.fs_encode(node.uri)
                short_name = node.short_name
                if conn.node_is_file(node):
                    if node.song:
                        short_name = node.song
                    if node.format in for_plugin.provides:
                        children[node_uri] = (node.typ,
                                              node.format, short_name, index)
                        
                elif conn.node_is_dir(node):
                    if not conn.node_is_empty(node,
                                              for_plugin):
                        children[node_uri] = ('directory', '', short_name,
                                              index)
        return children

    def has_children(self, uri, plugin_name):
        """
        Check whether the media located at `uri` has some leaves or not.

        :param uri: The media uri to scan
        :type uri: string
        :param plugin_name: The target plugin, we'll filter medias using provides list of the plugin
        :type plugin_name: string
        :rtype: bool
        """
        has_children = False
        
        conn = self.get_db_connection()        
        media_node = conn.get_media_for_uri(uri)
        
        plugin_manager = common.get_plugin_manager()
        for_plugin = plugin_manager.get_plugin_with_name(plugin_name)
        
        nodes = conn.get_nodes(media_node)
        for node in nodes:
            if node.deleted:
                continue
            if conn.node_is_file(node):
                if node.format in for_plugin.provides:
                    has_children = True
            elif conn.node_is_dir(node):
                if not conn.node_is_empty(node, for_plugin):
                    has_children = True
            if has_children:
                break
        return has_children

    def get_source_for_uri(self, uri):
        """
        Fetch the source located at given `uri`

        :param uri: The uri of the source to look for
        :type uri: string
        :rtype: elisa.core.media_db.Row
        """
        conn = self.get_db_connection()
        return conn.get_source_for_uri(uri)

    def get_media_for_uri(self, uri):
        """
        Fetch the media located at given `uri`

        :param uri: The uri of the media to look for
        :type uri: string
        :rtype: elisa.core.media_db.Row
        """
        conn = self.get_db_connection()
        return conn.get_media_for_uri(uri)

    get_media_node_for_uri = get_media_for_uri

    def get_media_root_id(self, media_type):
        conn = self.get_db_connection()
        #sources = conn.get_sources()
        #for source in sources:
        medias = conn.get_medias()
        for media in medias:
            if media.typ == media_type:
                return media.source_id
        return 0

    remote_get_media_root_id = get_media_root_id

    def get_media_node_with_id(self, media_node_id, conn=None):
        """
        Returns a dict with following keys:
        id = id in the media db
        parent_id = parent_id in the media db
        name = title, album name or basename
        mimetype = 'directory' or real mimetype
        size = in bytes (??)
        children = list of objects for which this item is the parent
        location = filesystem path if item is a file
        """
        node_dict = {}
        if not conn:
            conn = self.get_db_connection()

        if type(media_node_id) == int:
            # handle special case for Coherence
            if media_node_id == 0: 
                medias = conn.get_medias()
                child_list = []
                for media in medias:
                    node_dict = {'id': media.id,
                                 'parent_id': 0,
                                 'name': media.short_name,
                                 'mimetype': 'directory',
                                 'size': 1,
                                 'children': [],
                                 'location': {'internal': media.uri,
                                              'external': media.uri}
                                 }
                    child_list.append(node_dict)
                full_dict = {'id': 0,
                             'parent_id': 0,
                             'name': 'root',
                             'mimetype': 'directory',
                             'size': 1,
                             'children': child_list,
                             'location': {}
                             }
                #print "FULL DICT: ", full_dict
                return full_dict
            else:
                node = conn.get_media_with_id(media_node_id)
        else:
            node = media_node_id

        if node is not None:
            node_dict = self._node_to_dict(node, conn)
            node_dict['children'] = [self._node_to_dict(n, conn)
                                     for n in conn.get_nodes(node)]
            
        return node_dict
    
    remote_get_media_node_with_id = get_media_node_with_id

    def _node_to_dict(self, node, conn):
        parent = conn.get_parent(node)
        if parent:
            parent_id = parent.id
        else:
            parent_id = 0
        if node.typ == 'directory':
            mimetype = 'directory'
        else:
            # try to guess the uri mime-type
            mimetype, sub_type = mimetypes.guess_type(node.uri)
            if not mimetype:
                # fallback to incomplete mime-type based on node.format
                mimetype = '%s/' % node.format

        # get the plugin that will handle http://
        plugins = common.get_plugin_manager().get_singleton_plugins()
        location = ''
        for i in plugins:
            if ('http' in i.provides):
                location = 'http://%s:%s' % (i.get_host_address(),
                                             i.get_server_port())
                break
        if location == '':
            location = node.uri

        external_location = location + '/data/upnp/' + str(node.id)
        internal_location = node.uri
        node_dict = {'id': node.id,
                     'parent_id': parent_id,
                     'name': node.short_name,
                     'mimetype': mimetype,
                     'size': 0,
                     'children': [],
                     'location': {'internal': internal_location,
                                  'external': external_location}
                     }
        
        return node_dict

    ## /Public API
