# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.core import log, common
from elisa.core.plugin import Plugin
from elisa.core.component import InterfaceOmission, UnsupportedPlatform
from elisa.core import log
from sets import Set
import pkg_resources
import os, sys, locale
from twisted.spread import pb

from pkg_resources import EntryPoint

class PluginManager(pb.Referenceable):

    def __init__(self, config):
        self._main_plugins = []
        self._singleton_plugins = []
        self._plugins = {}
        self._config = config
        
        logger = log.Logger()
        config = self.get_config()


        if hasattr(sys, 'frozen') and sys.frozen:
            import elisa.core.plugins.config
            import elisa.core.plugins.daap_fs
            import elisa.core.plugins.dvd
            import elisa.core.plugins.flickr
            import elisa.core.plugins.hal
            import elisa.core.plugins.http_server
            import elisa.core.plugins.local_fs
            import elisa.core.plugins.meta_fs
            import elisa.core.plugins.movies
            import elisa.core.plugins.music
            import elisa.core.plugins.pictures
            import elisa.core.plugins.services
            import elisa.core.plugins.weather
            import elisa.core.plugins.web_radio
            import elisa.core.plugins.skins.default.skin

            self._entry_points = None
            try:
                f = open(os.path.join('elisa.egg-info','entry_points.txt'))
                self._entry_points = EntryPoint.parse_map(f.readlines())
                #print "m:", self._entry_points
            except Exception, ex:
                print "Error loading entrypoints", ex
                raise

        logger.set_level(config.get_option('log_level'))
        log_output = config.get_option('console_log_output')
        if log_output is None:
            logger.enable_console_output()
            logger.info("console_log_output option improperly set. "
                        "Please set it either to 1 or 0 in the config "
                        "file %s" % config.get_filename())
            
        console_log_output = int(log_output)
        if console_log_output:
            logger.enable_console_output()
        else:
            logger.disable_console_output()

        self.logger = logger
        self.logger.info("Using config file : %s" % config.get_filename())

        self.init_local_plugins_dir()

    def init_local_plugins_dir(self):
        working_set = pkg_resources.working_set

        # set up our custom environment where plugins are located
        plugins_dir = os.path.expanduser('~/.elisa/plugins')
        environment = pkg_resources.Environment([plugins_dir,])

        # cope with old versions of setuptools 
        if hasattr(working_set, 'find_plugins'):
            distributions, errors = working_set.find_plugins(environment)
            map(working_set.add, distributions)
            if errors:
                self.logger.info("Couldn't load: %s" % errors)
        else:
            self.logger.info("Please upgrade setuptools if you want to be able to use additional plugins from ~/.elisa/plugins")

    def get_config(self):
        return self._config

    def get_plugins(self):
        return self._plugins

    def get_plugin_with_name(self, name):
        found = None

        # first dig main plugins
        for plugin in self.get_main_plugins():
            if plugin.get_name() == name:
                found = plugin
                break
        
        if not found:
            # try singleton plugins ...
            for plugin in self.get_singleton_plugins():
                if plugin.get_name() == name:
                    found = plugin
                    break
                
            if not found:
                # try all un-instantiated plugins
                for entry_point_name, plugins in self._plugins.iteritems():
                    for plugin in plugins:
                        if plugin.name == name:
                            if plugin.enabled:
                                found = plugin()
                            else:
                                found = plugin
                            break
        return found

    remote_get_plugin_with_name = get_plugin_with_name

    def get_plugins_for_entry_point(self, entry_point_name):
        plugins = self._plugins.get(entry_point_name,[])
        return plugins

    def get_main_plugins(self):
        return self._main_plugins

    def set_main_plugins(self, plugins):
        self._main_plugins = plugins

    def get_singleton_plugins(self):
        return self._singleton_plugins

    def add_singleton_plugin(self, plugin):
        self.logger.debug("Adding singleton: %r" % plugin)
        self._singleton_plugins.append(plugin)

    def del_singleton_plugin(self, plugin):
        if plugin in self._singleton_plugins:
            self.logger.debug("Removing singleton: %r" % plugin)
            self._singleton_plugins.remove(plugin)

    def unload_plugins(self):
        for plugin in self.get_singleton_plugins() + self.get_main_plugins():
            plugin.unload()
            
        for entry_point_name, plugins in self._plugins.iteritems():
            for plugin in plugins:
                self.unregister_plugin(plugin)
        
    def load_plugins(self, load_main_plugins=True):
            
        options = self.get_config().get_section('general').keys()
        entry_points = [ opt for opt in options
                         if opt.startswith('plugins') ]

        # load all the plugins
        for entry_point in entry_points:
            self.load_plugins_for_entry_point('elisa.%s' % entry_point)

        # set the main plugins list
        plugin_names = self.get_config().get_option('plugins', default=[])

        if load_main_plugins:
            main_plugins = self.get_plugins().get('elisa.plugins',[])
            main_plugins = [ p() for p in main_plugins if p.enabled ]
            main_plugins.sort(key=lambda p: plugin_names.index(p.get_name()))
        else:
            main_plugins = []
        self.set_main_plugins(main_plugins)

        # post-init main plugins
        for plugin in self.get_main_plugins()[:]:
            plugin.post_init()

        # post-init singleton plugins
        singleton_plugins = self.get_singleton_plugins()[:]
        for plugin in singleton_plugins:
            plugin.post_init()            

    def load_plugins_for_entry_point(self, entry_point_name='elisa.plugins',
                                     plugin_names=None, check_in_config=True):
        """ Add the plugins given their entry point name.
        """
        
        # drop the first part of the entry_point name
        # elisa.plugins => plugins
        conf_plugins_name = '.'.join(entry_point_name.split('.')[1:])

        if not plugin_names:
            plugin_names = self.get_config().get_option(conf_plugins_name,
                                                        default=[])

        # TODO: refactor this duplicated code
        if hasattr(sys, 'frozen') and sys.frozen:
            for name, entrypoint in self._entry_points[entry_point_name].iteritems():
                if check_in_config and name not in plugin_names:
                    continue
                try:
                    plugin_class = entrypoint.load(False)
                except (InterfaceOmission, AssertionError, UnsupportedPlatform), error:
                    self.logger.info(error)
                    continue
                except ImportError, error:
                    self.logger.info("Not found: %s" % entrypoint.module_name)
                    self.logger.info("Error was %s" % error)
                    continue
                    #raise
                assert issubclass(plugin_class, Plugin), \
                       '%r is not a valid Plugin!' % plugin_class

                self.register_plugin(entry_point_name, plugin_class)
                self.logger.info("Loaded the plugin '%s'" % name)
        else:
            for entrypoint in pkg_resources.iter_entry_points(entry_point_name):
                try:
                    plugin_class = entrypoint.load()
                except (locale.Error,InterfaceOmission, AssertionError,
                        UnsupportedPlatform), error:
                    self.logger.info(error)
                    continue
                except ImportError, error:
                    self.logger.info("Not found: %s" % entrypoint.module_name)
                    self.logger.info("Error was %s" % error)
                    continue
                    #raise
                assert issubclass(plugin_class, Plugin), \
                       '%r is not a valid Plugin!' % plugin_class

                if check_in_config and entrypoint.name not in plugin_names:
                    enabled = False
                else:
                    enabled = True

                self.register_plugin(entry_point_name, plugin_class, enabled)
                self.logger.info("Loaded the plugin '%s'" % entrypoint.name)

        if check_in_config:
            loaded = [ ]
            for entry_point, plugins in self.get_plugins().iteritems():
                loaded.extend(plugins)
            loaded = [ p.name for p in loaded ]

            s1 = Set(loaded)
            s2 = Set(plugin_names)
            if not s2.issubset(s1):
                self.logger.info("Plugins not found : %s" % ', '.join(list(s2-s1)))


    def register_plugin(self, entry_point_name, plugin_class, enabled):
        """ Add the given plugin instance to our internal plugins list.
        """
        self._plugins.setdefault(entry_point_name,[]).append(plugin_class)
        plugin_class.enabled = enabled
        # strip "elisa."
        conf_entry_point_name = '.'.join(entry_point_name.split('.')[1:])
        plugin_class.entry_point_name = conf_entry_point_name
        if plugin_class.singleton == True and enabled:
            self.add_singleton_plugin(plugin_class())

    def unregister_plugin(self, plugin_class):
        for entry_point_name, plugins in self._plugins.iteritems():
            if plugin_class in plugins:
                index = plugins.index(plugin_class)
                del plugins[index]
                self._plugins[entry_point_name] = plugins

    def is_registered(self, plugin_class):
        for entry_point_name, plugins in self._plugins.iteritems():
            if plugin_class in plugins:
                return True
        return False
