# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2006, Mickael Sheldon <mike AT mikeasoft DOT com>

from elisa.core.base_plugins import DataAccessPlugin, DataAccessFile
from elisa.core.interfaces import IDataAccess
from elisa.core import media_uri
from elisa.core.utils.decode_string import code_to_string
from elisa.extern.flickrapi import FlickrAPI
from elisa.core import plugin
from elisa.core.utils.misc import get_resource_path
import re

class FlickrPlugin(DataAccessPlugin):

    __implements__ = IDataAccess

    name = "flickr"

    provides = [ "uri:flickr" ]

    config_doc = {'api_key': 'Flickr API key. you can get one at flickr.com',
                  'api_secret': 'Flickr API secret challenge',
                  'pictures_per_page': 'number of pictures to display by menu level'
                  }
    default_config = {"api_key" : "",
                      "api_secret" : "",
                      "pictures_per_page" : 8}

    singleton = True

    pictures = {}

    data_access = plugin.ExtensionPoint(IDataAccess)
    uncacheable_uri_schemes = ['flickr']

    def __init__(self):
        DataAccessPlugin.__init__(self)
        self.root_folder_icon = get_resource_path(self.__module__, 'data/flickr.png')

        self.fapi_key = self.get_config().get('api_key')
        self.fapi_secret = self.get_config().get('api_secret')
        self.per_page = str(self.get_config().get('pictures_per_page', 8))
        if not self.fapi_key or not self.fapi_secret or self.fapi_key == "" \
           or self.fapi_secret == "":
            self.logger.info("You must enter a Flickr API key and shared "\
                             "secret in to elisa.conf (api_key and "\
                             "api_secret) before using the Flickr plugin. "\
                             "These can be obtained from: "\
                             "http://www.flickr.com/services/api/keys")
            self.unload()
            
        self.fapi = FlickrAPI(self.fapi_key, self.fapi_secret)
        
    def is_dir(self, uri):
    	return not re.match("flickr://photo/.*", uri)

    def get_direct_children(self, uri):
        self.logger.info("Getting children for %s"%uri)
        children = {}
        match = re.search("flickr://([^/]*)/([^/]*)(/page/(\d+))?", uri)
        if match:
            type = match.groups()[0]
            terms = match.groups()[1]
            if match.groups()[3]:
                page = match.groups()[3]
            else:
                page = "1"
            if type == "search":
                self.logger.info("Performing flickr text search")
                resp = self.fapi.photos_search(api_key=self.fapi_key, 
                                               text=terms, 
                                               per_page=self.per_page, 
                                               page=page)
            elif type == "tags":
                self.logger.info("Performing flickr tag search")
                resp = self.fapi.photos_search(api_key=self.fapi_key, 
                                               tags=terms, 
                                               per_page=self.per_page, 
                                               page=page)
            elif type == "group":
                self.logger.info("Retrieving from flickr group: %s"%terms)
                resp = self.fapi.groups_pools_getPhotos(api_key=self.fapi_key,
                                                        group_id=terms, 
                                                        per_page=self.per_page,
                                                        page=page)
            elif type == "user":
                self.logger.info("Retrieving from flickr user: %s"%terms)
                #Find the user's NSID
                resp = self.fapi.people_findByUsername(api_key=self.fapi_key, 
                                                       username=terms)
                if resp['stat'] != "fail":
                    nsid = resp.user[0]["nsid"]
                    resp = self.fapi.people_getPublicPhotos(
                                                    api_key=self.fapi_key,
                                                    user_id=nsid,
                                                    per_page=self.per_page,
                                                    page=page)
            else:
                #Unsupported request
                return children
            if resp['stat'] == "fail":
                self.logger.info("Flickr request failed: %s"
                                %repr(resp.err[0]["msg"]))
                return children
            index = 0
            for photo in resp.photos[0].photo:
                child_uri = "flickr://photo/%s/%s_%s.jpg"\
                            %(photo['server'], photo['id'], photo['secret'])
                child_name = photo['title']
                index += 1 
                children[child_uri] = ("file", "directory", child_name, index)
            if int(resp.photos[0]['pages']) > int(page):
                index += 1
                child_uri ="flickr://%s/%s/page/%d"\
                            %(type, terms, int(page) + 1)
                children[child_uri] = ("directory", 
                                       "directory", 
                                       "Next Page", 
                                       index)

        return children

    def has_children(self, uri):
        is_photo = re.match("flickr://photo/.*", uri)
        return not is_photo

    def next_location(self, from_uri, item_filter=None, root=None):
        """
        Return the uri just next to `from_uri`.
        """
        next_uri = None
        short_name = None
        if not re.match("flickr://photo/.*", from_uri):
            #Get a list of all the pictures in this slideshow
            self.pictures = self.get_direct_children(from_uri)
            next_uri = self.pictures.keys()[0]
            short_name = self.pictures[next_uri][2]
        else:
            for i in range(0, len(self.pictures)):
                if self.pictures.keys()[i] == from_uri:
                    next_uri = self.pictures.keys()[i+1]
                    short_name = self.pictures[next_uri][2]
        if next_uri and not re.match("flickr://photo/.*", next_uri):
            #Recursively go through flickr pages
            return self.next_location(next_uri, item_filter, root)
        
        return (next_uri, short_name)

    def open(self, uri, mode=None):
        """
        Convert flickr://photo in to HTTP URL and pass on to Gnome VFS.
        We need this intermediary step so that the flickr plugin gets 
        queried for next_location in slideshows instead of the Gnome VFS
        plugin.
        """
        if not re.match("^flickr://photo/.*", uri):
            return None

        uri = re.sub("^flickr://photo", "http://static.flickr.com", uri)
        uri = media_uri.Uri(uri)

        access_plugin = self.data_access.get("uri:%s" % uri.scheme)

        return access_plugin.open(uri)

    def remove_location(self, uri):
        """
        Unimplemented.
        """
        return None
      
#Test the plugin
if __name__ == '__main__': 

    from elisa.core.plugin_manager import PluginManager
    from elisa.core.cache_manager import CacheManager
    from elisa.core.application import Application
    from elisa.core.config import Config
    from elisa.core import common
    from twisted.internet import reactor

    app = Application()
    #app.setup()

    def test():
        flickr = FlickrPlugin()
        print flickr.get_direct_children("flickr://tags/hamster")
        print flickr.get_direct_children("flickr://search/jokosher")
        print flickr.get_direct_children("flickr://user/bees")
        print flickr.get_direct_children("flickr://group/56878018@N00")
        print flickr.next_location("flickr://search/gstreamer")
        reactor.stop()

    reactor.callLater(0.4, test)
    reactor.run()
