# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright (C) 2006,2007, Elisa team <contact@fluendo.com>

from elisa.core import plugin
from elisa.core.signals import data
from elisa.core.signals.application import application_started

try:
    import dbus
    if getattr(dbus, 'version', (0,0,0)) >= (0,41,0):
        import dbus.glib
except ImportError:
    dbus = None

try:
    import gconf
except ImportError:
    gconf = None

class HalMonitor:

    def __init__(self, new_volume_cb=None, del_volume_cb=None):
        self.new_volume_cb = new_volume_cb
        self.del_volume_cb = del_volume_cb
        
        self.volume_manager_config = {}
        
        self.bus = dbus.SystemBus()
        self.hal_manager = self.bus.get_object('org.freedesktop.Hal', 
	                                            '/org/freedesktop/Hal/Manager')
        self.hal_iface = dbus.Interface(self.hal_manager,
	                                    'org.freedesktop.Hal.Manager')
        self._hotplugged_devices = {}
        self._unplugged_udis = []

        self._cbs = {}

    def _get_cb(self, udi):
        cb = self._cbs.get(udi)
        if not cb:
            extras = {'udi': udi}
            cb = lambda *args: apply(self.device_property_modified,
                                     args, extras)
            self._cbs[udi] = cb
        return cb

    def start(self):
        self.stop_volume_manager_monitoring()
        self.hal_iface.connect_to_signal('DeviceAdded',
                                         self.device_added_callback)
        self.hal_iface.connect_to_signal('DeviceRemoved',
                                         self.device_removed_callback)

    def detect_coldplugged(self):
        udis = self.hal_iface.FindDeviceByCapability('volume')

        for udi in udis:
            # get a hal object for the volume referenced by udi
            volume = self.get_device_with_udi(udi)
            parent_uri = volume.GetProperty('info.parent')
            parent = self.get_device_with_udi(parent_uri)
            
            if parent.GetProperty('storage.removable') \
                and parent.GetProperty('storage.hotpluggable'):
                # if its parent is removable and hotpluggable,
                # then the volume has been coldplugged
                self.device_added_callback(udi)
                properties = [(u'volume.mount_point', False, False),
                            (u'volume.is_mounted', False, False)]
                self.device_property_modified(len(properties), properties, udi=udi)

    def stop(self):
        self.restore_volume_manager_monitoring()

    def stop_volume_manager_monitoring(self):
        """
        Neutralize some of the volume_manager monitoring settings so
        that the user won't see rhythmbox pop up when an iPod is
        inserted (for example).
        """
        if gconf:
            client = gconf.client_get_default()
            path = '/desktop/gnome/volume_manager'
            autoplay_cda = client.get_bool('%s/autoplay_cda' % path)
            autoplay_dvd = client.get_bool('%s/autoplay_dvd' % path)
            autoipod = client.get_bool('%s/autoipod' % path)
            self.volume_manager_config = {'autoplay_cda': autoplay_cda,
                                          'autoplay_dvd': autoplay_dvd,
                                          'autoipod': autoipod}
            for prop in self.volume_manager_config.keys():
                client.set_bool('%s/%s' % (path, prop), False)
            
    def restore_volume_manager_monitoring(self):
        """
        Restore the volume_manager gconf settings
        """
        if gconf:
            client = gconf.client_get_default()
            path = '/desktop/gnome/volume_manager'
            for prop, value in self.volume_manager_config.iteritems():
                client.set_bool('%s/%s' % (path, prop), value)
        
    def get_device_with_udi(self, udi):
        obj = self.bus.get_object('org.freedesktop.Hal', udi)
        device = dbus.Interface(obj, 'org.freedesktop.Hal.Device')
        return device

    def parent_is_ipod(self, device):
        parent_udi = device.GetProperty(u'info.parent')
        parent = self.get_device_with_udi(parent_udi)
        return parent.GetProperty(u'storage.model') and \
               parent.GetProperty(u'info.product').lower() == 'ipod'

    def device_added_callback(self, udi):
        mount_point_prop = u'volume.policy.desired_mount_point'
        
        if udi not in self._hotplugged_devices:
            device = self.get_device_with_udi(udi)
            if device.QueryCapability("volume"):
                name = None
                if self.parent_is_ipod(device):
                    fstype = 'ipod'
                    if device.PropertyExists(mount_point_prop):
                        name = device.GetProperty(mount_point_prop)
                elif device.PropertyExists(u'volume.disc.is_videodvd'):
                    # No VideoDVD-Access-Plugin available for now
                    pass
##                     if device.GetProperty(u'volume.disc.is_videodvd'):
##                         name = ''
##                         fs_type = 'dvd'
##                         self.new_volume_cb('VideoDVD','dvd','') 
                elif device.PropertyExists(u'volume.disc.has_audio'):
                    # Checking for Audio-CDs
                    if device.GetProperty(u'volume.disc.has_audio'):
                        name = device.GetProperty(u'block.device')
                        fstype = 'audiocd'
                        self.new_volume_cb('AudioCD', fstype, name)
                else:
                    fstype = 'file'
                    if device.PropertyExists(mount_point_prop):
                        name = device.GetProperty(mount_point_prop)
                if name:
                    self._hotplugged_devices[udi] = (name, fstype, '')
                    if udi not in self._unplugged_udis:
                        device.connect_to_signal('PropertyModified',
                                                 self._get_cb(udi))

                
    def device_removed_callback(self, udi):
        if udi in self._hotplugged_devices:
            device = self.get_device_with_udi(udi)
##             try:
##                 self.bus.remove_signal_receiver(self._get_cb(udi),
##                                                 signal_name='PropertyModified',
##                                                 named_service=device._obj._named_service,
##                                                 path=device._obj._object_path)
##             except:
##                 import pdb; pdb.set_trace()
            name, fstype, mount_point = self._hotplugged_devices[udi]
            del self._hotplugged_devices[udi]
            if mount_point and self.del_volume_cb:
                self.del_volume_cb(name, fstype, mount_point)
            if udi not in self._unplugged_udis:
                self._unplugged_udis.append(udi)
                
    def device_property_modified(self, nb_props, properties, **kw):
        udi = kw.get('udi')
        device = self.get_device_with_udi(udi)
        values = {}
        for (name, added, removed) in properties:
            values[name] = device.GetProperty(name)

        if values.get(u'volume.is_mounted') and self.new_volume_cb:
            mount_point = values.get(u'volume.mount_point')
            if mount_point:
                name, fstype, unset = self._hotplugged_devices[udi]
                self._hotplugged_devices[udi] = (name, fstype, mount_point)
                self.new_volume_cb(name, fstype, mount_point)
        

class HalPlugin(plugin.Plugin):
    __implements__ = plugin.IPlugin

    name = "hal"

    singleton = True

    def __init__(self):
        plugin.Plugin.__init__(self)
        application_started.connect(self.application_started)
        
        self._monitor = None
        if not dbus:
            self.logger.info("Python-dbus not found. Disabling HAL plugin")
            self.unload()
        else:
            try:
                self._monitor = HalMonitor(self.new_volume, self.del_volume)
                self._monitor.start()
            except Exception, ex:
                self.logger.info("Couln't initialize HAL monitor: %s" % str(ex))
                self.unload()

    def unload(self):
        if self._monitor:
            self._monitor.stop()
        return plugin.Plugin.unload(self)
        
    def application_started(self):
        if self._monitor:
            self._monitor.detect_coldplugged()

    def _send_signal(self, signal_type, name, fstype, mount_point):
        volume_uri = str('%s://%s' % (fstype, mount_point))
        signal = getattr(data, signal_type)
        name = str(name)
        signal.emit(name, volume_uri, None)
                
    def new_volume(self, name, fstype, mount_point):
        self.logger.info('New volume appeared : %s' % mount_point)
        self._send_signal('new_device', name, fstype, mount_point)
        
    def del_volume(self, name, fstype, mount_point):
        self.logger.info('Volume disappeared  : %s' % mount_point)
        self._send_signal('del_device', name, fstype, mount_point)

if __name__ == '__main__':    
    from twisted.internet import glib2reactor
    glib2reactor.install()
    
    from twisted.internet import reactor
    
    def new_volume(name, fstype, mount_point):
        print 'New volume appeared :', name, fstype, mount_point

    def del_volume(name, fstype, mount_point):
        print 'Volume disappeared  :', name, fstype, mount_point

    monitor = HalMonitor(new_volume, del_volume)
    monitor.start()

    reactor.run()
