# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2007, Philippe Normand <phil at base-art dot net>

from elisa.core.base_plugins import DataAccessPlugin, PlayerPlugin, DataAccessFile
from elisa.core.base_plugins import TreePlugin
from elisa.core.interfaces import IDataAccess, IMediaPlayer
from elisa.core import media_uri, common, menu
from elisa.core.signals.application import application_started
from elisa.core.gst.lastfm_src import LastFMSource
import gst

class DummyFD:

    def close(self):
        pass

class LastFMPlayer(DataAccessPlugin, PlayerPlugin):

    __implements__ = (IMediaPlayer, IDataAccess)

    singleton = True
    name = 'lastfm'
    menu_name = _('Last.FM radios')
    
    parent = 'music'
    provides = ['audio',]

    default_config = {'locations': ['lastfm://globaltags/rock',
                                    'lastfm://globaltags/jazz']}
    
    folder_icon_path = 'folder-radios.png'
    item_icon_path = 'radio.png'

    folder_action = None

    icons = ['lastfm_love', 'lastfm_ban'] + PlayerPlugin.icons

    provides = [ "uri:lastfm" ]
    uncacheable_uri_schemes = ['lastfm']


    def __init__(self):
        DataAccessPlugin.__init__(self)
        PlayerPlugin.__init__(self)
        #self.set_icon_path('radios.png')
        self.login, self.password = '', ''
        self.lastfmsrc = None
        self.gst_pipeline = None
        application_started.connect(self._application_started)
        
    def _application_started(self):
        plugin_manager = common.get_plugin_manager()
        audioscrobbler = plugin_manager.get_plugin_with_name('audioscrobbler')
        if audioscrobbler and audioscrobbler.enabled:
            config = audioscrobbler.get_config()
            self.login = config.get('user')
            self.password = config.get('password')
        else:
            self.unload()

    def get_parent(self):
        return PlayerPlugin.get_parent(self)
            
    def add_wizard_shortcuts(self):
        pass

    def play_item(self, menu_item):
        if not self.lock():
            return
        
        music_file = str(media_uri.Uri(location=menu_item.get_target_path()))
        player = self.get_player()

        def launch(r):
            self.logger.info('Playing radio %s' % music_file)
            player.play(start=True)
            player.un_mute()
            PlayerPlugin.play_item(self, menu_item)
            self.unlock()
            
        dfr = player.set_uri(music_file)
        if dfr:
            dfr.addCallback(launch)
        else:
            self.unlock()
            
    item_action = play_item

    def item_action_menu(self, menu_item):        
        play_pause = menu.MenuItem(short_name=_("Play"))
        play_pause.set_target_path("play/pause")
        play_pause.set_icon_path(self.play_icon)
        play_pause.set_action_callback(self.play_parent_item)

        love = menu.MenuItem(short_name=_("Love"))
        love.set_target_path("love")
        love.set_icon_path(self.lastfm_love_icon)
        love.set_action_callback(self.love_parent_item)
        love.disable()

        ban = menu.MenuItem(short_name=_("Ban"))
        ban.set_target_path("ban")
        ban.set_icon_path(self.lastfm_ban_icon)
        ban.set_action_callback(self.ban_parent_item)
        ban.disable()

        skip = menu.MenuItem(short_name=_("Skip"))
        skip.set_target_path("skip")
        skip.set_icon_path(self.seekf_icon)
        skip.set_action_callback(self.skip_parent_item)
        skip.disable()

        stop = menu.MenuItem(short_name=_("Stop"))
        stop.set_target_path("stop")
        stop.set_icon_path(self.stop_icon)
        stop.set_action_callback(self.stop_parent_item)
        stop.disable()

        volp = menu.MenuItem(short_name=_("Volume+"))
        volp.set_target_path("vol+")
        volp.set_icon_path(self.volume_up_icon)
        volp.set_action_callback(self.volp_parent_item)
        volp.disable()

        volm = menu.MenuItem(short_name=_("Volume-"))
        volm.set_target_path("vol-")
        volm.set_icon_path(self.volume_down_icon)
        volm.set_action_callback(self.volm_parent_item)
        volm.disable()

        menu_item.add_item(stop)
        menu_item.add_item(love)
        menu_item.add_item(ban)
        menu_item.add_item(play_pause)
        menu_item.add_item(skip)        
        menu_item.add_item(volm)
        menu_item.add_item(volp)
        
        TreePlugin.item_action_menu(self, menu_item)
        
        player = self.get_player()
        if player and str(menu_item.get_target_path()) == \
               str(player.get_playing_uri()):
            self.playing(menu_item)

    def playing(self, menu_item):
        PlayerPlugin.playing(self, menu_item)
        for target in ("love", "ban", "skip"):
            item = menu_item.get_item_with_target(target)
            if item:
                item.enable()
                
        #common.get_application().get_task_manager().add_task(task.Task(menu_item))
        
    def love_parent_item(self, menu_item):
        if self.lastfmsrc:
            self.lastfmsrc.love()
            
    def ban_parent_item(self, menu_item):
        if self.lastfmsrc:
            self.lastfmsrc.ban()

    def skip_parent_item(self, menu_item):
        if self.lastfmsrc:
            self.lastfmsrc.skip()

    # IDataAccess implementation

    def is_dir(self, uri):
    	return False

    def get_direct_children(self, uri):
        children = {}
        return children

    def has_children(self, uri):
        return False

    def next_location(self, from_uri, item_filter=None, root=None):
        """
        Return the uri just next to `from_uri`.
        """
        next_uri = None
        short_name = None
        return (next_uri, short_name)

    def open(self, uri, mode=None):
        """
        """
        handle = DataAccessFile(self, DummyFD())
        return handle

    def remove_location(self, uri):
        """
        Unimplemented.
        """
        return None

    def get_gst_pipeline(self):
        if not self.gst_pipeline:            
            self.gst_pipeline = gst.Pipeline('pipeline')

            self.lastfmsrc = LastFMSource('src', self.login, self.password)
            decoder = gst.element_factory_make('decodebin')
            queue = gst.element_factory_make('queue')
            convert = gst.element_factory_make('audioconvert')

            def on_new_decoded_pad(element, pad, last):
                caps = pad.get_caps()
                name = caps[0].get_name()
                apad = convert.get_pad('sink')
                if 'audio' in name:
                    if not apad.is_linked(): # Only link once
                        pad.link(apad)

            decoder.connect('new-decoded-pad', on_new_decoded_pad)
            
            volume = gst.element_factory_make('volume', 'vlm')
            sink = gst.element_factory_make('autoaudiosink', 'sink')

            self.gst_pipeline.add(self.lastfmsrc, decoder, queue, convert,
                                  volume, sink)

            self.lastfmsrc.link(decoder)
            convert.link(volume)
            volume.link(sink)
            
        return self.gst_pipeline
