# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright (C) 2006,2007, Elisa team <contact@fluendo.com>

from elisa.core.base_plugins import DataAccessPlugin, DataAccessFile
from elisa.core.interfaces import IDataAccess
from elisa.core import media_uri
from elisa.core.utils import misc
from elisa.extern import path
import os, re, sys

try:
    from elisa.extern import inotify
except ImportError:
    inotify = None

import mimetypes

class LocalFSPlugin(DataAccessPlugin):

    __implements__ = IDataAccess

    name = 'local_fs'

    provides = ["uri:file",]

    hidden_file_pattern = re.compile("\..*")
    hidden_fnmatch_pattern = '[!.]*'
    singleton = True

    def __init__(self):
        DataAccessPlugin.__init__(self)
        self._paths = {}
        try:
            self._inotify = inotify.INotify()
        except Exception, e:
            self._inotify = None
            self.logger.info(str(e))
        
    def get_local_path(self, uri):
        uri = misc.fs_encode(uri)
        m_uri = media_uri.Uri(location=uri)
        if m_uri.parts['params'].get('label'):
            del m_uri.parts['params']['label']
        
        uri = str(m_uri)
        if uri.startswith('file://'):
            uri = uri[7:]
        uri = misc.fs_encode(uri)
        return uri

    def validate_uri(self, uri):
        uri = media_uri.Uri(location=misc.fs_encode(uri))
        valid = False
        if uri.host and uri.host[0] in ('.', '/'):
            valid = True
        elif os.path.exists(uri.path):
            valid = True
        elif sys.platform == "win32":
            '''
            if (len(uri.path) >= 1) and ((uri.path[0] != '/') or (uri.path[0] != '\\')):
                uri.path = '/' + uri.path
            '''
            p = uri.host + uri.path
            if os.path.exists(p):
                valid = True
        return valid

    def create_menu_for_uri(self, uri, for_plugin, short_name=None):
        if short_name and self.hidden_file_pattern.match(short_name):
            return None

        return DataAccessPlugin.create_menu_for_uri(self, uri, for_plugin,
                                                    short_name=short_name)

    def is_dir(self, uri):
        path = self.get_local_path(uri)
        return os.path.isdir(path)

    def get_file_type(self, uri):
        """
        Return the type of file. If directory return empty string, either
        return the content-type of the uri
        """
        if isinstance(uri, media_uri.Uri):
            uri = misc.fs_encode(str(uri))

        file_type = ''
        if not self.is_dir(uri):
            file_type, sub_type = mimetypes.guess_type(uri)
            if file_type:
                parts = file_type.split('/')
                file_type = parts[0]
                if file_type == 'application':
                    # filter on extension
                    _, ext = os.path.splitext(uri)
                    ext = ext[1:]
                    custom_extensions = {'ogg': 'audio',
                                         }
                    file_type = custom_extensions.get(ext)
            else:
                # in case mimetypes did not return anything, filter on extension
                _, ext = os.path.splitext(uri)
                ext = ext[1:]
                custom_extensions = {'flv': 'video',
                                     'ogm': 'video'
                                     }
                file_type = custom_extensions.get(ext)
                
        return file_type

    def get_direct_children(self, uri):
        path = self.get_local_path(uri)
        children = {}
        
        if self.is_dir(uri):
            try:
                contents = os.listdir(path)
            except OSError, error:
                self.logger.info("Could not list directory %r: %s" % (path,
                                                                      error))
            else:
                contents.sort()
                uri = misc.fs_encode(uri)
                index = 0
                for fname in contents:
                    full_path = os.path.join(path, fname)
                    if self.hidden_file_pattern.match(fname) or \
                           not os.path.exists(full_path):
                        continue
                    try:
                        if os.path.isdir(full_path) and not os.listdir(full_path):
                            continue
                    except OSError:
                        continue
                    child_uri = "file://%s" % full_path
                    fname = misc.fs_decode(fname)
                    child_type = 'file'
                    if self.is_dir(child_uri):
                        child_type = 'directory'
                    file_type = self.get_file_type(child_uri)
                    if file_type or child_type == 'directory':
                        children[child_uri] = (child_type, file_type, fname,
                                               index)
                    index += 1
        return children

    def next_location(self, uri, root=None, files_only=True):
        """
        Return the uri just next to `from_uri`.
        """
        location = None
        short_name = None

        local = self.get_local_path(uri)
            
        if os.path.isdir(local):
            dir_path = local            
        else:
            dir_path = os.path.dirname(local)

        root_path = None
        walk_gen = None
        if root:
            root_path = self.get_local_path(root)
            walk_gen =  self._paths.get(root_path)
        else:
            for walker_root, walker in self._paths.iteritems():
                if local.startswith(walker_root):
                    walk_gen = walker
                    root_path = walker_root
                    break

        if not root_path:
            root_path = dir_path
            
        if not walk_gen:
            p = path.path(root_path)
            walk_gen = p.walk(self.hidden_fnmatch_pattern, errors='ignore')
            self._paths[root_path] = walk_gen
            if root_path != local:
                n = walk_gen.next()
                s_name = media_uri.Uri(n).get_label()
                if n != local:
                    self._add_to_history(root, n, s_name)
                while n != local:
                    try:
                        n = walk_gen.next()
                    except StopIteration:
                        return (None, None)
                    s_name = media_uri.Uri(n).get_label()
                    self._add_to_history(root, n, s_name)
                    
        try:
            next_file = misc.fs_decode(walk_gen.next())
        except StopIteration:
            del self._paths[root_path]
            next_file = None
                
        if next_file:
            location = 'file://' + next_file
            short_name = media_uri.Uri(location).get_label()
            if root:
                self._add_to_history(root, location, short_name)
                
        return location, short_name

    def get_parent(self, uri):
        parent = media_uri.Uri(uri).get_parent()
        return parent
        

    def open(self, uri, mode=None):
        """
        Open an uri and return DataAccessFile file
        """
        file = self.get_local_path(uri)
        if not mode:
            mode = 'r'
        handle = None
        if os.path.exists(file) and not os.path.isdir(file):
            try:
                handle = open(file, mode)
            except Exception, exc:
                self.logger.info("Couln't open %s : %s" % (file, exc))
            else:
                handle = DataAccessFile(self, handle)
        return handle

    def read(self, handle, size=-1):
        if size == -1:
            data = handle.get_file_handle().read()
        else:
            data = handle.get_file_handle().read(size)
        return data

    def remove_location(self, uri):
        """
        Delete data at location identified by its uri.
        """
        self.logger.info("Removing %s" % str(uri))
        os.unlink(self.get_local_path(uri))

    def monitor_uri(self, uri, callback):
        # FIXME: should fail properly if uri not monitorable for whatever reason
        if self.uri_is_monitorable(uri):
            inotify_cb = lambda iwp, filename, mask, parameter: \
                        callback(self._inotify_event_type(mask), \
                                "file://%s/%s" % (iwp.path, filename))
            inotify_instance = inotify.INotify()
            inotify_instance.watch(uri.path, auto_add = True, \
                                    mask = inotify.IN_CREATE \
                                            | inotify.IN_DELETE \
                                            | inotify.IN_ATTRIB \
                                            | inotify.IN_MOVED_TO \
                                            | inotify.IN_MOVED_FROM, \
                                    callbacks=(inotify_cb, None), \
                                    recursive=True)

    def unmonitor_uri(self, uri):
        # everything handled by inotify
        pass

    def _inotify_event_type(self, mask):
        if mask & inotify.IN_CREATE or mask & inotify.IN_MOVED_TO:
            return "ADDED"
        elif mask & inotify.IN_DELETE or mask & inotify.IN_MOVED_FROM:
            return "REMOVED"
        elif mask & inotify.IN_ATTRIB:
            return "CHANGED"
        else:
            inotify_instance = inotify.INotify()
            #uri = "file://%s/%s" % (iwp.path, filename)
            human = inotify_instance.self.flag_to_human(mask)
            self.logger.info("Unexpected inotify event: %s" % human)
            
    def uri_is_monitorable(self, uri):
        # FIXME: should ask inotify module if uri is monitorable or not instead
        if not self._inotify:
            return False
	
        return inotify != None
