# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.core.widgets import widget_level_zoomed, widget_item, video_widget_item
from pgm.timing import controller, modifier, keyframe
from pgm.form import image
from elisa.core.signals.menu import *
from elisa.core.signals.data import data_updated
from elisa.core import log, menu, media_uri, common


class WidgetLevel(widget_level_zoomed.WidgetLevelZoomed):
    
    def __init__(self, current_skin, menu_level, arrow_image=None,
                 visible_item_count=5, zoomed_items=3, zoom_factor=2):
        width = current_skin.get_window().canvas_size[0]
        widget_level_zoomed.WidgetLevelZoomed.__init__(self, arrow_image,
                                                       visible_item_count,
                                                       zoomed_items,
                                                       zoom_factor,
                                                       width=width)
        self._skin = current_skin
        self._mm = current_skin.get_menu_manager()
        self._menu_level = menu_level

        item_added.connect(self._item_added)
        item_removed.connect(self._item_removed)
        #data_updated.connect(self._data_updated)
        
    def __repr__(self):
        params = (self._added_widget_item, self._menu_level)
        return "<CachedWidgetLevel %d children on %r>" % params

    def lock_animation(self):
        return self._mm.is_locked()
    
    def is_locked(self):    
        return self._mm.is_locked()
    
    def unlock_animation(self):
        self._mm.unlock()
        
    def rotate_image(self, uri, angle):
        w_item = self.get_widget_item(self._active_item_index)
        w_item.rotate(uri, angle)

    def first_draw_level(self):
        widget_level_zoomed.WidgetLevelZoomed.first_draw_level(self)
        self._mm.go_menu_item_with_index(self.get_selected_rank())
        w_item = self.get_widget_item(self._active_item_index)
        menu_item = self._mm.get_menu_item_for_widget(w_item)
        if menu_item and w_item:
            w_item.set_current(True)
            w_item.show_label(True)
            if menu_item.has_items():
                w_item.show_arrow()
            else:
                w_item.hide_arrow()
                              
        if isinstance(self._menu_level, menu.MenuTree):
            self.fade(True)
            
    def fade(self, is_in, end_callback=None):
        if is_in:
            # Fade the menu in
            fade_timeline = [keyframe.KeyFrame(-255, 0), keyframe.KeyFrame(1, 1)]
        else:
            # Fade out
            fade_timeline = [keyframe.KeyFrame(1, 0), keyframe.KeyFrame(-255, 1)]

        widgets = self._skin.get_status_bar().get_widgets()
        widgets.extend([self._skin._background_widget, self])
        mod = modifier.Modifier(widgets, 'opacity_offset', fade_timeline,
                                operation=modifier.RELATIVE)
        ctrl = controller.Controller(begin=0, duration=1000, resolution=10,
                                     modifiers=[mod, ],
                                     end_behavior=controller.END,
                                     end_callback=end_callback,
                                     interpolation=controller.SINUSOIDAL)
        ctrl.start()
        
    def unlock_animation(self):
        widget_level_zoomed.WidgetLevelZoomed.unlock_animation(self)
        self._skin.get_menu_widget().unlock_animation()

    def get_widget_count(self):
        return len(self._mm.get_menu_item_children(self._menu_level))
    
    def add_item(self, menu_item):
        
        witem = None
        if menu_item.get_parent() != self._menu_level:
            return None

        if self.is_visible( self._mm.get_index_from_menu_item(menu_item) ):
            icon_path = menu_item.get_icon_path()
            thumbnailer = menu_item.get_thumbnailer()
            caption = menu_item.get_short_name()
            arrow_image = self.get_arrow_image()
            cached_widget_item = self._mm.get_widget_item_for_menu_item(menu_item)
            
            if cached_widget_item == None:
                
                # FIXME: plainly wrong; a type could be added to MenuItem to fix that
                # maybe short_name is enough ?
                if icon_path == 'movie.png':
                    witem = video_widget_item.VideoWidgetItem(menu_item,
                                                              caption=caption,
                                                              start_size=(self._item_width, self._item_width),
                                                              arrow_image=arrow_image)
                else:

                    witem = widget_item.WidgetItem( start_size=(self._item_width, self._item_width),
                                                   caption=caption,
                                                   arrow_image=arrow_image)

                if not icon_path:
                    icon_path = 'unknown.png'

                # display icon_path immediatly
                icon_path = self._skin.get_picture_path(icon_path)
                img = image.Image(icon_path)
                witem.set_background_image(img)
                
                # and defer thumbnail generation if a thumbnailer is available
                if thumbnailer:
                    thumbnail_size = int(self._skin.icon_size[0:3])

                    th_manager = common.get_application().get_thumbnail_manager()
                    uri_to_thumbnail = media_uri.Uri(menu_item.get_target_path())
                    dfr = th_manager.enqueue_action(thumbnailer,
                                                    uri_to_thumbnail,
                                                    witem,
                                                    thumbnail_size)
                    def error_case(failure):
                        log.Logger().info(str(failure))

                    dfr.addErrback(error_case)


                witem.set_menu_item(menu_item)
                self._mm.bind_menu_item_to_widget(menu_item, witem)
                self._added_widget_item += 1
            else:
                witem = cached_widget_item
                self.logger.debug("WARNING: try to add yet cached widget: %s"%(menu_item))
            if self._first_draw_level_done:
                index = self._mm.get_index_from_menu_item(menu_item)
                self._draw_widget_item(witem, index)
        else:
            self.logger.debug("WARNING: try to add invisible widget: %s"%(menu_item))

        return witem
    
    def remove_item(self, menu_item):
        parent = menu_item.get_parent()
        if parent != self._menu_level:
            return

        cached_widget_item = self._mm.get_widget_item_for_menu_item(menu_item) 
        if cached_widget_item:
            self._added_widget_item -= 1
            self.remove_widget_item(cached_widget_item)

        current_menu_item  = self._mm.get_current_menu_item()
        if current_menu_item == menu_item:
            if len(self.children) > 0:
                self._skin.move_left()
            else:
                self._skin.move_up()
            
    def get_widget_items(self, start , count):
        witem_list = []
        max_index = self.get_widget_count() - 1
        for index_offset in range(count):
            index = start + index_offset
            if index > max_index:
                break
            witem = self.get_widget_item(index)
            witem_list.append(witem)
        return witem_list
    
    def get_widget_item(self, index, create_in_not_exist=False):
        menu_item = self._mm.get_menu_item_from_index(index, self._menu_level)
        witem = self._mm.get_widget_item_for_menu_item(menu_item)   
        if not witem and create_in_not_exist:
            witem =self.add_item(menu_item)
        return witem
        
    def anim_end(self, anim_label):
        rank = self._active_item_index
        w_item = self.get_widget_item(rank)
        if w_item:
            menu_item = self._mm.get_menu_item_for_widget(w_item)

            if anim_label == 'left':
                previous = self.get_widget_item(rank+1)
            else:
                previous = self.get_widget_item(rank-1)

            previous.set_current(False)    
            w_item.set_current(True)

            if menu_item:
                w_item.show_label()
                # update arrow
                if menu_item.has_items():
                    w_item.show_arrow()
                else:
                    w_item.hide_arrow()

    def _item_added(self, menu_item, index):
        self.add_item(menu_item)    

    def _item_removed(self, menu_item):
        self.remove_item(menu_item)

    def _data_updated(self, filename, evt_type, location, plugin_name,
                      parent_uri, index):
        # This is WiP
        manager = self._mm
        menu_item = None
        try:
            level_uri = str(media_uri.Uri(location=self._menu_level.get_target_path()))
        except TypeError:
            pass
        else:
            if parent_uri == level_uri:
                menu_item = manager.get_menu_item_from_index(index,
                                                             menu_level=self._menu_level)
        #get_menu_item_from_uri(str(location))
        print 'menu_item : ', menu_item
        if menu_item:
            if evt_type == "new_file":
                self.add_item(menu_item)
            elif evt_type == "del_file":
                try:
                    self.remove_item(menu_item)
                except Exception, ex:
                    print ex
                    raise
