# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright (C) 2006,2007, Elisa team <contact@fluendo.com>

from elisa.core.base_plugins import DataAccessPlugin
from elisa.core import interfaces
from elisa.core import menu, common, media_uri
from elisa.core.signals.data import new_device, del_device
from elisa.core.signals.application import application_started
from elisa.core.utils.misc import handle_failure
from elisa.extern import upnp

class UPnPPlugin(DataAccessPlugin):
    """

    """
    __implements__ = interfaces.IDataAccess

    name = "upnp"
    config_doc = {'enable_log': 'boolean flag to enable/disable the UPnP logging'}
    default_config = {'enable_log': 0}
    provides = ['uri:upnp',]
    singleton = True
    uncacheable_uri_schemes = ['upnp']

    def __init__(self):
        DataAccessPlugin.__init__(self)
        self.cache = {}
        self.device_cache = {}
        application_started.connect(self._application_started)

    def _application_started(self):
        coherence = common.get_application().get_coherence()
        if not coherence:
            self.unload("UPnP client can't work without Coherence framework")
        else:
            self._coherence = coherence
            import louie
            # connect to louie signals
            louie.connect(self.detected_media_server,
                          'Coherence.UPnP.ControlPoint.MediaServer.detected',
                          louie.Any)
            louie.connect(self.removed_media_server,
                          'Coherence.UPnP.ControlPoint.MediaServer.removed',
                          louie.Any)
            
    def _get_upnp_target(self, uri):
        if uri.startswith('upnp:'):
            uri = "http:%s" % uri[5:]
        return uri

    def _cache_enabled(self):
        media_manager = common.get_application().get_media_manager()
        return media_manager.enabled()

    def _get_control_url(self, upnp_uri):
        url = upnp_uri.split('#')[0]
        return url

    def _service_in_db(self, control_url):
        uri = "upnp%s" % control_url[4:]
        media_manager = common.get_application().get_media_manager()
        if media_manager.enabled():
            media = media_manager.get_source_for_uri(uri)
        else:
            media = None
        return media is not None

    def detected_media_server(self, client, usn):
        device = client.device
        friendly_name = device.get_friendly_name()
        
        msg = 'Browsing services on device %s' % friendly_name
        self.logger.info(msg)

        services = device.get_services()
        self.logger.info("Device has %s services" % len(services))
        for service in services:
            self.logger.info("Service type: %s" % service.get_type())
            cds1 = "urn:schemas-upnp-org:service:ContentDirectory:1"
            cds2 = "urn:schemas-upnp-org:service:ContentDirectory:2"
            if service.get_type() in (cds1, cds2):
                #service.subscribe()
                control_url = service.get_control_url()
                if not self._service_in_db(control_url):
                    self.device_cache[usn] = (control_url, friendly_name)
                    dfr = self.browse(client, 0, requested_count=0,
                                      friendly_name=friendly_name)
                    dfr.addCallback(self.build_menus, friendly_name,
                                    control_url, 0)
                else:
                    self.build_menus(None, friendly_name, control_url, 0)

    def removed_media_server(self, usn):
        control_url, friendly_name = self.device_cache[usn]
        del self.device_cache[usn]
        uri = "upnp:%s#0" % control_url[5:]
        del_device.emit(str(friendly_name), uri, None)

    def browse(self, client, container_id, starting_index=0,
               requested_count=0, friendly_name=""):
        """ Create an UPnP client and return a Deferred
        """
        try:
            client = client.content_directory
        except:
            pass
        control_url = client.url
        
        recursive = self._cache_enabled()

        dfr = client.browse(container_id, starting_index=starting_index,
                            requested_count=requested_count,
                            recursive=recursive)
        
        def add_to_cache(results, child=None):
            if not child:
                root = upnp.content_directory.buildHierarchy(results,
                                                             container_id,
                                                             recursive)
                self.store_in_cache(control_url, container_id, friendly_name,
                                    root, client)
                
            else:
                root = child
            self.store_in_cache(control_url, root.id, friendly_name, root,client)
            for child in root.children:
                if isinstance(child, upnp.content_directory.Folder):
                    add_to_cache(None, child=child)
                else:
                    for url in child.urls.keys():
                        self.store_in_cache(url,None,
                                            friendly_name, child,client)
                    self.store_in_cache(control_url, child.id, friendly_name,
                                        child,client)
            return results

        dfr.addCallback(add_to_cache)
        return dfr

    def browse_more(self, friendly_name, upnp_uri, container_id,
                    item, for_plugin, client,starting_index=0,
                    requested_count=0):
        """ Browse a container and hookup the corresponding menu item
            to the loadmore callback.
        """
        control_url = self._get_control_url(self._get_upnp_target(upnp_uri))
        name, data,_ = self.get_from_cache(control_url, container_id)

        # -1 in Elisa's incremental browsing <=> 0 in UPnP specs
        if requested_count == -1:
            requested_count = 0

        def process(contents):
            callback = self.create_child_menuitems_for_uri
            item.set_loadmore_callback(callback, start_item=0, nb_items=-1,
                                       for_plugin=for_plugin,
                                       short_name=name)

        def got_results(results):
            contents = upnp.content_directory.buildHierarchy(results,
                                                             container_id,
                                                             self._cache_enabled())
            process(contents)
        dfr = self.browse(client, container_id, starting_index,
                          requested_count)
        dfr.addCallback(got_results)
        dfr.addErrback(handle_failure)
        return dfr


    def store_in_cache(self, control_url, container_id, friendly_name, data,
                       client=None):
        if container_id is not None:
            uri = "%s#%s" % (control_url, container_id)
        else:
            uri = control_url
        self.logger.debug("Storing %s to cache" % uri)
        self.cache[uri] = (friendly_name, data,client)

    def get_from_cache(self, control_url, container_id):
        uri = "%s#%s" % (control_url, container_id)
        return self.cache.get(uri,(None,None,None))
        
    def build_menus(self, results, friendly_name, control_url, container_id):
        """
        Build a menu according browsing results and dispatch it to the plugins
        """
        upnp_uri = "upnp:%s#%s" % (control_url[5:], container_id)
        new_device.emit(str(friendly_name), upnp_uri, None)

    def create_menu_for_uri(self, uri, for_plugin, short_name=None):
        item_filter = for_plugin.item_filter
        folder_icon_path = for_plugin.folder_icon_path
        item_action = for_plugin.item_action
        item_icon_path = for_plugin.item_icon_path
        item_action_menu = for_plugin.item_action_menu
        folder_action_menu = for_plugin.folder_action_menu
        provides = for_plugin.provides

        if self._cache_enabled():
            return DataAccessPlugin.create_menu_for_uri(self, uri,
                                                        for_plugin, short_name)

        upnp_target = self._get_upnp_target(uri)


        #print upnp_target, self.cache[upnp_target], self.cache.keys()
        friendly_name, root, client = self.cache.get(upnp_target,
                                                     (None, None, None))
        if not isinstance(root, upnp.content_directory.Item):
            child_nb = len(root.children)
            if child_nb:
                self.logger.info("Found an UPnP tree with %s children" % child_nb)
        target_path = uri

        if isinstance(root, upnp.content_directory.Item):
            media_type = root.get_media_type()
            self.logger.info("Media type: %s" % media_type)
            is_item = media_type in provides
        else:
            is_item = False
         
        is_dir = isinstance(root, upnp.content_directory.Folder)
                  
        if not short_name:
            short_name = friendly_name

        item = None
        if is_dir or is_item:
            item = menu.MenuItem(short_name=short_name)

            if is_dir:
                icon_path = folder_icon_path
                children_nb = root.child_count
            else:

                # print out infos about item's resources
                urls = root.get_urls()
                self.logger.info("Item %r has %s resources" % (short_name,
                                                               len(urls)))
                for k, i in urls.iteritems():
                    self.logger.info(" - %s %r" % (k,i))

                url = root.get_url(for_plugin.preferred_formats)
                if not url:
                    if len(urls) == 1:
                        url = urls.keys()[0]
                icon_path = item_icon_path or url
                children_nb = 0
                item.set_action_callback(item_action)
                target_path = url

            if callable(icon_path):
                icon_path = icon_path(uri)

            #item.set_items_count(children_nb)
            item.set_target_path(target_path)
            item.set_icon_path(icon_path)

            if children_nb:
                self.logger.debug("%s has %s children" % (repr(root.name),
                                                          children_nb))
                # check if we need to browse that container
                if children_nb != len(root.children):
                    self.browse_more(friendly_name, target_path,
                                     root.id, item, for_plugin, client)
                else:
                    callback = self.create_child_menuitems_for_uri
                    item.set_loadmore_callback(callback, 
                                               start_item=0,
                                               nb_items=-1,
                                               for_plugin=for_plugin,
                                               short_name=short_name)
            elif is_dir:
                return None

            # build action menu if leaf
            if is_item and item_action_menu:
                item_action_menu(item)

        return item

    def has_children(self, uri):
        has_children = False
        media_manager = common.get_application().get_media_manager()
        classic_has_children = True
        if media_manager.enabled():
            # this ain't so because data_access.has_children() doesn't accept
            # a plugin so that we could filter content. So we use the
            # menu_manager to know in which plugin we are...
            skin = common.get_application().get_skin()
            if skin:
                menu_manager = skin.get_menu_manager()
                menu_item = menu_manager.get_current_menu_item().get_parent()
                target_path = menu_item.get_target_path()
                if target_path in ('pictures','movies','music'):
                    has_children = media_manager.has_children(uri, target_path)
                    classic_has_children = False
        if classic_has_children:
            has_children = DataAccessPlugin.has_children(self, uri)
        return has_children

    def is_dir(self, uri):
        """
        return True if a directory
        """
        is_dir = True
        upnp_uri = self._get_upnp_target(uri)
        friendly_name, folder, client = self.cache.get(upnp_uri,(None,None,
                                                                 None))
        if folder:
            is_dir = isinstance(folder, upnp.content_directory.Folder)
        return is_dir

    def get_file_type(self, uri):
        """
        Return the type of file. If directory return empty string, either
        return the content-type of the uri
        """
        media_type = 'audio'
        parent_uri = self._get_upnp_target(self.get_parent(uri))
        uri = self._get_upnp_target(uri)
        friendly_name, folder,_ = self.cache.get(parent_uri,(None,None, None))
        if folder:
            for child in folder.children:
                if not isinstance(child, upnp.content_directory.Folder):
                    urls = child.get_urls()
                    keys = urls.keys()
                    keys.sort()
                    child_uri = keys[0]
                    if child_uri == uri:
                        media_type = child.get_media_type()
                        break
                
        return media_type

    def get_parent(self, for_uri):
        cached_item = None
        for_uri = self._get_upnp_target(for_uri)
        _, cached_itemn, _ = self.cache.get(for_uri)
        if cached_item:
            for uri, (short_name, item) in self.cache.iteritems():
                if item.id == cached_item.parentID:
                    return uri.replace('http','upnp')
        return None
    
    def get_direct_children(self, uri):
        uri = self._get_upnp_target(uri)
        children = {}
        control_url = self._get_control_url(uri)
        friendly_name, folder, client = self.cache.get(uri,(None,None,None))
        if folder and isinstance(folder, upnp.content_directory.Folder):
            index = 0
            for child in folder.children:
                is_dir = isinstance(child, upnp.content_directory.Folder)
                if is_dir:
                    child_uri = "upnp:%s#%s" % (control_url[5:], child.id)
                    child_type = 'directory'
                    media_type = ''
                else:
                    child_type = 'file'
                    urls = child.get_urls()
                    keys = urls.keys()
                    keys.sort()
                    child_uri = keys[0]
                    media_type = child.get_media_type()

                if type(child_uri) == unicode:
                    child_uri = child_uri.encode('utf-8')
                children[child_uri] = (child_type, media_type, child.name,
                                       index)
                index += 1
        return children
        
    def open(self, uri):
        """
        Open an uri and return DataAccessFile file
        """
        uri = media_uri.Uri(location=uri)
        uri.scheme = 'http'
        
        return self.get_data_plugin(str(uri)).open(uri)
    
    def close(self, data_access_file):
        pass
        
    def remove_location(self, uri):
        """
        Delete data at location identified by its uri.
        """

    def next_location(self, from_uri, root=None, files_only=True):
        """
        Return the uri just next to `from_uri`.
        """
        location, short_name = self._next_location_from_cache(from_uri)
        if location and root:
            self._add_to_history(root, location, short_name)
            
        return location, short_name
    
    def _next_location_from_cache(self, from_uri, files_only=True):
        uri = self._get_upnp_target(from_uri)
        next_uri, short_name = '', ''

        is_file = False
        cached_item = self.cache.get(uri)
        if not cached_item:
            uri = uri.replace('upnp','http')
            cached_item = self.cache.get(uri)
            is_file = True

        if cached_item:
            _, container, _ = cached_item
            short_name = container.name
            
            if container.get_id() == 0:
                first_child = container.children[0]
                next_uri, short_name, item = self._get_cached_item_with_id(first_child.id)
                
                if not is_file:
                    next_uri = next_uri.replace('http','upnp')
                short_name = first_child.name

            else:
                next_child = None
                if isinstance(container, upnp.content_directory.Folder):
                    try:
                        next_child = container.children[0]
                    except:
                        pass

                if not next_child:
                    parent_id = container.parentID

                    parent = self._get_cached_item_with_id(parent_id)
                    parent_uri, parent_name, parent_item = parent
                    
                    item = None
                    #next_child = None
                    
                    if parent_item:
                        idx = parent_item.children.index(container)
                        if isinstance(parent_item, upnp.content_directory.Folder):
                            try:
                                next_child = parent_item.children[idx+1]
                            except IndexError:
                                if isinstance(container,
                                              upnp.content_directory.Folder) and container.children:
                                    
                                    next_child = container.children[0]

                if not next_child:
                    parent_id = container.parentID
                    _, _, parent = self._get_cached_item_with_id(parent_id)
                    ok = False
                    while not ok:
                        try:
                            parent_parent_id = parent.parentID
                        except AttributeError:
                            break
                        _, _, parent_parent = self._get_cached_item_with_id(parent_parent_id)
                        idx = parent_parent.children.index(parent)
                        try:
                            next_child = parent_parent.children[idx+1]
                        except IndexError:
                            parent = parent_parent
                        else:
                            ok = True


                if next_child:
                    next_uri, short_name, item = self._get_cached_item_with_id(next_child.id)
                    if not isinstance(item, upnp.content_directory.Folder):
                        urls = item.get_urls()
                        keys = urls.keys()
                        keys.sort()
                        next_uri = keys[0]

                
                if item and not is_file:
                    next_uri = next_uri.replace('http','upnp')
                    
        return (next_uri, short_name)

    def _get_cached_item_with_id(self, id):
        for item_id, (uri, item, client) in self.cache.iteritems():
            if item.id == id:
                return (uri, item.name, item)
        return (None, None, None)
