# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2006, Florian Boucault <florian@boucault.ath.cx>
# Modified by Colin Laplace for use with the Metar library

from elisa.core.plugin import IPlugin
from elisa.core.base_plugins import TreePlugin
from elisa.core import common
from elisa.core.signals.application import application_started
from elisa.core.signals.menu import item_activated
from elisa.core.utils.sun import suntime
from elisa.core.widgets.dialog_box import DialogBox
from elisa.core.utils.misc import get_resource_path

import os
import time, calendar, datetime
import csv

try:
    from elisa.extern.metar import Metar
except ImportError:
    Metar = None

import string
import urllib
import math

BASE_URL = "http://weather.noaa.gov/pub/data/observations/metar/stations"

class WeatherPlugin(TreePlugin):
    """ Weather plugin fetching standard METAR data from the Internet and
        displaying them thanks to WeatherWidget
    """

    __implements__ = IPlugin

    name = "weather"
    menu_name = _("Weather")
    menu_help_string = _("Display Weather Information")

    config_doc = {'ICAO': 'Four letters designing your location; ' \
                           'a complete list can be found at ' \
                           'http://en.wikipedia.org/wiki/ICAO_airport_code',
                  'latitude': 'Automatically generated; ' \
                               'remove if you change your ICAO code',
                  'longitude': 'Automatically generated; ' \
                               'remove if you change your ICAO code',
                  'location': 'Automatically generated; ' \
                               'remove if you change your ICAO code'
                  }

    default_config = {"ICAO": "LEBL",
                      "latitude": "41.270",
                      "longitude": "2.070",
                      "location": "BARCELONA AIRPOR,ES"
                      }
    
    
    # translation table between weather indications returned by pymetar to
    # custom pixmaps; these are in data/weather/ and supposed to be png files
    weather_to_pixmap = {"day": {
                            "sun": "clear",
                            "suncloud": "clear_cloud", 
                            "cloud": "cloud", 
                            "clear": "clear", 
                            "fog": "fog", 
                            "storm": "storm", 
                            "rain": "rain", 
                            "snow": "snow",
                            "unknown": "clear"
                            },
                        "night": {
                            "sun": "night_clear",
                            "suncloud": "night_clear_cloud", 
                            "cloud": "night_cloud", 
                            "clear": "night_clear", 
                            "fog": "night_fog", 
                            "storm": "night_storm", 
                            "rain": "night_rain", 
                            "snow": "night_snow",
                            "unknown": "night_clear"
                            }
                        }

    def __init__(self):
        TreePlugin.__init__(self)
        application_started.connect(self.application_started)

        self._station_code = self.get_config().get("ICAO")
        self._latitude = self.get_config().get("latitude", 0)
        self._longitude = self.get_config().get("longitude", 0)
        self._location = self.get_config().get("location")

        stations_path = get_resource_path(self.__module__, "data/weather/stations.csv")                                                        
        
        # FIXME: do not check each time when weather config is made
        try:
            stations = csv.reader(open(stations_path, 'rb'))
            for row in stations:
                if len(row) > 7:
                    if (row[3] == self.get_config().get("ICAO")):
                        self._location = row[2] + ',' + row[0]
                        self._latitude = float(row[6])
                        self._longitude = float(row[7])
                        self.config['latitude'] = row[5].replace(' ', '')
                        self.config['longitude'] = row[6].replace(' ', '')
                        self.config['location'] = self._location
                        break
        except:
            self.logger.info("Error loading %s" % stations_path)

    def post_init(self):
        if Metar == None:
            url = "http://homepage.mac.com/wtpollard/Software/FileSharing4.html"
            self.logger.info("Metar module not found. Install it from %s" % url)
            self.get_parent().del_item(self)
            
    def application_started(self):
        self._skin = common.get_application().get_skin()
        if self._skin:
            self.set_action_callback(WeatherPlugin.toggle_weather_display)
        else:
            self.logger.info("No skin found, disabling plugin")
            self.unload()
    
    def retrieve_weather(self):
        """ Fetches the weather information on the Internet thanks to metar
            and fill in the weather report self._report
        """

        url = "%s/%s.TXT" % (BASE_URL, self._station_code)
        # FIXME: can hang the entire software
        urlh = urllib.urlopen(url)
        report = ''
        try:
            for line in urlh:
                if line.startswith(self._station_code):
                    report = line.strip()
                    self._report = Metar.Metar(line)
                    break
        except:
            raise Exception

    def get_icon_path_from_weather(self, weather_report):
        """ Returns a pixmap path corresponding to a given weather and time

            weather_report: report to consider containing weather and time
        """

        t = datetime.datetime.now()
        
        rise_time = suntime('rise', float(self._latitude),
                            float(self._longitude), 1, time.localtime()[7])
        set_time = suntime('set', float(self._latitude),
                           float(self._longitude), 1, time.localtime()[7])
        current_time = time.localtime()[3:5]

        night_time = (current_time > set_time) or (current_time < rise_time)

        weather = None

        for i in self._report.sky:
            for j in i:
                if (j == "SKC") or (j == "CLR"):
                    sky_weather = "sun"
                elif (j == "FEW") or (j == "SCT"):
                    sky_weather = "suncloud"
                elif (j == "FEW") or (j == "SCT"):
                    weather = "cloud"
                    break

        for i in self._report.weather:
            for j in i:
                if (j == "TS"):
                    weather = "storm"
                elif (j == "DZ") or (j == "RA"):
                    weather = "rain"
                    break
                elif (j == "SN") or (j == "SG") or (j == "IC") or (j == "PL") or (j == "GS"):
                    weather = "snow"
                    break
                elif (j == "BR") or (j == "FG") or (j == "FU"):
                    weather = "fog"

        if weather == None:
            weather = "unknown"

        if not night_time:
            pixmap = self.weather_to_pixmap["day"][weather]
        else:
            pixmap = self.weather_to_pixmap["night"][weather]
        
        return get_resource_path(self.__module__, "data/weather/%s.png" % pixmap)

    def toggle_weather_display(self):
        """ Show/Hide a WeatherWidget containing an updated weather report """

        if hasattr(self, "_weather_widget"):
            if self._weather_widget.visible:
                self._weather_widget.hide(animate=True)
                return
            else:
                self._skin._window.remove(self._weather_widget)
                del self._weather_widget

        try:
            self.retrieve_weather()
        except Exception, exception:
            self.logger.info("Unable to retrieve weather: %s" % str(exception))
            common.get_application().get_skin().notify(["Could not retrieve weather information"])
            return

        # Labels initialisation
        labels = []

        if not hasattr(self._report.temp, "string"): #Metar lib doesn't handle exceptions yet
            common.get_application().get_skin().notify(["Could not retrieve weather information"])
            return

        labels.append(self._location)
        labels.append(_("Temperature: %(tempC)s / %(tempF)s") % {"tempC": self._report.temp.string("C"),
                                                                 "tempF": self._report.temp.string("F")})

        # compute humidity: http://www.faqs.org/faqs/meteorology/temp-dewpoint/
        dp  = self._report.dewpt.value(units="K")
        t = self._report.temp.value(units="K")
        e = 6.11 * math.exp((2500000 / 461.5) * ((1 / 273.15) - (1 / dp)))
        es = 6.11 * math.exp((2500000 / 461.5) * ((1 / 273.15) - (1 / t)))
        RH = (e / es) * 100
        
        labels.append(_("Relative Humidity: %.0f%%") % (RH)) 
        if self._report.wind_speed is not None:
            labels.append(_("Wind speed: %(windS)s") % {"windS": self._report.wind_speed.string("KMH")})

        if self._report.wind_dir is not None:
            labels.append(_("Wind direction: %(windD)s (%(windC)s)") %
                          {"windD": self._report.wind_dir, "windC": self._report.wind_dir.compass()})

        if self._report.press is not None:
            labels.append(_("Pressure: %s") % (self._report.press.string("HPA")))

        labels.append(_("Dew Point: %(dewC)s / %(dewF)s") % {"dewC": self._report.dewpt.string("C"),
                                                             "dewF": self._report.dewpt.string("F")})

        labels.append(_("Sky Conditions:"))
        labels.append(self._report.sky_conditions("\n"))                

        # Icon path initialisation
        icon_path = self.get_icon_path_from_weather(self._report)

        #print "icon_path:", icon_path

        # Update the MenuItem icon
        self.set_icon_path(icon_path, send_event=True)

        # Create the weather widget, center it and add it to the window
        (win_width, win_height) = self._skin._window.canvas_size
        # FIXME: not too happy about that constant in the code
        weather_width = 600
        self._weather_widget = WeatherWidget(size=(weather_width, weather_width),
                                        position=(0, 30),
                                        icon_path=icon_path,
                                        labels=labels)
        self._weather_widget._skin = self._skin
        self._weather_widget.visible = False
        self._skin._window.add(self._weather_widget)
        self._weather_widget.show(animate=True, autohide=True)
   
 
# TODO: separate in another python source file
from pgm.form import rectangle, image, label, color
from pgm.timing import controller, modifier, keyframe
from pgm.layout import group
from pgm import widget

from twisted.internet import reactor

class WeatherWidget(group.Group):
    """ Weather widget displaying weather information: a picture which should
        correspond to the current weather and a list of labels which should
        contain various information about the weather
    """

    def __init__(self, size, position=(0, 0), icon_path="", labels=[]):
        """ icon_path: path to the picture to display
            labels: list of strings to display
        """
        group.Group.__init__(self, size, position)
    
        self.hide_timeout = 8
        self._timed_out = None
        self.visible = False

        # Create the icon widget
        icon_width = size[0] * 0.40
        self._icon_widget = rectangle.Rectangle(size=(icon_width, icon_width),
                                                position=(0, 0))
        if icon_path:
            self._icon_widget.media = image.Image(icon_path)
            self._icon_widget.ordering = widget.NEAR
            self._icon_widget.opacity = 255
            self.add(self._icon_widget)
       
        # Create the labels widgets
        self._labels = []
        space_between_labels = 8
        pos_x = icon_width + 10
        pos_y = 0

        for caption in labels:
            # FIXME: if width of label is too small, text can get squeezed
            # Should investigate in Cairo/Pango to pre-determine rendered size
            # and permit to dynamically set a size to a Label
            current_label = label.Label(size=(size[0], 50),
                                        position=(pos_x, pos_y),
                                        text=caption,
                                        font_size=20)

            current_label.ordering = widget.NEAR
            current_label.opacity = 255
            current_label.x_align = 0.
            
            self.add(current_label)
            self._labels.append(current_label)
            pos_y = pos_y + current_label.text_size[1] + space_between_labels
        
#    def size__set(self, size):
#        if hasattr(self, "_icon_widget") and hasattr(self, "_labels"):
#            icon_width = size[0] * 0.40
#            self._icon_widget.size = (icon_width, icon_width)
#
#            for label in self._labels:
#                label.size = (size[0]-icon_width, 50)

#    def size__get(self):
#        return self.size

    def hide(self, animate=False):
        self._timed_out = None
        def _hide(c):
            self.visible = False
        if animate:
            self.animate(cb=_hide)
        else:
            self.visible = False

        skin = common.get_application().get_skin()
        skin.show_menu()

    def show(self, animate=True, autohide=True):
        if animate:
            self.animate(autohide=autohide)
        else:
            self.visible = True

        # hide skin menu
        skin = common.get_application().get_skin()
        skin.hide_menu()
        
        if autohide:
            self._timed_out = reactor.callLater(self.hide_timeout, self.hide,
                                                True)

    def animate(self, autohide=False, cb=None):
        if autohide:
            self.loop = True
        else:
            self.loop = False
        self.timeout = 1

        if self.timeout:
            win_width = self._skin._window.canvas_size[0]
            if self.visible:
                start = win_width / 2.0 - self.width / 2.0 - 100
                end = win_width
            else:
                start = -self.width
                end = win_width / 2.0 - self.width / 2.0 - 100
                self.visible = True

            duration = self.timeout * 1000
            timeline = [ keyframe.KeyFrame(start, 0),
                             keyframe.KeyFrame(end, 1) ]

            prop = 'x'
            mod = modifier.Modifier([self,], prop, timeline)
            self._ctrl = controller.Controller(begin=0, duration=duration,
                                               resolution=10,
                                               modifiers=[mod],
                                               end_behavior=controller.END,
                                               interpolation=controller.SINUSOIDAL,
                                               end_callback=cb)
            if self._skin:
                self._ctrl.start()
