# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from twisted.trial import unittest
from twisted.internet import defer
from elisa.core import menu_manager
from elisa.core import menu, log
from elisa.core.signals.menu import *
from pgm import log as pgm_log

class Widget:
    """
    I don't want yet to use Pigment's Widget. Because! :)
    """
    pass

log.Logger().disable_console_output()
log.Logger().set_level('DEBUG')

#pgm_log.set_level(pgm_log.DEBUG)

class TestMenuManager(unittest.TestCase):

    def new_menu_manager(self, menu_tree):
        return menu_manager.MenuManager(menu_tree)
    
    def test_empty_menu(self):
        manager = self.new_menu_manager(menu.MenuTree())
        self.assertEqual(len(manager._history), 1)
        self.assertEqual(manager._history[0].get_menuitem(),
                         manager.get_root_item())

    def test_menu_bind(self):
        tree = menu.MenuTree()
        a = menu.MenuItem(short_name='a')
        b = menu.MenuItem(short_name='b')
        tree.add_item(a)
        tree.add_item(b)

        widget_a = Widget()
        level_widget_a = Widget()
        widget_b = Widget()
        manager = self.new_menu_manager(tree)

        def _test_bind(r):
            manager.bind_menu_item_to_widget(a, widget_a)
            self.assertEqual(manager.get_widget_item_for_menu_item(a), widget_a)

            manager.bind_menu_level_to_widget(a, level_widget_a)
            self.assertEqual(manager.get_level_widget_for_menu_item(a),
                             level_widget_a)
     
        dfr = manager.build_root_menu()
        dfr.addCallback(_test_bind)
        return dfr

    def test_menu_item_children(self):
        tree = menu.MenuTree()
        a = menu.MenuItem(short_name='a')
        b = menu.MenuItem(short_name='b')
        tree.add_item(a)
        tree.add_item(b)
                      
        manager = self.new_menu_manager(tree)

        def check_children(r):
            children = manager.get_menu_item_children(tree)
            self.assertEqual(len(children), 2)
            self.assertEqual(children[0], a)
            self.assertEqual(children[1], b)
            self.assertEqual(a.get_parent(), tree)
            self.assertEqual(a.get_root(), tree)
            
        dfr = manager.build_root_menu()
        dfr.addCallback(check_children)
        return dfr

    def test_bind_level(self):
        tree = menu.MenuTree()
        manager = self.new_menu_manager(tree)

        def _root_menu_built(root_item):
            root_menu_item = manager._get_current_hist_item()
            item_widget = Widget()
            manager.bind_menu_level_to_widget(root_menu_item, item_widget)
            _root_menu_built.called = True
              
            
        def check_calls(r):
            self.assertEqual(_root_menu_built.called, True)
            root_menu_built.disconnect(_root_menu_built)
            
        _root_menu_built.called = False
        root_menu_built.connect(_root_menu_built)
        dfr = manager.build_root_menu()
        dfr.addCallback(check_calls)
        return dfr
        
    def test_bind_item(self):
        tree = menu.MenuTree()
        manager = self.new_menu_manager(tree)

        def _root_menu_built(root_item):
            root_menu_item = manager._get_current_hist_item()
            item_widget = Widget()
            manager.bind_menu_item_to_widget(root_menu_item, item_widget)
            _root_menu_built.called = True
              
            
        def check_calls(r):
            self.assertEqual(_root_menu_built.called, True)
            root_menu_built.disconnect(_root_menu_built)
            
        _root_menu_built.called = False
        root_menu_built.connect(_root_menu_built)
        dfr = manager.build_root_menu()
        dfr.addCallback(check_calls)
        return dfr
    
    def test_one_level(self):
        tree = menu.MenuTree()
        manager = self.new_menu_manager(tree)
        first = menu.MenuItem(short_name="first")
        tree.add_item(first)

        def _root_menu_built(root_item):
            self.assertEqual(root_item, tree)
##             self.assertEqual(len(root_item.get_children()), 1)
##             first_bis = root_item.get_children()[0]
##             self.assertEqual(first_bis.get_menuitem(), first)
            _root_menu_built.called = True
            
        def _item_added(item, index):
            self.assertEqual(item, first)
            self.assertEqual(index, -1)
            _item_added.called = True
            
        def _level_added(menu_item):
            self.assertEqual(menu_item, tree)
            _level_added.called = True

        def check_calls(dummy):
            self.assertEqual(_root_menu_built.called, True)
            self.assertEqual(_item_added.called, True)
            self.assertEqual(_level_added.called, True)
            level_added.disconnect(_level_added)
            item_added.disconnect(_item_added)
            root_menu_built.disconnect(_root_menu_built)
            
        _root_menu_built.called = False
        _item_added.called = False
        _level_added.called = False
        root_menu_built.connect(_root_menu_built)
        item_added.connect(_item_added)
        level_added.connect(_level_added)
        dfr = manager.build_root_menu()
        dfr.addCallback(check_calls)
        return dfr

    def test_navigate_forward_backward(self):
        """
        Put 3 items in our menu, and select the first one, then the
        second one, and navigate back and forth.
        """
        tree = menu.MenuTree()
        manager = self.new_menu_manager(tree)
        first = menu.MenuItem(short_name="first")
        second = menu.MenuItem(short_name="second")
        last = menu.MenuItem(short_name="last")
        tree.add_item(first)
        tree.add_item(second)
        tree.add_item(last)
        
        def got_first(old_menu_item, old_widget,
                      menu_item, widget, direction, step, forward):
            self.assertEqual(direction, u'horizontal')
            if forward:
                self.assertEqual(step, 1)
            else:
                self.assertEqual(step, -1)
                self.assertEqual(old_menu_item, second)
            self.assertEqual(menu_item, first)
            got_first.called = True
            
        def got_second(old_menu_item, old_widget,
                       menu_item, widget, direction, step, forward):
            self.assertEqual(direction, u'horizontal')
            if forward:
                self.assertEqual(old_menu_item, first)
                self.assertEqual(step, 1)
            else:
                self.assertEqual(step, -1)
                self.assertEqual(old_menu_item, last)
            self.assertEqual(menu_item, second)
            got_second.called = True
            
        def navigate_forward(result):
            got_first.called = False
            got_second.called = False

            go = manager.go_next_menu_item
            forward = True

            # we should get the first item
            active_changed.connect(got_first, forward,
                                   type_check=False)
            go()
            active_changed.disconnect(got_first)
            manager.unlock()
            
            # then the second one!
            active_changed.connect(got_second,
                                   forward, type_check=False)
            go()
            active_changed.disconnect(got_second)
            manager.unlock()
            self.assertEqual(got_first.called, True)
            self.assertEqual(got_second.called, True)
            
            # go to last item
            go()
            manager.unlock()
            
        def navigate_backward(result):
            got_first.called = False
            got_second.called = False

            forward = False
            go = manager.go_previous_menu_item

            # let's see if we get the second item
            active_changed.connect(got_second,
                                   forward, type_check=False)
            go()
            active_changed.disconnect(got_second)
            manager.unlock()

            # and now the first!
            active_changed.connect(got_first, forward,
                                   type_check=False)
            go()
            active_changed.disconnect(got_first)
            manager.unlock()
            self.assertEqual(got_second.called, True)
            self.assertEqual(got_first.called, True)
                 
        dfr = manager.build_root_menu()
        dfr.addCallback(navigate_forward)
        dfr.addCallback(navigate_backward)
        return dfr

    def test_item_activation(self):
        """
        Put one item in the menu and activate it
        """
        tree = menu.MenuTree()
        manager = self.new_menu_manager(tree)
        first = menu.MenuItem(short_name="first")
        tree.add_item(first)
        
        def action_callback(menu_item):
            self.assertEqual(menu_item, first)
        
        def _item_activated(menu_item, widget):
            self.assertEqual(menu_item, first)
            _item_activated.called = True
            
        def activate_first(dummy):
            item_activated.called = False
            
            item_activated.connect(_item_activated, type_check=False)
            manager.go_next_menu_item()
            called = manager.activate_current_menu_item()
            self.assertEqual(called, False)
            
            first.set_action_callback(action_callback)
            called = manager.activate_current_menu_item()
            self.assertEqual(called, True)

            self.assertEqual(_item_activated.called, True)
            
        dfr = manager.build_root_menu()
        dfr.addCallback(activate_first)
        return dfr

    def test_level_navigation(self):
        tree = menu.MenuTree()        
        first = menu.MenuItem(short_name="level0")
        sub_level = menu.MenuItem(short_name="level1")
        sub_sub = menu.MenuItem(short_name="level2")
        sub3 = menu.MenuItem(short_name="level3")
        
        tree.add_item(first)
        first.add_item(sub_level)
        sub_level.add_item(sub_sub)
        sub_sub.add_item(sub3)
        
        manager = self.new_menu_manager(tree)

        def got_first(old_menu_item, old_widget,
                      menu_item, widget, direction, step, forward):
            self.assertEqual(len(active_changed.get_receivers()), 1)
            self.assertEqual(direction, u'vertical')
            self.assertEqual(menu_item, sub_level)
            if forward:
                self.assertEqual(step, 1)
                self.assertEqual(old_menu_item, first)
            else:
                self.assertEqual(step, -1)
                self.assertEqual(old_menu_item, sub_sub)
            got_first.called = True
            
        def got_second(old_menu_item, old_widget,
                       menu_item, widget, direction, step, forward):
            self.assertEqual(direction, u'vertical')
            self.assertEqual(len(active_changed.get_receivers()), 1)
            self.assertEqual(menu_item, sub_sub)
            if forward:
                self.assertEqual(old_menu_item, sub_level)
                self.assertEqual(step, 1)
            else:
                self.assertEqual(step, -1)
                self.assertEqual(old_menu_item, sub3)
            got_second.called = True

        def navigate_backward(result):
            got_first.called = False
            got_second.called = False

            forward = False
            go = manager.go_up_level

            # let's see if we get the second item
            active_changed.connect(got_second,
                                   forward, type_check=False)
            go()         
            active_changed.disconnect(got_second)

            # and now the first!
            active_changed.connect(got_first, forward,
                                   type_check=False)
            go()
            active_changed.disconnect(got_first)
         
            self.assertEqual(got_first.called, True)
            self.assertEqual(got_second.called, True)

            
        def navigate_forward(result):
            got_first.called = False
            got_second.called = False

            go = manager.go_down_level
            forward = True

            # we should get the first item
            active_changed.connect(got_first, forward,
                                   type_check=False)
            dfr = go()
            
            def final_part(dummy):
                active_changed.disconnect(got_second)
                self.assertEqual(got_first.called, True)
                self.assertEqual(got_second.called, True)
                
                # go to last item
                last = go()
                last.addCallback(navigate_backward)
                return last
            
            def next_part(dummy):
                active_changed.disconnect(got_first)
                
                # then the second one!
                active_changed.connect(got_second,
                                       forward, type_check=False)
                dfr = go()
                dfr.addCallback(final_part)
                return dfr
            
            dfr.addCallback(next_part)            
            return dfr
            
            
        dfr = manager.build_root_menu()
        dfr.addCallback(navigate_forward)
        return dfr
    test_level_navigation.skip = "this is working like hell"
    
    def test_widget_binding(self):
        tree = menu.MenuTree()
        manager = self.new_menu_manager(tree)
        first = menu.MenuItem(short_name="first")
        tree.add_item(first)

        # fake widgets!
        item_widget = Widget()
        level_widget = Widget()

        def _item_added(item, index):
            manager.bind_menu_item_to_widget(item, item_widget)
            
        def _level_added(menu_item):
            manager.bind_menu_level_to_widget(menu_item, level_widget)
            
        def _item_bound(menu_item, widget):
            self.assertEqual(widget, item_widget)
            _item_bound.called = True

        def _level_bound(menu_item, widget):
            self.assertEqual(widget, level_widget)
            self.assertEqual(menu_item, tree)
            _level_bound.called = True

        def _item_unbound(menu_item, widget):
            self.assertEqual(widget, item_widget)
            self.assertEqual(menu_item, first)            
            _item_unbound.called = True

        def _level_unbound(menu_item, widget):
            self.assertEqual(widget, None)
            _level_unbound.called = True

        def check_bound(dummy):
            self.assertEqual(_level_bound.called, True)
            self.assertEqual(_item_bound.called, True)

        def check_unbound(dummy):
            #self.assertEqual(_level_unbound.called, True)
            #self.assertEqual(_item_unbound.called, True)            
            item_added.disconnect(_item_added)
            level_added.disconnect(_level_added)
            item_bound.disconnect(_item_bound)
            level_bound.disconnect(_level_bound)
            item_bound.disconnect(_item_unbound)
            level_unbound.disconnect(_level_unbound)
        

        def unbind(dummy):
            # print manager.get_root_item().get_widgets()
            # TODO
            pass
        

        _level_bound.called = False
        _item_bound.called = False
        _level_unbound.called = False
        _item_unbound.called = False
        item_added.connect(_item_added)
        level_added.connect(_level_added)
        item_bound.connect(_item_bound, type_check=False)
        level_bound.connect(_level_bound, type_check=False)
        item_unbound.connect(_item_unbound)
        level_unbound.connect(_level_unbound)
        dfr = manager.build_root_menu()
        dfr.addCallback(check_bound)
        dfr.addCallback(unbind)
        dfr.addCallback(check_unbound)
        return dfr

    def test_navigation_history(self):
        tree = menu.MenuTree()
        manager = self.new_menu_manager(tree)
        
        first_a = menu.MenuItem(short_name="1a")
        first_b = menu.MenuItem(short_name="1b")
        first_c = menu.MenuItem(short_name="1c")
        tree.add_item(first_a)
        tree.add_item(first_b)
        tree.add_item(first_c)

        second_a = menu.MenuItem(short_name="2a")
        second_b = menu.MenuItem(short_name="2b")
        first_b.add_item(second_a)
        first_b.add_item(second_b)

        def check_called(dummy, callback):
            self.assertEqual(callback.called, True)
            active_changed.disconnect(callback)
            manager.unlock()
            
        def got_first_b(old_menu_item, old_widget,
                        menu_item, widget, direction, step):
            self.assertEqual(menu_item, first_b)
            if direction == u'horizontal':
                # coming from first_a horizontal neighbour
                if step == 1:
                    self.assertEqual(old_menu_item, first_a)
            else:
                # we're coming from second_b
                if step == -1:
                    self.assertEqual(old_menu_item, second_b)
            got_first_b.called = True

        def got_second_a(old_menu_item, old_widget,
                         menu_item, widget, direction, step):
            self.assertEqual(menu_item, second_a)
            got_second_a.called = True

        @log.debug("second")
        def got_second_b(old_menu_item, old_widget,
                         menu_item, widget, direction, step):
            self.assertEqual(menu_item, second_b)
            got_second_b.called = True

        def go_to_first_b(dummy):
            manager.go_next_menu_item()
            manager.unlock()
            got_first_b.called = False
            active_changed.connect(got_first_b, type_check=False)
            manager.go_next_menu_item()
            check_called(None, got_first_b)

        def go_to_second_a(dummy):
            got_second_a.called = False
            dfr = manager.go_down_level()
            active_changed.connect(got_second_a, type_check=False)   
            dfr.addCallback(lambda dummy: manager.go_next_menu_item())
            dfr.addCallback(check_called, got_second_a)
            return dfr

        def go_to_second_b(dummy):
            got_second_b.called = False
            active_changed.connect(got_second_b, type_check=False)
            manager.go_next_menu_item()
            check_called(None, got_second_b)

        def go_back_to_first_b(dummy):
            got_first_b.called = False
            active_changed.connect(got_first_b, type_check=False)
            manager.go_up_level()
            check_called(None, got_first_b)

        def go_again_to_second_b(dummy):
            got_second_b.called = False
            active_changed.connect(got_second_b, type_check=False)   
            dfr = manager.go_down_level()
            dfr.addCallback(check_called, got_second_b)
            return dfr
            
        dfr = manager.build_root_menu()
        dfr.addCallback(go_to_first_b)
        dfr.addCallback(go_to_second_a)
        #dfr.addCallback(go_to_second_b)
        #dfr.addCallback(go_back_to_first_b)
        #dfr.addCallback(go_again_to_second_b)
        return dfr

    def test_dynamic_items(self):
        tree = menu.MenuTree()
        manager = self.new_menu_manager(tree)
        
        first_a = menu.MenuItem(short_name="1a")
        first_b = menu.MenuItem(short_name="1b")
        first_b2 = menu.MenuItem(short_name="1b2")
        first_c = menu.MenuItem(short_name="1c")
        tree.add_item(first_a)
        tree.add_item(first_b)
        tree.add_item(first_c)

        def check_called(dummy, callback, signal):
            self.assertEqual(callback.called, True)
            signal.disconnect(callback)
            manager.unlock()
            
        def got_first_a(old_menu_item, old_widget,
                        menu_item, widget, direction, step):
            self.assertEqual(menu_item, first_a)
            self.assertEqual(old_menu_item, first_b2)
            got_first_a.called = True
            
        def got_first_b(old_menu_item, old_widget,
                        menu_item, widget, direction, step):
            self.assertEqual(menu_item, first_b)
            self.assertEqual(old_menu_item, first_a)
            got_first_b.called = True

        def got_first_b2(old_menu_item, old_widget,
                         menu_item, widget, direction, step):
            self.assertEqual(menu_item, first_b2)
            self.assertEqual(old_menu_item, first_b)
            got_first_b2.called = True

        def first_b2_added(hist_item, index):
            self.assertEqual(hist_item, first_b2)
            self.assertEqual(index, 2)
            first_b2_added.called = True

        def first_b_removed(hist_item):
            self.assertEqual(hist_item, first_b)
            first_b_removed.called = True

        def go_to_first_b(dummy):
            manager.go_next_menu_item()
            manager.unlock()
            got_first_b.called = False
            active_changed.connect(got_first_b, type_check=False)
            manager.go_next_menu_item()
            check_called(None, got_first_b, active_changed)
        
        def add_first_b2(dummy):
            first_b2_added.called = False
            current = manager._get_current_hist_item().get_parent()
            item_added.connect(first_b2_added)
            manager._add_hist_item(first_b2, current, 2)
            check_called(None, first_b2_added, item_added)

        def go_to_first_b2(dummy):
            got_first_b2.called = False
            active_changed.connect(got_first_b2, type_check=False)
            manager.go_next_menu_item()
            check_called(None, got_first_b2, active_changed)

        def del_first_b(dummy):
            first_b_removed.called = False
            current = manager._get_current_hist_item().get_parent()
            item_removed.connect(first_b_removed)
            manager._remove_hist_item(first_b, current)
            check_called(None, first_b_removed, item_removed)

        def go_back_to_first_a(dummy):
            got_first_a.called = False
            active_changed.connect(got_first_a, type_check=False)
            manager.go_previous_menu_item()
            check_called(None, got_first_a, active_changed)
            
        dfr = manager.build_root_menu()
        dfr.addCallback(go_to_first_b)
        dfr.addCallback(add_first_b2)
        dfr.addCallback(go_to_first_b2)
        dfr.addCallback(del_first_b)
        dfr.addCallback(go_back_to_first_a)
        return dfr

    def test_dynamic_levels(self):
        """
        - Create a menu with one level including 3 items (1a, 1b, 1c)
        - Go to 1b
        - Add a new level under 1b containing 2a and 2b items
        
        """
        tree = menu.MenuTree()
        manager = self.new_menu_manager(tree)
        
        first_a = menu.MenuItem(short_name="1a")
        first_b = menu.MenuItem(short_name="1b")
        first_c = menu.MenuItem(short_name="1c")
        tree.add_item(first_a)
        tree.add_item(first_b)
        tree.add_item(first_c)

        second_a = menu.MenuItem(short_name="2a")
        second_b = menu.MenuItem(short_name="2b")

        def check_called(dummy, callback, signal):
            self.assertEqual(callback.called, True)
            signal.disconnect(callback)
            manager.unlock()
            
        def got_first_b(old_menu_item, old_widget,
                        menu_item, widget, direction, step):
            self.assertEqual(menu_item, first_b)
            self.assertEqual(old_menu_item, first_a)
            got_first_b.called = True

        def _level_added(menu_item):
            self.assertEqual(menu_item, first_b)
            _level_added.called = True

        def _level_removed(menu_item):
            self.assertEqual(menu_item, first_b)
            _level_removed.called = True
            
        def second_a_added(hist_item, index):
            self.assertEqual(hist_item, second_a)
            second_a_added.called = True
                             
        def second_b_added(hist_item, index):
            self.assertEqual(hist_item, second_b)
            second_b_added.called = True
                             
        def go_to_first_b(dummy):
            manager.go_next_menu_item()
            manager.unlock()
            got_first_b.called = False
            active_changed.connect(got_first_b, type_check=False)
            manager.go_next_menu_item()
            check_called(None, got_first_b, active_changed)
            
        def add_level(dummy):
            for cb in (level_added, second_a_added, second_b_added):
                cb.called = False
            current = manager._get_current_hist_item()
            level_added.connect(_level_added)
            level = manager._add_hist_level(current)
            check_called(None, _level_added, level_added)
            item_added.connect(second_a_added)
            manager._add_hist_item(second_a, current)
            check_called(None, second_a_added, item_added)
            item_added.connect(second_b_added)
            manager._add_hist_item(second_b, current)
            check_called(None, second_b_added, item_added)

        def remove_level(dummy):
            current = manager._get_current_hist_item()
            _level_removed.called = False
            level_removed.connect(_level_removed)
            manager._remove_hist_level(current)
            check_called(None, _level_removed, level_removed)
            
        dfr = manager.build_root_menu()
        dfr.addCallback(go_to_first_b)
        dfr.addCallback(add_level)
        dfr.addCallback(remove_level)
        return dfr


    def test_widget_cache(self):
        tree = menu.MenuTree()
        manager = self.new_menu_manager(tree)

        level1 = []
        for i in range(10):
            item = menu.MenuItem(short_name=str(i))
            level1.append(item)
            tree.add_item(item)
        
        level_widget = Widget()

        def _item_added(item, index):
            item_widget = Widget()
            item_widget.index = int(item.get_short_name())
            item.widget = item_widget
            manager.bind_menu_item_to_widget(item, item_widget)
            
        def _level_added(menu_item):
            manager.bind_menu_level_to_widget(menu_item, level_widget)

        def check_cached_widgets(dummy):
##             for i in range(5):
##                 self.assertEqual(level1[i].widget
            for i in range(5, 10):
                self.assertEqual(level1[i].widget.index, i)
            item_added.disconnect(_item_added)
            level_added.disconnect(_level_added)
            
        
        item_added.connect(_item_added)
        level_added.connect(_level_added)
        manager.set_widget_cache_size(5)
        dfr = manager.build_root_menu()
        dfr.addCallback(check_cached_widgets)
        return dfr

    def test_index_based_horizontal_navigation(self):
        tree = menu.MenuTree()
        a = menu.MenuItem(short_name="a")
        b = menu.MenuItem(short_name="b")
        c = menu.MenuItem(short_name="c")
        tree.add_item(a)
        tree.add_item(b)
        tree.add_item(c)
        
        manager = self.new_menu_manager(tree)

        def got_a(old_menu_item, old_widget, menu_item, widget,
                  direction, step):
            self.assertEqual(menu_item, a)
            got_a.called = True
            active_changed.disconnect(got_a)
        got_a.called = False

        def got_b(old_menu_item, old_widget, menu_item, widget,
                  direction, step):
            self.assertEqual(menu_item, b)
            got_b.called = True
            active_changed.disconnect(got_b)
        got_b.called = False

        def got_c(old_menu_item, old_widget, menu_item, widget,
                  direction, step):
            self.assertEqual(menu_item, c)
            got_c.called = True
            active_changed.disconnect(got_c)            
        got_c.called = False

        def go_to_index_two(r):
            active_changed.connect(got_c, type_check=False)
            manager.go_menu_item_with_index(2)
        
        def go_to_index_zero(r):
            active_changed.connect(got_a, type_check=False)
            manager.go_menu_item_with_index(0)

        def go_to_index_one(r):
            active_changed.connect(got_b, type_check=False)
            manager.go_menu_item_with_index(1)

        def check_calls(r):
            self.assertEqual(got_b.called, True)
            self.assertEqual(got_a.called, True)
            self.assertEqual(got_c.called, True)

        dfr = manager.build_root_menu()
        dfr.addCallback(go_to_index_two)
        dfr.addCallback(go_to_index_zero)
        dfr.addCallback(go_to_index_one)
        dfr.addCallback(check_calls)
        return dfr
        
    def test_circular_navigation(self):
        tree = menu.MenuTree()
        a = menu.MenuItem(short_name="a")
        b = menu.MenuItem(short_name="b")
        c = menu.MenuItem(short_name="c")
        tree.add_item(a)
        tree.add_item(b)
        tree.add_item(c)
        
        manager = self.new_menu_manager(tree)

        def got_a(old_menu_item, old_widget, menu_item, widget,
                  direction, step):
            self.assertEqual(menu_item, a)
            got_a.called = True
            active_changed.disconnect(got_a)
        got_a.called = False

        def got_a_again(old_menu_item, old_widget, menu_item, widget,
                        direction, step):
            self.assertEqual(menu_item, a)
            got_a_again.called = True
            active_changed.disconnect(got_a_again)
        got_a_again.called = False

        def got_c(old_menu_item, old_widget, menu_item, widget,
                  direction, step):
            self.assertEqual(menu_item, c)
            got_c.called = True
            active_changed.disconnect(got_c)            
        got_c.called = False

        def go_to_a(r):
            active_changed.connect(got_a, type_check=False)
            manager.go_next_menu_item()

        def go_back_to_c(r):
            active_changed.connect(got_c, type_check=False)
            manager.go_previous_menu_item()

        def go_to_a_again(r):
            active_changed.connect(got_a_again, type_check=False)
            manager.go_next_menu_item()
                        
        def check_calls(r):
            self.assertEqual(got_a_again.called, True)
            self.assertEqual(got_a.called, True)
            self.assertEqual(got_c.called, True)            

        dfr = manager.build_root_menu()
        dfr.addCallback(go_to_a)
        dfr.addCallback(go_back_to_c)
        dfr.addCallback(go_to_a_again)
        dfr.addCallback(check_calls)
        return dfr
    test_circular_navigation.skip = "Disabled in history_item"
    
    def test_menu_item_index(self):
        tree = menu.MenuTree()
        a = menu.MenuItem(short_name="a")
        b = menu.MenuItem(short_name="b")
        c = menu.MenuItem(short_name="c")
        tree.add_item(a)
        tree.add_item(b)
        tree.add_item(c)
        
        manager = self.new_menu_manager(tree)

        def check(r):

            # This is required because tested methods operate on current_level
            manager.go_next_menu_item()
            
            self.assertEqual(manager.get_menu_item_from_index(0), a)
            self.assertEqual(manager.get_index_from_menu_item(a), 0)
            self.assertEqual(manager.get_menu_item_from_index(1), b)
            self.assertEqual(manager.get_index_from_menu_item(b), 1)
            self.assertEqual(manager.get_menu_item_from_index(2), c)
            self.assertEqual(manager.get_index_from_menu_item(c), 2)

            # circular navigation is disabled
            self.assertEqual(manager.get_menu_item_from_index(-1), None)
            self.assertEqual(manager.get_menu_item_from_index(-2), None)

            self.assertEqual(None,
                             manager.get_menu_item_from_index(4))
            self.assertEqual(None,
                             manager.get_menu_item_from_index(-4))
            
        dfr = manager.build_root_menu()
        dfr.addCallback(check)
        return dfr
