# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

import unittest
import StringIO, urlparse
from elisa.core import media_uri
from elisa.core.utils import misc

media_uri.register_scheme('upnp')
media_uri.register_scheme('smb')


class TestMediaURI(unittest.TestCase):

    samples = ['http://google.com',
               'smb://philn/phil',
               'smb:///',
               'smb://phil:foo@philn/phil',
               'http://192.168.1.0:10243/foo.mp3',
               ]

    def test_samples(self):
        for url in self.samples:
            uri = media_uri.Uri(location=url)
            self.assertEqual(str(uri), url)
            if not url.endswith('/'):
                url2 = "%s/" % url
                uri = media_uri.Uri(location=url2)
                self.assertEqual(str(uri), url2)

    def test_smb(self):
        uri = media_uri.Uri(location="smb://philn:foo@phil/bare")
        expected_parts = {'password': 'foo', 'fragment': '',
                          'host': 'phil', 'path':'/bare',
                          'params': {},
                          'user': 'philn',
                          'scheme': 'smb', 'port': ''}
        parts = uri.get_parts()
        self.assertEqual(parts, expected_parts)

    def test_label(self):
        uri = media_uri.Uri(location='file:///home/phil/')
        print "path: ", uri.path
        print "label: ", uri.get_label()
        self.assertEqual(uri.path,'/home/phil/')
        self.assertEqual(uri.get_label(),'phil')

        uri = media_uri.Uri(location='file:///home/phil/?label=foo')
        self.assertEqual(uri.path,'/home/phil/')
        self.assertEqual(uri.get_label(),'foo')

    def test_escape(self):
        uri = media_uri.Uri(location='file:///home/phil/foo?.mp3')
        self.assertEqual(uri.path,'/home/phil/foo?.mp3')
        

    def test_unicode(self):
        u = 'file:///home/phil/\ufffd'
        uri = media_uri.Uri(location=u)
        print "str : ", str(uri)
        self.assertEqual(str(uri), u)
        u = u'file:///home/phil/\ufffd'
        uri = media_uri.Uri(location=u)
        self.assertEqual(str(uri), misc.fs_encode(u))
        
    def test_parameters(self):
        location = 'http://192.168.2.69:10243/WMCv2/%7b9B3A580F-68E8-4701-B8D8-9708E6E5E4F6%7d/103-1.wma?format=PCM'
        uri = media_uri.Uri(location=location)
        self.assertEqual(uri.params, {'format':'PCM'})

        # Windows Media Connect fancy urls
        location = 'http://192.168.1.87:10243/WMPNSSv4/3560585789/ezA1NDZBRDk5LTU3QkEtNENFOS04QjBFLThCQTUxRDFENUU0Qn0uMC5GRDNGRDY3Ng.jpg?width=684,formatID=24,height=456'
        uri = media_uri.Uri(location=location)
        self.assertEqual(uri.params, {'formatID':'24', 'width': '684',
                                      'height':'456'})
        self.assertEqual(str(uri), location)

    def test_join(self):
        location = 'http://foobar.com/foo'
        uri = media_uri.Uri(location=location)
        self.assertEqual(uri.path, '/foo')

        uri2 = uri.join('bar')
        self.assertEqual(uri2.path, '/foo/bar')
        self.assertEqual(uri.path, '/foo')

    def test_from_parts(self):
        location = 'http://foobar.com/foo'
        uri = media_uri.Uri(location=location)
        parts = uri.get_parts()

        uri2 = media_uri.Uri(parts=parts)
        self.assertEqual(str(uri2), location)

    def test_quote(self):
        location = 'upnp://192.168.2.171:2869/upnphost/udhisapi.dll?control=uuid:faf5c4e3-2a55-4a4c-a578-763f7823e1e0+urn:upnp-org:serviceId:ContentDirectory#0'
        uri = media_uri.Uri(location=location, quote=False)
        self.assertEqual(str(uri), location)

    def test_relative(self):
        location = 'file://./test/toto/'
        uri = media_uri.Uri(location=location)
        import os
        self.assertEqual(str(uri), 'file://%s/test/toto' % os.getcwd())

        location = 'file://./test/toto'
        uri = media_uri.Uri(location=location)
        self.assertEqual(str(uri), 'file://%s/test/toto' % os.getcwd())


        
if __name__ == "__main__":
    unittest.main()
