# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.core import media_uri
from elisa.core import log
from elisa.core.utils import deferred_action

from pgm.form import image

import os
import md5 # FIXME: md5 is deprecated in python 2.5, use hashlib instead

import PIL.Image
import PIL.PngImagePlugin


class ThumbnailManager(deferred_action.DeferredActionsManager):

    """
    Generate and retrieve thumbnails for any kind of file.

    It aims at respecting freedesktop thumbnails specification:
    http://www.freedesktop.org/wiki/Standards
    Its inheritance from deferred_action.DeferredActionsManager makes it
    accepting actions which will be processed sequentially in a separate thread.
    To add an action, use the method enqueue_action from DeferredActionsManager.
    """
    

    def _get_thumbnail_location(self, file_uri, size):
        """
        Return the thumbnail's location of a file for a particular size.

        :parameters:
            `file_uri` : media_uri.Uri
                Uri of the file
            `size` : int
                Size of the thumbnail (in pixels); must be <= 512
        
        :return:
            - Absolute path to the directory containing the thumbnail
            - Name of the png file containing the thumbnail
            - Maximum size of the thumbnail (it might be smaller)
        :rtype: (string, string, int)

        :exceptions:
            - `Exception` Raise an exception if size is superior to 512
        """
        # FIXME: if file_uri refers to ~/.thumbnails/*, return str(file_uri)

        if size <= 128:
            thumbnail_dir_size = "normal"
            thumbnail_max_size = 128
        elif size <= 256:
            thumbnail_dir_size = "large"
            thumbnail_max_size = 256
        elif size <= 512:
            # NOTE: this is not supported by the freedesktop specification
            thumbnail_dir_size = "extra_large"
            thumbnail_max_size = 512
        else:
            raise Exception("ThumbnailManager._get_thumbnail_location(): size given \
                             too large: %d" % size)

        thumbnail_dir = os.path.join(os.path.expanduser("~"),
                                    ".thumbnails",
                                    thumbnail_dir_size)
        thumbnail_filename = md5.new(str(file_uri)).hexdigest() + ".png"

        return (thumbnail_dir, thumbnail_filename, thumbnail_max_size)

    def _save_thumbnail_at(self, file_uri, thumbnail, thumbnail_dir, thumbnail_filename):
        """
        Save a thumbnail of a file at a given location.

        :parameters:
            `file_uri` : media_uri.Uri
                Uri of the file
            `thumbnail` : pgm.form.image.Image
                Thumbnail of the file
            `thumbnail_dir` : string
                Absolute path to the directory where to save the thumbnail
            `thumbnail_filename` : string
                Filename of the thumbnail
        """
        # convert from pgm.form.image to PIL.Image
        # FIXME: should use thumbnail.format
        thumbnail = PIL.Image.frombuffer("RGBA",
                                    thumbnail.size,
                                    thumbnail.pixels, "raw", "RGBA", 0, 1)

        info = PIL.PngImagePlugin.PngInfo()

        # required metadata
        info.add_text("Thumb::URI", str(file_uri))
        # FIXME: requires network transparent mtime
        #info.add_text("Thumb::MTime", os.path.getmtime(file_uri))

        # TODO: think about some useful (but optional) metadata to add
        # cf: http://jens.triq.net/thumbnailnail-spec/creation.html

        thumbnail_path = os.path.join(thumbnail_dir, thumbnail_filename)

        # if the thumbnail directory does not exist yet, create it
        if not os.path.exists(thumbnail_dir):
            os.makedirs(thumbnail_dir, 0700)

        # FIXME: access permissions on thumbnail must be set to 600
        thumbnail.save(thumbnail_path, "png", pnginfo=info)

    def _get_thumbnail_from(self, file_uri, thumbnail_dir, thumbnail_filename):
        """
        Retrieve a thumbnail of a file from a given location.

        :parameters:
            `file_uri` : media_uri.Uri
                Uri of the file
            `thumbnail_dir` : string
                Absolute path to the directory where to save the thumbnail
            `thumbnail_filename` : string
                Filename of the thumbnail

        :return: Thumbnail of the file
        :rtype: pgm.form.image.Image

        :exceptions:
            - `Exception` Raise an exception if an error occurs while loading
            - `Exception` Raise an exception if the thumbnail does not exist
        """
        thumbnail_path = os.path.join(thumbnail_dir, thumbnail_filename)

        # FIXME: should check modification time of original file and thumbnail
        # requires network transparent mtime
        thumbnail_is_valid = os.path.exists(thumbnail_path)
        
        if thumbnail_is_valid:
            # the thumbnail file exists and is valid, just load it
            try:
                thumbnail = image.Image(thumbnail_path)
                return thumbnail
            except IOError, error:
                raise Exception("Error loading %s : %s" % (thumbnail_path, error))
        else:
            raise Exception("Thumbnail for %s not generated yet" % file_uri)
    
    def get_thumbnail(self, file_uri, size, thumbnailer):
        """
        Retrieve the thumbnail of a file or generate it if it does not exist.

        :parameters:
            `file_uri` : media_uri.Uri
                Uri of the file
            `size` : int
                Size of the thumbnail (in pixels); must be <= 512
            `thumbnailer` : (file_uri, size) -> pgm.form.image.Image
                Function able to make a thumbnail of the file at the required size

        :return: Thumbnail of the file
        :rtype: pgm.form.image.Image
            
        :exceptions:
            - `Exception` Raise an exception if the thumbnail could not be generated
        """
        # TODO: check for local thumbnail directory as specified at:
        # http://jens.triq.net/thumbnail-spec/local.html

        # FIXME: catch exception if size is wrong
        thumbnail_dir, thumbnail_filename, thumbnail_max_size = self._get_thumbnail_location(file_uri, size)
        try:
            # retrieve thumbnail from cache if it exists
            thumbnail = self._get_thumbnail_from(file_uri, thumbnail_dir, thumbnail_filename)
        except:
            # generate the thumbnail
            try:
                # FIXME: not functional and should be rethinked if we want to respect freedesktop spec.
                """
                # if an empty png exists in ~/.thumbnails/fail/elisa/, do not try generation
                # FIXME: application name (elisa) hardcoded
                fail_dir = os.path.join(os.path.expanduser("~"), ".thumbnails", "fail", "elisa")
                fail_path = os.path.join(fail_dir, thumbnail_filename)
                #if os.path.exists(fail_path):
                #    raise Exception("Thumbnail generation failed previously for %s" \
                #                                    % file_uri)
                """
                thumbnail = thumbnailer(file_uri, thumbnail_max_size)
                if thumbnail:
                    self._save_thumbnail_at(file_uri, thumbnail, thumbnail_dir, thumbnail_filename)
            except Exception, e:
                # FIXME: create an empty png in ~/.thumbnails/fail/elisa/ if it failed
                #        instead of an empty file
                """
                if not os.path.exists(fail_dir):
                    os.makedirs(fail_dir, 0700)
                if not os.path.exists(fail_path):
                    os.mknod(fail_path, 0600)
                """
                raise
              
        # resize the thumbnail to the exact desired size if necessary
        if size != thumbnail_max_size:
            thumbnail.resize(size)

        return thumbnail


if __name__ == "__main__":
    from elisa.core import application
    from elisa.core.utils import image
    import sys

    if len(sys.argv) != 2:
        print "Usage: %s image_uri" % sys.argv[0]
        sys.exit()

    file_path = sys.argv[1]
    file_uri = media_uri.Uri(file_path)

    app = application.Application()
    app.setup()

    def img_thumbnailer(uri, size):
        thumbnail = image.Image(uri)
        thumbnail.resize(size)
        return thumbnail

    thumbnail_mgr = ThumbnailManager()
    thumb = thumbnail_mgr.get_thumbnail(file_uri, 256, img_thumbnailer)
    thumb_dir = os.path.join(os.path.expanduser("~"), "Desktop")
    thumb_filename = "test.png"
    thumbnail_mgr._save_thumbnail_at(file_uri, thumb, thumb_dir, thumb_filename)
