# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

#export GST_DEBUG=pygst:5
#ulimit -c unlimited
import sys, os
import pygst
pygst.require('0.10')
import gst, gobject, time

try:
    import gnomevfs
except ImportError:
    gnomevfs = None

try:
    import tagpy
except ImportError:
    tagpy = None

class TagBrowser(object):


    def __init__(self):
        gobject.threads_init()
        self._mp3_pipe = None 
        self._ogg_pipe = None   
        self._tags = {}
        self._lasterror = None
        self._rebuild_pipeline = False
        self.mimetype = None
        
    def get_type(self, filename):
        ext = filename[len(filename)-3:]
        if ext == 'mp3':
            return 'mp3'
        if ext == 'ogg':
            return 'ogg'

        return 'unknown'

    def pad_added(self, element, pad, pipeline):
        if not pad.is_linked():
            fakesink = pipeline.get_by_name('sink')
            pad.link(fakesink.get_pad('sink'))
    
    def get_pipeline(self, filename):
        ext = self.get_type(filename)
        self._tags = {}

        scheme, path = filename.split('://')
        if scheme == 'file':
            src = 'filesrc'
        else:
            src = 'gnomevfssrc'
        
        if ext == 'mp3':
            if self._mp3_pipe == None or self._rebuild_pipeline:
                self._mp3_pipe = gst.parse_launch('%s name=source ! id3demux name=demux ! fakesink name=sink' % src)
                demux = self._mp3_pipe.get_by_name('demux')
                if demux:
                    demux.connect ('pad-added', self.pad_added, self._mp3_pipe)

            return self._mp3_pipe
        elif ext == 'ogg':
            if self._ogg_pipe == None or self._rebuild_pipeline:
                self._ogg_pipe = gst.parse_launch('%s name=source ! oggdemux name=demux ! vorbisdec name=sink ! fakesink' % src)
                demux = self._ogg_pipe.get_by_name('demux')
                demux.connect ('pad-added', self.pad_added, self._ogg_pipe)
            return self._ogg_pipe
        return None

    def _have_type_cb(self, typefind, prob, caps):
        self.mimetype = caps.to_string()
  
                
    def get_tags(self, filename):
        scheme, path = filename.split('://')

        use_tagpy = False
        if scheme == 'file':
            filename = path
            if tagpy:
                use_tagpy = True
                
        elif gnomevfs:
            filename = gnomevfs.escape_host_and_path_string(filename)

        if use_tagpy:
            _, ext = os.path.splitext(filename)
            tags = {}
            # TODO: complete this list
            if ext[1:].lower() in ('ogg','mp3','flac'):
                f = tagpy.FileRef(filename)
                t = f.tag()
                if t:
                    artist = t.artist or ''
                    album = t.album or ''
                    title = t.title or ''
                    track = t.track or ''
                    tags = {'artist': artist.rstrip(),
                            'album': album.rstrip(), 'title': title.rstrip(),
                            'track-number': track}
            return tags

        #build or get existing pipeline
        pipeline = self.get_pipeline("%s://%s" % (scheme,path))
        if not pipeline:
            return {}
        
        self.typefind = pipeline.get_by_name("typefind")
        if self.typefind:
            self.typefind.connect("have-type", self._have_type_cb)
        
        
        #stop bus flushing
        bus = pipeline.get_bus()
        bus.set_flushing(False)
    
        #if pipeline is not rebuilded, put to READY state    
        if not self._rebuild_pipeline:
            if not self._change_state(pipeline, gst.STATE_READY):
                self.clean(pipeline)
                return {}
        
        #set filename and put to paused state
        #message will processed and tags will found if they are in first buffer (most cases)
        source = pipeline.get_by_name('source')
        source.set_property('location', filename)
        if not self._change_state(pipeline, gst.STATE_PAUSED):
            self.clean(pipeline)
            return {}
        
        #flushing and empty the bus
        #bus.set_flushing(True)
        #self._process_bus(pipeline)
        
        self.clean(pipeline)
        
        return self._tags
        
    def clean(self, pipeline):
        #clean if we want to rebuild pipeline each time
        if self._rebuild_pipeline:
            self._change_state(pipeline, gst.STATE_NULL)
        
        
    def _change_state(self, pipeline, state):
        r = self._change_state_no_bus(pipeline, state)
        self._process_bus(pipeline)
        return r
        
    def _change_state_no_bus(self, pipeline, state):
        pipeline.set_state(state)
        max_step = 10
        step = 0
        current_states = pipeline.get_state(50*gst.MSECOND)
        while gst.STATE_CHANGE_ASYNC == current_states[0]:
            step +=1
            time.sleep(0.01)
            if step > max_step:
                return False
            current_states = pipeline.get_state(50*gst.MSECOND)

        if current_states == (gst.STATE_CHANGE_SUCCESS, state,
                              gst.STATE_VOID_PENDING):
            return True
        
        return False

    def _process_bus(self, pipeline):
        bus = pipeline.get_bus()
        while True:
            time.sleep(0.01)
            msg = bus.pop()
            if not msg or msg.type==gst.MESSAGE_EOS:
                return True
            elif msg.type==gst.MESSAGE_TAG:
                self._tags = {}
                parsed_tags = dict(msg.parse_tag())
                for key, value in parsed_tags.iteritems():
                    if type(value) == str:
                        value = value.decode('utf-8')
                    self._tags[key] = value
                if pipeline != self._ogg_pipe:
                    self._tags['audio-codec'] = u''
            elif msg.type==gst.MESSAGE_ERROR:
                self._lasterror = msg.parse_error() 
                gerror, debug = self._lasterror
                print "error %s: %s"%(gerror, debug)
                return False
                
        return True

    def get_content_type(self, uri):
        tags = self.get_tags(uri)
        keys = tags.keys()
        if 'audio-codec' in keys:
            typ = 'audio'
        elif 'video-codec' in keys:
            typ = 'video'
        else:
            typ = None
        return typ
        
if __name__ == '__main__':
    import threading
    
    def fct():
        f=TagBrowser()
        for fname in sys.argv[1:]:
            print "-"*35
            print "for %s tags=%s"%(fname,f.get_tags(fname))
        if not sys.argv[1:]:
            fname="file:///home/yoyo/Elisa/sample_data/music/mp3-tags/Eels/Beautiful Freak/Eels - 05 - Not Ready Yet.mp3"
            print "for %s tags=%s"%(fname,f.get_tags(fname))
            print "-"*35
            #fname = "file:///home/yoyo/Elisa/sample_data/music/mp3-tags/Coldplay/Coldplay - Parachute/01 - Don't Panic.mp3"
            #print "for %s tags=%s"%(fname,f.get_tags(fname))
            #print "-"*35
            fname =  "file:///home/yoyo/Elisa/sample_data/music/mp3-tags/Coldplay/Coldplay - Parachute/02 - Shiver.mp3"
            print "for %s tags=%s"%(fname,f.get_tags(fname))
            print "-"*35
            fname="file:///home/yoyo/Elisa/sample_data/music/mp3-tags/B.O. Kill Bill 1/05_dialogue_-_queen_of_the_crime_council-xmr.mp3"
            print "for %s tags=%s"%(fname,f.get_tags(fname))
            fname="file:///home/yoyo/Elisa/sample_data/music/02_icarus_grounded__jeffrey_altergott.ogg"
            print "for %s tags=%s"%(fname,f.get_tags(fname))
            print "-"*35
            fname="file:///home/yoyo/Elisa/sample_data/music/01_open_source__magic_mushrooms.ogg"
            print "for %s tags=%s"%(fname,f.get_tags(fname))
            print "-"*35
            
    t = threading.Thread(target=fct)
    t.start()
