# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from pgm.layout import group

SCROLL_NORMAL = 0
SCROLL_MIDDLE = 1

from pgm import widget, render
from pgm.message.signal import Signal
from pgm.widget import Widget
from pgm.form import image
from elisa.core.widgets.action_item import ActionItem
from elisa.core.widgets import rectangle

class List(group.Group):
    active_changed = Signal('list-active-changed', (Widget, Widget, int))
    keys = {}

    def __init__(self, size, spacing, background_picture_path=None):
        group.Group.__init__(self, size)
        self.bg = rectangle.Rectangle(size=size)
        self.spacing = spacing
        self._max_items = 0
        self._scroll_prev_idx = -1
        self._scroll_next_idx = -1
        self._scroll_index = 1
        self._scroll_mode = SCROLL_NORMAL
        self._scroll_enabled = False
        self._scroll_visible = False
        self._selected = 0
        self._background_picture_path = background_picture_path
        self.set_key_actions()

        if self._background_picture_path:
            self.bg.media = image.Image(self._background_picture_path)

            # background
            self.bg.ordering = widget.FAR
            group.Group.add(self, self.bg)
            self.size = self.bg.size
            
    def _add_at_index(self, widget, idx):
        # Already got a parent ?
        if widget.parent == None:
            # Add the group position to the widget one            
            widget.position = map(lambda x,y: x+y,
                                  widget.position, self.position)
            self._children.insert(idx, widget)
            widget._parent = self
            # Already binded to a Window ? Let's add the Widget.
            if self._window:
                widget._set_window(self._window)
            return True
        else:
            return False

    def add(self, widget, index=-1):
        if widget == self.bg:
            group.Group.add(self, widget)
            return

        real_index = None
        if index != -1:
            real_index = index
            if self.bg in self:
                real_index += 1
            if self._scroll_enabled:
                real_index += 1
            selected = real_index
        else:
            selected = self._selected
            
        if self._scroll_enabled:
            if real_index is None:
                real_index = len(self._children) - 1
            self._add_at_index(widget, real_index)
        else:
            if real_index is None:
                group.Group.add(self, widget)
            else:
                self._add_at_index(widget, real_index)
                
        self.set_selected(selected)
        self.reorder()

    def dump(self):
        print 'Dumping list'
        for widget in self:
            print widget
        print 'Dumped'

    def remove(self, widget):
        if widget != self.bg:
            group.Group.remove(self, widget)
        self.set_selected(self._selected)
        self.reorder()

    def set_scrollbars(self, max_items, widget_prev, widget_next,
                       mode=SCROLL_NORMAL):
        if self._scroll_enabled:
            self.remove_scrollbars()

        self._add_at_index(widget_prev, 0)
        group.Group.add(self, widget_next)

        widget_prev.x = self.x + (self.width / 2.) - (widget_prev.width / 2.)
        widget_next.x = self.x + (self.width / 2.) - (widget_next.width / 2.)
        
        self._scroll_prev_idx = 0
        self._scroll_next_idx = self.index(widget_next)        
        
        self._scroll_mode = mode
        self._max_items = max_items
        self._scroll_index = 1
        self._scroll_enabled = True
        self._scroll_visible = True
        self.reorder()

    def show_scrollbars(self, show=True):
        self._scroll_visible = show
        self.reorder()

    def current_item__get(self):
        idx = self.get_selected()
        if self._scroll_enabled:
            idx += 1
        return self[idx]

    def get_selected(self):
        return self._selected

    def set_selected(self, idx):
        if (idx > -1) and (idx < len(self.children)):
            self._selected = idx
            if self._scroll_enabled:
                #skip scrollbars
                #print "idx:",self._scroll_index," sel:", self._selected
                if (self._scroll_mode == SCROLL_MIDDLE):
                    if self._selected >= self._scroll_index + round(float(self._max_items) / 2.)- 1:
                        self._scroll_index += 1
                    elif (self._selected < self._scroll_index) and self._scroll_index > 1:
                        self._scroll_index -= 1
                #elif (self._selected < len(self.children) - 1 - self._max_items):
                elif (self._selected >= self._scroll_index + self._max_items - 1) or \
                     (self._selected + 1 < self._scroll_index):
                    self._scroll_index = self._selected
                    
                start = 1
                end = len(self.children) - 1
                idx += 1
            else:
                start = 0
                end = len(self.children)
            i = start
            selected = -1
            while i < end:
                if self.bg in self and idx == self.index(self.bg):
                    self.children[i].opacity = 128
                    i+= 1
                    continue
                if i == idx:
                    self.children[i].opacity = 255
                    selected = i
                else:
                    self.children[i].opacity = 128
                i += 1

            if selected == -1:
                self.set_selected(idx+1)
            self.reorder()

    def select(self, idx):
        if self.bg in self:
            idx += 1
        
        current = 0
        if self._scroll_enabled:
            current += 1
            
        while current <= idx:
            self.next()
            self._selected = current
            current += 1

    def remove_scrollbars(self):
        self._scroll_enabled = False
        group.Group.remove(self, self._children[self._scroll_next_idx])
        group.Group.remove(self, self._children[self._scroll_prev_idx])
        self.reorder()

    def reorder(self):
        for child in self.children:
            position = self.get_item_position(child)
            idx = self.index(child)

            if self._scroll_enabled:
                child.visible = self.is_item_visible(idx)
            else:
                child.visible = True
                
            if self._scroll_enabled and (idx == 0 or idx == len(self)-1):
                child.y = position[1]
            else:
                child.position = position

    def is_item_visible(self, idx):
        if self.bg in self and idx == self.index(self.bg):
            return True
        visible = ((idx >= self._scroll_index) and \
                   (idx <= self._scroll_index + self._max_items - 1)) or \
                   (self._scroll_visible and \
                    ((idx == 0) \
                     or (idx == len(self._children) - 1))) # scrollbars
        return visible
    
    def next(self):
        if self._scroll_enabled and (self._selected + 1 < (len(self._children) - 2)):
            old_widget = self._children[self._selected + 1]
            #print old_widget.media
            self._selected += 1
            self.set_selected(self._selected)
            
            self.reorder()
            new_widget = self._children[self._selected + 1]
            self.active_changed.emit(old_widget, new_widget, 1);
        elif (not self._scroll_enabled) and (self._selected + 1 < len(self._children)):
            old_widget = self._children[self._selected]
            #self._selected += 1
            self.set_selected(self._selected + 1)
            
            self.reorder()
            new_widget = self._children[self._selected]
            self.active_changed.emit(old_widget, new_widget, 1);            
                                      
    def previous(self):
        first = 0
        if self.bg in self:
            first = 1
        #scroll index cannot be the scroll bar index
        if self._scroll_enabled and (self._selected > first): 
            old_widget = self._children[self._selected + 1]
            self._selected -= 1
            self.set_selected(self._selected)
            
            self.reorder()
            new_widget = self._children[self._selected + 1]
            self.active_changed.emit(old_widget, new_widget, -1);            
        elif (not self._scroll_enabled) and (self._selected >= 1):
            if self._children[self._selected-1] != self.bg:
                old_widget = self._children[self._selected]

                self.set_selected(self._selected - 1)

                self.reorder()
                new_widget = self._children[self._selected]
                self.active_changed.emit(old_widget, new_widget, -1);
    
    def index(self, child):
        # FIXME: add this in Group
        idx = self._children.index(child)

        return idx

    def __len__(self):
        # FIXME: add this in Group
        return len(self._children)

    def __getitem__(self, index):
        # FIXME: add this in Group
        return self._children[index]

    def set_key_actions(self):
        pgmrender = render.get_renderer()
        self.key_actions = {}
        for action, k in self.directions.iteritems():
            key_value = getattr(pgmrender, 'KEY_%s' % k.upper())
            self.key_actions[key_value] = getattr(self, '%s' % action)
        self.key_actions[pgmrender.KEY_RETURN] = self.fire_action
        
    def key_pressed(self, key_event):
        value = key_event.keyval
        action = self.key_actions.get(value)
        if action:
            action()

    def fire_action(self):
        if self._scroll_enabled and isinstance(self._children[self._selected+1], ActionItem):
            self._children[self._selected+1].fire_action()
        elif isinstance(self._children[self._selected], ActionItem):
            self._children[self._selected].fire_action()

class HList(List):
    """
    Horizontal List    
    """
    directions = {'next': u'right', 'previous': u'left'}

    def get_item_position(self, item):
        # TODO: self.bg support (see VList)
        index = self.index(item)
        total = len(self)

        if self._scroll_enabled:
            i = 0
            x = self.x

            while i < index:
                current_item = self.children[i]

                if self.is_item_visible(i):
                    x += current_item.width + self.spacing
                i += 1
        else:
            taken = 0
            for child in self.children:
                taken += child.width + self.spacing
            if taken:
                taken -= self.spacing

            remain = (self.width - taken) / 2.
            i = 0
            x = self.x + remain
            current_item = self.children[0]
            while i < index:
                x += current_item.width + self.spacing
                i += 1
                current_item = self.children[i]
            
        y = self.y
        return (x, y)


class VList(List):
    """
    Vertical list
    """
    directions = {'next': u'down', 'previous': u'up'}

    def get_item_position(self, item):
        index = self.index(item)
        total = len(self)

        if item == self.bg:
            return (self.x, self.y)

        if self._scroll_enabled:
            i = 0
            y = self.y

            while i < index:
                current_item = self.children[i]
                if current_item != self.bg and self.is_item_visible(i):
                    y += current_item.height + self.spacing
                i += 1
        else:
            taken = 0
            for child in self.children:
                if child != self.bg:
                    taken += child.height + self.spacing
            if taken and self.bg not in self:
                taken -= self.spacing

            remain = (self.height - taken) / 2.
            i = 0
            if remain >0:
                y = self.y + remain
            else:
                y = self.y
            current_item = self.children[0]
            while i < index:
                if current_item != self.bg:
                    y += current_item.height + self.spacing
                i += 1
                current_item = self.children[i]
                
        x = self.x

        if self.bg in self:
            # FIXME: don't hardcode item alignment values
            x += 8
            y += 6
        
        return (x, y)
