# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from pgm.timing import controller, modifier, keyframe
from pgm.layout import group
from elisa.core import log
from pgm.form import rectangle, image
import mutex

class PictureSelector(group.Group):
    
    logger = log.Logger()
    
    def __init__(self, size=(800,600), columns_count=5, item_apect_ratio=4/float(3)):
        group.Group.__init__(self)
        #storing a tuple of image_path and pigment image if exist and widget
        #(image_path, image, widget)
        self._items_pictures = []
        
        self._animation_mutex = mutex.mutex()
        self.size = size
        self._unselected_opacity = 100
        self._selected_item_index = 0
        width = size[0] / float(columns_count)
        self._item_size = ( width , width / float(item_apect_ratio) )
        self._visible_items = ( columns_count, int(size[1] / self._item_size[1]) )
        self._line_animation_offset = 0
        self._anim_up = False

    def get_visible_lines_index(self):
        current_line_index =  self._line_animation_offset
        last_line_index = current_line_index + self._visible_items[1] - 1
        return (current_line_index, last_line_index)
    
    def is_visible_item(self, index):
        current_line_index = self.get_cell_from_index(index)[1]
        start,end = self.get_visible_lines_index()
        if current_line_index in range( start,end +1):
            return True
        return False
        
    def lock_animation(self):
        return self._animation_mutex.testandset()
    
    def is_locked(self):    
        return self._animation_mutex.locked
    
    def unlock_animation(self):
        self._animation_mutex.unlock()
    
    def get_cell_from_index(self, index):
        line_index = index / self._visible_items[0]
        column_index = round( ( ( index / float(self._visible_items[0]) ) - line_index ) * self._visible_items[0])
        return column_index, line_index
    
    def get_index_from_cell(self, cell):
        #cell is (column_index, line_index)
        index = cell[1] * self._visible_items[0] + cell[0]
        return index
        
        
    def get_item_position(self, index):
        column_index, line_index = self.get_cell_from_index(index)
        pos = (column_index * self._item_size[0], (line_index - self._line_animation_offset) * self._item_size[1] )
        #print column_index ,line_index , pos
        return pos
        
    def add_item(self, picture_path):
        start_line, end_line = self.get_visible_lines_index()
        new_index = len(self._items_pictures)
        items_by_line = self._visible_items[0]
        
        #calculate index to create (take one line top and down ofscreen)
        first_index_to_create = (start_line - 1) * items_by_line
        last_index_to_create = (end_line + 1) * items_by_line + items_by_line - 1
        if new_index in range( first_index_to_create, last_index_to_create + 1):
            witem = rectangle.Rectangle(self._item_size)
            img = image.Image(picture_path)
            witem.media = img
            self._items_pictures.append( [picture_path, img, witem ] )
            if new_index != self._selected_item_index:
                witem.opacity = self._unselected_opacity
            witem.position = self.get_item_position(new_index)
            self.add(witem)
        else:
            self._items_pictures.append( [picture_path, None, None ] )
    
    def move_left(self):
        if self.lock_animation():
            if self._selected_item_index >0:
                next_selected_item_index = self._selected_item_index - 1
                self._set_selected_item_index(next_selected_item_index, True)
            else:
                self.unlock_animation()
    
    def move_right(self):
        if self.lock_animation():
            if self._selected_item_index < len(self._items_pictures) - 1:
                next_selected_item_index = self._selected_item_index + 1
                self._set_selected_item_index(next_selected_item_index, True)
            else:
                self.unlock_animation()
    
    def move_up(self):
        if self.lock_animation():
            (col, line) = self.get_cell_from_index(self._selected_item_index)
            new_index = self.get_index_from_cell( (col, line - 1) )
            if new_index >= 0:
                next_selected_item_index = self._selected_item_index - self._visible_items[0]
                self._set_selected_item_index(next_selected_item_index, True)
            else:
                self.unlock_animation()
        
    def move_down(self):
        if self.lock_animation():
            (col, line) = self.get_cell_from_index(self._selected_item_index)
            new_index = self.get_index_from_cell( (col, line + 1) )
            if len(self._items_pictures) > new_index:
                next_selected_item_index = self._selected_item_index + self._visible_items[0]
                if next_selected_item_index < len(self._items_pictures):
                    self._set_selected_item_index(next_selected_item_index, True)
                else:
                    self.unlock_animation()
            else:
                self.unlock_animation()
    
    def _vert_anim(self, anim_up):
        #anm_up is a boolean
        if anim_up:
            end_frame = keyframe.KeyFrame(self._item_size[1], 1.0)
            self._line_animation_offset -=1
        else:
            end_frame = keyframe.KeyFrame(-self._item_size[1], 1.0)
            self._line_animation_offset +=1
        
        timeline_pos = [keyframe.KeyFrame(0, 0.0), end_frame]
        mdf_pos = modifier.Modifier(self.children, 'y',
                                    timeline_pos, 
                                    operation=modifier.RELATIVE)
        ctrl = controller.Controller(begin=0, duration=250,
                                     resolution=15,
                                     end_behavior=controller.END,
                                     modifiers=[mdf_pos],
                                     end_callback=self._anim_end)
        ctrl.start()
    
    def _anim_end(self, ctrl):
        line_index_to_create = self.get_cell_from_index(self._selected_item_index)[1]
        if self._anim_up == False:
            line_index_to_create += 1
            line_index_to_delete = line_index_to_create - self._visible_items[1] - 2
        else:
            line_index_to_create -= 1
            line_index_to_delete = line_index_to_create + self._visible_items[1] + 2
        self._add_line(line_index_to_create)
        self._remove_line(line_index_to_delete)
        
        self.unlock_animation()
            
    def anim_end(self, anim_label):
        pass
    
    def _add_line(self, line_index):
        start_index = self.get_index_from_cell( (0, line_index) )
        end_index = start_index + self._visible_items[0] - 1
        for index in range(start_index, end_index + 1 ):
            if index < len(self._items_pictures) and index >= 0: 
                img = self._items_pictures[index][1]
                if not img:
                    img = image.Image(self._items_pictures[index][0])
                    self._items_pictures[index][1] = img
                witem = rectangle.Rectangle(self._item_size)    
                self._items_pictures[index][2] = witem
                witem.media = img
                witem.opacity = self._unselected_opacity
                witem.position = self.get_item_position(index)
                self.add(witem)
    
    def _remove_line(self, line_index):
        start_index = self.get_index_from_cell( (0, line_index) )
        end_index = start_index + self._visible_items[0] - 1
        for index in range(start_index, end_index + 1 ):
            if index < len(self._items_pictures) and index >= 0:
                witem = self._items_pictures[index][2]
                if witem:
                    self._items_pictures[index][2] = None
                    self.remove(witem)
                
        
    def _set_selected_item_index(self, index, auto_unlock_if_no_animation = False):
        #index must be near the current index !!!!
        start_animation = False
        self._items_pictures[self._selected_item_index][2].opacity = self._unselected_opacity
        if not self.is_visible_item(index):
            line_index_to_create = self.get_cell_from_index(index)[1]
            start_animation = True
            if index > self._selected_item_index:
                self._anim_up = False
            else:
                self._anim_up = True
            self._vert_anim(self._anim_up)
        self._selected_item_index = index
        self._items_pictures[self._selected_item_index][2].opacity = 255
        if not start_animation and auto_unlock_if_no_animation and self.is_locked:
            self.unlock_animation()

if __name__ == '__main__':  
    global ps
    from pgm import render, widget
    from pgm.layout import window
    from pgm.form import rectangle, image
    import sys
    
    try:
        from twisted.internet import glib2reactor
        glib2reactor.install()
    except AssertionError:
        # already installed...
        pass
    
    from twisted.internet import reactor, task
    from twisted.spread import pb
    import gobject
    
    
    def exit_event_cb(event):    
        win.main_quit()
        reactor.stop()
    
    def key_press_cb(event):
        global ps
        if event.keyval == 22:
            ps.move_right()
        elif event.keyval == 20:
            ps.move_left()
        elif event.keyval == 21:
            ps.move_up()
        elif event.keyval == 23:
            ps.move_down()
        elif event.keyval == render.get_renderer().KEY_q:
            reactor.stop()
            
    render.init('opengl1')
    title = 'picture selector test'
    global win, ps
    win = window.Window(title=title)
    

    
    #ps = PictureSelector(size=(800,120), columns_count = 10)
    #ps.position= (0,200)
    ps = PictureSelector()
    base ='elisa/core/'
    #base ='../'
    
    
    for i in range(0,30):
        ps.add_item(base + 'plugins/skins/default/pictures/256X256/remove.png')
        ps.add_item(base + 'plugins/skins/default/pictures/256X256/music.png')
        ps.add_item(base + 'plugins/skins/default/pictures/256X256/movies.png')
        ps.add_item(base + 'plugins/skins/default/pictures/256X256/config.png')

    win.add(ps)
                    
    win.sig_key_press_event.connect(key_press_cb)
    win.sig_exit_event.connect(exit_event_cb)
    gobject.timeout_add(15, win.update)

    win.main(reactor.loop)
 
    reactor.run()
    
    
