# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from pgm import widget
from pgm.timing import controller, modifier, keyframe
import widget_level

class WidgetLevelZoomed(widget_level.WidgetLevel):

    def __init__(self, arrow_image=None, visible_item_count=5, zoomed_items=3,
                 zoom_factor=2, width=800):
        self._enable_osx_animation = True
        self._zoomed_items = zoomed_items
        self._zoom_factor = zoom_factor
        widget_level.WidgetLevel.__init__(self, arrow_image,
                                          visible_item_count, width=width)
        
    def width__set(self, width):
        widget_level.WidgetLevel.width__set(self, width)
        if self._enable_osx_animation:
            self._item_width = self._get_base_item_width()
            self._item_height = self._item_width
            self._relative_xlist = self._get_relative_widget_x_list()


    def _get_zoom_factor_from_rank(self, relative_rank):
        #rank is a relative position
        first_zoomed_rank = (self._active_item_position - ( self._zoomed_items - 1 ) / 2 )
        last_zoomed_rank = (self._active_item_position + ( self._zoomed_items - 1 ) / 2 )
        partially_zoomed_items = self._active_item_position - first_zoomed_rank
        if relative_rank >= first_zoomed_rank and relative_rank < self._active_item_position:
            coef = ( relative_rank - first_zoomed_rank + 1 ) / float(partially_zoomed_items + 1)
            return (self._zoom_factor - 1) * coef + 1
        elif relative_rank <= last_zoomed_rank and relative_rank > self._active_item_position:
            coef = ( last_zoomed_rank -relative_rank + 1 ) / float(partially_zoomed_items + 1)
            return (self._zoom_factor - 1) * coef + 1
        elif relative_rank == self._active_item_position:
            return self._zoom_factor

        return 1

    def _get_base_item_width(self):
        x = 0
        for rank in range(self._visible_item_count):
            x += self._get_zoom_factor_from_rank(rank)
        return (self._width - self._space * (self._visible_item_count - 1)) / float(x)

    def _get_relative_widget_x_list(self):
        l = []
        l.append(0)
        pos = 0
        for rank in range(self._visible_item_count-1):
            zoom_factor = self._get_zoom_factor_from_rank(rank)
            pos += self._item_width * zoom_factor + self._space
            l.append(pos)
        return l

    def get_x_from_index(self, index):
        if not self._enable_osx_animation:
            return widget_level.WidgetLevel.get_x_from_index(self, index)

        xlist_rank = self.get_relative_index(index)
        xlist_size = len(self._relative_xlist)
        if xlist_rank > -1 and xlist_rank < xlist_size:
            x = self._relative_xlist[xlist_rank]
        elif xlist_rank < 0:
            x = (xlist_rank) * (self._item_width + self._space)
        else:
            x = ( xlist_size - xlist_rank + 1) * (self._item_width + self._space) + self._relative_xlist[xlist_size-1]

        return x

    def first_draw_level(self):
        #FIXME : animation is unlocked by parent call !!!
        widget_level.WidgetLevel.first_draw_level(self)
        if self._enable_osx_animation:
            #place and zoom items at start
            item_count = min(self._visible_item_count, self.get_widget_count())
            relative_rank = 0
            absolute_rank = self.get_relative_index(relative_rank)
            selected_rank = self.get_selected_rank()
            for w in self.get_widget_items( 0, item_count):
                w = self.get_widget_item(relative_rank)
                if not w:
                    continue
                w.x = self.get_x_from_index(relative_rank)
                zoom_factor = self._get_zoom_factor_from_rank(absolute_rank)
                w.size_mode = widget.UP_RIGHT
                size = w.size
                w.size = ( size[0] * zoom_factor, size[1] * zoom_factor)
                if relative_rank == selected_rank:
                    w.opacity = 255
                    w.hide_label()
                    w.show_label()
                else:
                    w.opacity = self._unselected_opacity
                relative_rank += 1
                absolute_rank += 1
            selected_item = self.get_widget_item(self._active_item_index)
            if selected_item:
                selected_item.show_label()
                self.parent.height = selected_item.size[1]

    def _draw_widget_item(self, witem, index):
        if witem:
            witem.opacity = self._unselected_opacity
            widget_level.WidgetLevel._draw_widget_item(self, witem, index)

    def _get_translate_modifier(self, widget_list, start_x, end_x, operation,
                                property='x'):
        timeline_pos = [ keyframe.KeyFrame(start_x, 0.0),
                         keyframe.KeyFrame(end_x, 1.0) ]
        return modifier.Modifier(widget_list, property,
                                 timeline_pos,
                                 operation=operation)

    def _get_size_modifier(self, widget_list, start_size, end_size):
        timeline_pos = [ keyframe.KeyFrame(start_size, 0.0),
                         keyframe.KeyFrame(end_size, 1.0) ]
        return modifier.Modifier(widget_list, 'size',
                                 timeline_pos,
                                 operation=modifier.ABSOLUTE)

    def _anim(self, direction):
        #interpolation=controller.LINEAR
        interpolation=controller.SINUSOIDAL
        relative_widget_list = []
        modifiers_list = []
        futur_unselected_widget = None
        futur_selected_widget = None
        first_absolute_index = self.get_first_visible_item_index()
        last_absolute_index = self.get_last_visible_item_index()
        count = last_absolute_index - first_absolute_index + 1

        if direction == 'right':
            next_rank_modifier = -1
            witem_offset = - (self._item_width + self._space)
        else:
            next_rank_modifier = 1
            witem_offset = self._item_width + self._space

        absolute_index = first_absolute_index
        relative_index = self.get_relative_index(absolute_index)
        selected_rank = self.get_selected_rank()

        for i in range(count):
            witem = self.get_widget_item(absolute_index)
            if witem:
                current_zoom_factor = self._get_zoom_factor_from_rank(relative_index)
                new_zoom_factor = self._get_zoom_factor_from_rank(relative_index + next_rank_modifier)
                zoom_factor = new_zoom_factor / float(current_zoom_factor)

                if zoom_factor == 1:
                    relative_widget_list.append(witem)
                else:
                    current_x = self.get_x_from_index(absolute_index)
                    new_x = self.get_x_from_index(absolute_index + next_rank_modifier)
                    mdf = self._get_translate_modifier([witem], current_x,
                                                       new_x, modifier.ABSOLUTE)
                    modifiers_list.append(mdf)
                    witem.size_mode = widget.UP_RIGHT
                    current_size = witem.size
                    new_width = current_size[0] * float(zoom_factor)
                    new_height = current_size[1] * float(zoom_factor)
                    mdf = self._get_size_modifier([witem], current_size,
                                                  (new_width, new_height))
                    modifiers_list.append(mdf)
                if absolute_index == selected_rank - next_rank_modifier:
                    futur_selected_widget = witem
                elif absolute_index == selected_rank:
                    futur_unselected_widget = witem
            absolute_index += 1
            relative_index += 1

        if len(relative_widget_list) > 0:
            mdf = self._get_translate_modifier(relative_widget_list, 0,
                                               witem_offset, modifier.RELATIVE)
            modifiers_list.append(mdf)
        if futur_selected_widget:
            mdf = self._get_translate_modifier([futur_selected_widget],
                                               self._unselected_opacity,
                                               255, modifier.ABSOLUTE, 'opacity')
            modifiers_list.append(mdf)

        if futur_unselected_widget:
            mdf = self._get_translate_modifier([futur_unselected_widget], 255,
                                               self._unselected_opacity,
                                               modifier.ABSOLUTE, 'opacity')
            modifiers_list.append(mdf)

        ctrl = controller.Controller(begin=0, duration=250,
                                     resolution=15,
                                     end_behavior=controller.END,
                                     modifiers=modifiers_list,
                                     end_callback=self._anim_end,
                                     interpolation=interpolation)
        ctrl.start()


    def _horiz_anim(self, direction):
        if self._enable_osx_animation:
            self._anim(direction)
        else:
            widget_level.WidgetLevel._horiz_anim(self, direction)

if __name__ == '__main__':
    from pgm import render
    from pgm.layout import window
    from pgm.form import rectangle, image
    import sys

    try:
        from twisted.internet import glib2reactor
        glib2reactor.install()
    except AssertionError:
        # already installed...
        pass

    from twisted.internet import reactor, task
    from twisted.spread import pb

    def refresh():
        win.dispatch_events()
        win.dispatch_controllers()
        win.update()

    def exit_event_cb(event):
        win.quit_loop()
        reactor.stop()

    def key_press_cb(event):
        if event.keyval == 22:
            wl.move_right()
        elif event.keyval == 20:
            wl.move_left()
        elif event.keyval == render.get_renderer().KEY_q:
            reactor.stop()


    render.init('opengl1')
    title = 'widget_level test'
    win = window.Window(title=title)

    win.start_loop()


    _background_widget = rectangle.Rectangle((800,600))
    img = image.Image('pictures/background.png')
    _background_widget.media = img
    _background_widget.ordering = widget.FAR
    win.add(_background_widget)


    wl = WidgetLevelZoomed('pictures/arrow.png',zoomed_items=5,visible_item_count=7)
    win.add(wl)
    wl.y = 150

    #wl.add_item(image.Image('pictures/256X256/pictures.png'))
    #wl.add_item(image.Image('pictures/256X256/dvd.png'))
    #wl.add_item(image.Image('pictures/256X256/movies.png'))
    #wl.add_item(image.Image('pictures/256X256/remove.png'))
    #wl.add_item(image.Image('pictures/256X256/music.png'))
    #wl.add_item(image.Image('pictures/256X256/play.png'))
    #wl.add_item(image.Image('pictures/256X256/pause.png'))
    #wl.add_item(image.Image('pictures/256X256/dvd.png'))
    #wl.add_item(image.Image('pictures/256X256/music.png'))
    #wl.add_item(image.Image('pictures/256X256/music.png'))
    #wl.add_item(image.Image('pictures/256X256/remove.png'))
    #wl.add_item('pictures/256X256/pause.png')
    #wl.add_item('pictures/256X256/play.png')
    #wl.add_item('pictures/256X256/generic-usb.png')
    #wl.add_item(image.Image('pictures/256X256/pictures.png'))

    for i in range(20):
        wl.add_item(image.Image('pictures/256X256/music.png'))

    wl.set_status(widget_level.NORMAL)
    wl.first_draw()

    _LoopingCall = task.LoopingCall( refresh )
    _LoopingCall.start( 1.0/60.0 )
    win.sig_key_press_event.connect(key_press_cb)
    win.sig_exit_event.connect(exit_event_cb)
    reactor.run()
